@@ -19,6 +19,7 @@
 #include "RecordTask.h"
 #include "Session.h"
 #include "Task.h"
+#include "core.h"
 #include "log.h"
 
 using namespace std;
@@ -39,45 +40,48 @@ static const char* trim_leading_blanks(const char* str) {
 }
 
 /**
- * The following helper is used to iterate over a tracee's memory
- * map.
+ * Returns true if a task in t's task-group other than t is doing an exec.
  */
-class KernelMapIterator {
-public:
-  KernelMapIterator(Task* t) : tid(t->tid) { init(); }
-  KernelMapIterator(pid_t tid) : tid(tid) { init(); }
-  ~KernelMapIterator() {
-    if (maps_file) {
-      fclose(maps_file);
+static bool thread_group_in_exec(Task* t) {
+  if (!t->session().is_recording()) {
+    return false;
+  }
+  for (Task* tt : t->thread_group()->task_set()) {
+    if (tt == t) {
+      continue;
+    }
+    RecordTask* rt = static_cast<RecordTask*>(tt);
+    Event& ev = rt->ev();
+    if (ev.is_syscall_event() &&
+        is_execve_syscall(ev.Syscall().number, ev.Syscall().arch())) {
+      return true;
     }
   }
+  return false;
+}
 
-  // It's very important to keep in mind that btrfs files can have the wrong
-  // device number!
-  const KernelMapping& current(string* raw_line = nullptr) {
-    if (raw_line) {
-      *raw_line = this->raw_line;
-    }
-    return km;
+KernelMapIterator::KernelMapIterator(Task* t) : tid(t->tid) {
+  // See https://lkml.org/lkml/2016/9/21/423
+  ASSERT(t, !thread_group_in_exec(t)) << "Task-group in execve, so reading "
+                                         "/proc/.../maps may trigger kernel "
+                                         "deadlock!";
+  init();
+}
+
+KernelMapIterator::~KernelMapIterator() {
+  if (maps_file) {
+    fclose(maps_file);
   }
-  bool at_end() { return !maps_file; }
-  void operator++();
+}
 
-private:
-  void init() {
-    char maps_path[PATH_MAX];
-    sprintf(maps_path, "/proc/%d/maps", tid);
-    if (!(maps_file = fopen(maps_path, "r"))) {
-      FATAL() << "Failed to open " << maps_path;
-    }
-    ++*this;
+void KernelMapIterator::init() {
+  char maps_path[PATH_MAX];
+  sprintf(maps_path, "/proc/%d/maps", tid);
+  if (!(maps_file = fopen(maps_path, "r"))) {
+    FATAL() << "Failed to open " << maps_path;
   }
-
-  pid_t tid;
-  FILE* maps_file;
-  string raw_line;
-  KernelMapping km;
-};
+  ++*this;
+}
 
 void KernelMapIterator::operator++() {
   char line[PATH_MAX * 2];
@@ -95,8 +99,8 @@ void KernelMapIterator::operator++() {
                              " %x:%x %" SCNu64 " %n",
                        &start, &end, flags, &offset, &dev_major, &dev_minor,
                        &inode, &chars_scanned);
-  assert(8 /*number of info fields*/ == nparsed ||
-         7 /*num fields if name is blank*/ == nparsed);
+  DEBUG_ASSERT(8 /*number of info fields*/ == nparsed ||
+               7 /*num fields if name is blank*/ == nparsed);
 
   // trim trailing newline, if any
   int last_char = strlen(line) - 1;
@@ -117,7 +121,10 @@ void KernelMapIterator::operator++() {
     char proc_exe[PATH_MAX];
     char exe[PATH_MAX];
     snprintf(proc_exe, sizeof(proc_exe), "/proc/%d/exe", tid);
-    readlink(proc_exe, exe, sizeof(exe));
+    ssize_t size = readlink(proc_exe, exe, sizeof(exe));
+    if (size < 0) {
+      FATAL() << "readlink failed";
+    }
     FATAL() << "Sorry, tracee " << tid << " has x86-64 image " << exe
             << " and that's not supported with a 32-bit rr.";
   }
@@ -128,6 +135,21 @@ void KernelMapIterator::operator++() {
   int f = (strchr(flags, 'p') ? MAP_PRIVATE : 0) |
           (strchr(flags, 's') ? MAP_SHARED : 0);
 
+  string tmp_name;
+  if (strchr(name, '\\')) {
+    // Unescape any '\012' sequences
+    while (*name) {
+      if (strncmp(name, "\\012", 4) == 0) {
+        tmp_name.push_back('\n');
+        name += 4;
+      } else {
+        tmp_name.push_back(*name);
+        ++name;
+      }
+    }
+    name = tmp_name.c_str();
+  }
+
   km = KernelMapping(start, end, name, MKDEV(dev_major, dev_minor), inode, prot,
                      f, offset);
 }
@@ -193,7 +215,17 @@ AddressSpace::Mapping::Mapping(const Mapping& other)
 
 AddressSpace::Mapping::~Mapping() {}
 
-AddressSpace::~AddressSpace() { session_->on_destroy(this); }
+AddressSpace::~AddressSpace() {
+  for (auto& m : mem) {
+    if (m.second.local_addr) {
+      int ret = munmap(m.second.local_addr, m.second.map.size());
+      if (ret < 0) {
+        FATAL() << "Can't munmap";
+      }
+    }
+  }
+  session_->on_destroy(this);
+}
 
 void AddressSpace::after_clone() { allocate_watchpoints(); }
 
@@ -294,6 +326,7 @@ void AddressSpace::map_rr_page(AutoRemoteSyscalls& remote) {
 
     map(t, rr_page_start(), rr_page_size(), prot, flags, 0, file_name, 0, 0);
   }
+  mapping_flags_of(rr_page_start()) = Mapping::IS_RR_PAGE;
 
   if (t->session().is_recording()) {
     // brk() will not have been called yet so the brk area is empty.
@@ -330,7 +363,11 @@ static const AddressSpace::SyscallType entry_points[] = {
     AddressSpace::RECORDING_ONLY },
 };
 
-static remote_code_ptr ip_from_index(size_t i) {
+static remote_code_ptr entry_ip_from_index(size_t i) {
+  return remote_code_ptr(RR_PAGE_ADDR + RR_PAGE_SYSCALL_STUB_SIZE * i);
+}
+
+static remote_code_ptr exit_ip_from_index(size_t i) {
   return remote_code_ptr(RR_PAGE_ADDR + RR_PAGE_SYSCALL_STUB_SIZE * i +
                          RR_PAGE_SYSCALL_INSTRUCTION_END);
 }
@@ -341,17 +378,7 @@ remote_code_ptr AddressSpace::rr_page_syscall_exit_point(Traced traced,
   for (auto& e : entry_points) {
     if (e.traced == traced && e.privileged == privileged &&
         e.enabled == enabled) {
-      return ip_from_index(&e - entry_points);
-    }
-  }
-  return nullptr;
-}
-
-const AddressSpace::SyscallType* AddressSpace::rr_page_syscall_from_exit_point(
-    remote_code_ptr ip) {
-  for (size_t i = 0; i < array_length(entry_points); ++i) {
-    if (ip_from_index(i) == ip) {
-      return &entry_points[i];
+      return exit_ip_from_index(&e - entry_points);
     }
   }
   return nullptr;
@@ -360,10 +387,34 @@ const AddressSpace::SyscallType* AddressSpace::rr_page_syscall_from_exit_point(
 remote_code_ptr AddressSpace::rr_page_syscall_entry_point(Traced traced,
                                                           Privileged privileged,
                                                           Enabled enabled,
-                                                          SupportedArch arch) {
-  remote_code_ptr ip = rr_page_syscall_exit_point(traced, privileged, enabled);
-  return ip.is_null() ? remote_code_ptr()
-                      : ip.decrement_by_syscall_insn_length(arch);
+                                                          SupportedArch) {
+  for (auto& e : entry_points) {
+    if (e.traced == traced && e.privileged == privileged &&
+        e.enabled == enabled) {
+      return entry_ip_from_index(&e - entry_points);
+    }
+  }
+  return nullptr;
+}
+
+const AddressSpace::SyscallType* AddressSpace::rr_page_syscall_from_exit_point(
+    remote_code_ptr ip) {
+  for (size_t i = 0; i < array_length(entry_points); ++i) {
+    if (exit_ip_from_index(i) == ip) {
+      return &entry_points[i];
+    }
+  }
+  return nullptr;
+}
+
+const AddressSpace::SyscallType* AddressSpace::rr_page_syscall_from_entry_point(
+    remote_code_ptr ip) {
+  for (size_t i = 0; i < array_length(entry_points); ++i) {
+    if (entry_ip_from_index(i) == ip) {
+      return &entry_points[i];
+    }
+  }
+  return nullptr;
 }
 
 vector<AddressSpace::SyscallType> AddressSpace::rr_page_syscalls() {
@@ -382,7 +433,7 @@ template <typename Arch> static vector<uint8_t> read_auxv_arch(Task* t) {
 
   // Check final NULL in argv
   auto null_ptr = t->read_mem(stack_ptr);
-  assert(null_ptr == 0);
+  ASSERT(t, null_ptr == 0);
   stack_ptr++;
 
   // Should now point to envp
@@ -559,7 +610,7 @@ KernelMapping AddressSpace::map(Task* t, remote_ptr<void> addr,
                                 EmuFile::shr_ptr emu_file, void* local_addr,
                                 shared_ptr<MonitoredSharedMemory>&& monitored) {
   LOG(debug) << "mmap(" << addr << ", " << num_bytes << ", " << HEX(prot)
-             << ", " << HEX(flags) << ", " << HEX(offset_bytes);
+             << ", " << HEX(flags) << ", " << HEX(offset_bytes) << ")";
   num_bytes = ceil_page_size(num_bytes);
   KernelMapping m(addr, addr + num_bytes, fsname, device, inode, prot, flags,
                   offset_bytes);
@@ -581,13 +632,6 @@ KernelMapping AddressSpace::map(Task* t, remote_ptr<void> addr,
   map_and_coalesce(t, m, actual_recorded_map, emu_file, move(mapped_file_stat),
                    move(local_addr), move(monitored));
 
-  if ((prot & PROT_EXEC) &&
-      (fsname.find(SYSCALLBUF_LIB_FILENAME) != string::npos ||
-       fsname.find(SYSCALLBUF_LIB_FILENAME_32) != string::npos)) {
-    syscallbuf_lib_start_ = addr;
-    syscallbuf_lib_end_ = addr + num_bytes;
-  }
-
   // During an emulated exec, we will explicitly map in a (copy of) the VDSO
   // at the recorded address.
   if (actual_recorded_map.is_vdso()) {
@@ -602,8 +646,9 @@ template <typename Arch> void AddressSpace::at_preload_init_arch(Task* t) {
       remote_ptr<rrcall_init_preload_params<Arch>>(t->regs().arg1()));
 
   if (t->session().is_recording()) {
-    ASSERT(t, t->session().as_record()->use_syscall_buffer() ==
-                  params.syscallbuf_enabled)
+    ASSERT(t,
+           t->session().as_record()->use_syscall_buffer() ==
+               params.syscallbuf_enabled)
         << "Tracee thinks syscallbuf is "
         << (params.syscallbuf_enabled ? "en" : "dis")
         << "abled, but tracer thinks "
@@ -623,17 +668,15 @@ template <typename Arch> void AddressSpace::at_preload_init_arch(Task* t) {
 }
 
 void AddressSpace::at_preload_init(Task* t) {
-  ASSERT(t, !syscallbuf_lib_start_.is_null())
-      << "should have found preload library already";
   RR_ARCH_FUNCTION(at_preload_init_arch, t->arch(), t);
 }
 
 const AddressSpace::Mapping& AddressSpace::mapping_of(
     remote_ptr<void> addr) const {
-  MemoryRange range(floor_page_size(addr), page_size());
+  MemoryRange range(floor_page_size(addr), 1);
   auto it = mem.find(range);
-  assert(it != mem.end());
-  assert(it->second.map.contains(range));
+  DEBUG_ASSERT(it != mem.end());
+  DEBUG_ASSERT(it->second.map.contains(range));
   return it->second;
 }
 uint32_t& AddressSpace::mapping_flags_of(remote_ptr<void> addr) {
@@ -642,16 +685,48 @@ uint32_t& AddressSpace::mapping_flags_of(remote_ptr<void> addr) {
       .flags;
 }
 
+uint8_t* AddressSpace::local_mapping(remote_ptr<void> addr, size_t size) {
+  MemoryRange range(floor_page_size(addr), 1);
+  auto it = mem.find(range);
+  if (it == mem.end()) {
+    return nullptr;
+  }
+  DEBUG_ASSERT(it->second.map.contains(range));
+  const Mapping& map = it->second;
+  // Fall back to the slow path if we can't get the entire region
+  if (size > static_cast<size_t>(map.map.end() - addr)) {
+    return nullptr;
+  }
+  if (map.local_addr != nullptr) {
+    size_t offset = addr - map.map.start();
+    return static_cast<uint8_t*>(map.local_addr) + offset;
+  }
+  return nullptr;
+}
+
+void* AddressSpace::detach_local_mapping(remote_ptr<void> addr) {
+  auto m = const_cast<AddressSpace::Mapping&>(mapping_of(addr));
+  void* p = m.local_addr;
+  m.local_addr = nullptr;
+  return p;
+}
+
 bool AddressSpace::has_mapping(remote_ptr<void> addr) const {
   if (addr + page_size() < addr) {
     // Assume the last byte in the address space is never mapped; avoid overflow
     return false;
   }
-  MemoryRange m(floor_page_size(addr), page_size());
+  MemoryRange m(floor_page_size(addr), 1);
   auto it = mem.find(m);
   return it != mem.end() && it->first.contains(m);
 }
 
+bool AddressSpace::has_rr_page() const {
+  MemoryRange m(RR_PAGE_ADDR, 1);
+  auto it = mem.find(m);
+  return it != mem.end() && (it->second.flags & Mapping::IS_RR_PAGE);
+}
+
 void AddressSpace::protect(Task* t, remote_ptr<void> addr, size_t num_bytes,
                            int prot) {
   LOG(debug) << "mprotect(" << addr << ", " << num_bytes << ", " << HEX(prot)
@@ -759,7 +834,7 @@ void AddressSpace::remap(Task* t, remote_ptr<void> old_addr,
              << new_addr << ", " << new_num_bytes << ")";
 
   Mapping mr = mapping_of(old_addr);
-  assert(!mr.monitored_shared_memory);
+  DEBUG_ASSERT(!mr.monitored_shared_memory);
   const KernelMapping& m = mr.map;
 
   old_num_bytes = ceil_page_size(old_num_bytes);
@@ -772,7 +847,7 @@ void AddressSpace::remap(Task* t, remote_ptr<void> old_addr,
   auto it = dont_fork.lower_bound(MemoryRange(old_addr, old_num_bytes));
   if (it != dont_fork.end() && it->start() < old_addr + old_num_bytes) {
     // mremap fails if some but not all pages are marked DONTFORK
-    assert(*it == MemoryRange(old_addr, old_num_bytes));
+    DEBUG_ASSERT(*it == MemoryRange(old_addr, old_num_bytes));
     remove_range(dont_fork, MemoryRange(old_addr, old_num_bytes));
     add_range(dont_fork, MemoryRange(new_addr, new_num_bytes));
   } else {
@@ -807,10 +882,11 @@ bool AddressSpace::add_breakpoint(remote_code_ptr addr, BreakpointType type) {
                                sizeof(overwritten_data), &overwritten_data)) {
       return false;
     }
-    t->write_mem(addr.to_data_ptr<uint8_t>(), breakpoint_insn);
+    t->write_mem(addr.to_data_ptr<uint8_t>(), breakpoint_insn, nullptr,
+                 Task::IS_BREAKPOINT_RELATED);
 
     auto it_and_is_new = breakpoints.insert(make_pair(addr, Breakpoint()));
-    assert(it_and_is_new.second);
+    DEBUG_ASSERT(it_and_is_new.second);
     it_and_is_new.first->second.overwritten_data = overwritten_data;
     it = it_and_is_new.first;
   }
@@ -854,9 +930,25 @@ int AddressSpace::access_bits_of(WatchType type) {
   }
 }
 
+/**
+ * We do not allow a watchpoint to watch the last byte of memory addressable
+ * by rr. This avoids constructing a MemoryRange that wraps around.
+ * For 64-bit builds this is no problem because addresses at the top of memory
+ * are in kernel space. For 32-bit builds it seems impossible to map the last
+ * page of memory in Linux so we should be OK there too.
+ * Note that zero-length watchpoints are OK. configure_watch_registers just
+ * ignores them.
+ */
+static MemoryRange range_for_watchpoint(remote_ptr<void> addr,
+                                        size_t num_bytes) {
+  uintptr_t p = addr.as_int();
+  uintptr_t max_len = UINTPTR_MAX - p;
+  return MemoryRange(addr, min<uintptr_t>(num_bytes, max_len));
+}
+
 void AddressSpace::remove_watchpoint(remote_ptr<void> addr, size_t num_bytes,
                                      WatchType type) {
-  auto it = watchpoints.find(MemoryRange(addr, num_bytes));
+  auto it = watchpoints.find(range_for_watchpoint(addr, num_bytes));
   if (it != watchpoints.end() &&
       0 == it->second.unwatch(access_bits_of(type))) {
     watchpoints.erase(it);
@@ -866,12 +958,12 @@ void AddressSpace::remove_watchpoint(remote_ptr<void> addr, size_t num_bytes,
 
 bool AddressSpace::add_watchpoint(remote_ptr<void> addr, size_t num_bytes,
                                   WatchType type) {
-  MemoryRange key(addr, num_bytes);
+  MemoryRange key = range_for_watchpoint(addr, num_bytes);
   auto it = watchpoints.find(key);
   if (it == watchpoints.end()) {
     auto it_and_is_new =
         watchpoints.insert(make_pair(key, Watchpoint(num_bytes)));
-    assert(it_and_is_new.second);
+    DEBUG_ASSERT(it_and_is_new.second);
     it = it_and_is_new.first;
     update_watchpoint_value(it->first, it->second);
   }
@@ -884,7 +976,7 @@ void AddressSpace::save_watchpoints() {
 }
 
 bool AddressSpace::restore_watchpoints() {
-  assert(!saved_watchpoints.empty());
+  DEBUG_ASSERT(!saved_watchpoints.empty());
   watchpoints = saved_watchpoints[saved_watchpoints.size() - 1];
   saved_watchpoints.pop_back();
   return allocate_watchpoints();
@@ -961,8 +1053,11 @@ bool AddressSpace::notify_watchpoint_fired(uintptr_t debug_status) {
   return triggered;
 }
 
-void AddressSpace::notify_written(remote_ptr<void> addr, size_t num_bytes) {
-  update_watchpoint_values(addr, addr + num_bytes);
+void AddressSpace::notify_written(remote_ptr<void> addr, size_t num_bytes,
+                                  uint32_t flags) {
+  if (!(flags & Task::IS_BREAKPOINT_RELATED)) {
+    update_watchpoint_values(addr, addr + num_bytes);
+  }
   session()->accumulate_bytes_written(num_bytes);
 }
 
@@ -1021,6 +1116,14 @@ void AddressSpace::unmap_internal(Task*, remote_ptr<void> addr,
       overflow.flags = m.flags;
       add_to_map(overflow);
     }
+
+    if (m.local_addr) {
+      int ret =
+          munmap(m.local_addr + (rem.start() - m.map.start()), rem.size());
+      if (ret < 0) {
+        FATAL() << "Can't munmap";
+      }
+    }
   };
   for_each_in_range(addr, num_bytes, unmapper);
   update_watchpoint_values(addr, addr + num_bytes);
@@ -1197,7 +1300,7 @@ static void assert_segments_match(Task* t, const KernelMapping& input_m,
 }
 
 KernelMapping AddressSpace::vdso() const {
-  assert(!vdso_start_addr.is_null());
+  DEBUG_ASSERT(!vdso_start_addr.is_null());
   return mapping_of(vdso_start_addr).map;
 }
 
@@ -1209,6 +1312,10 @@ KernelMapping AddressSpace::vdso() const {
 void AddressSpace::verify(Task* t) const {
   ASSERT(t, task_set().end() != task_set().find(t));
 
+  if (thread_group_in_exec(t)) {
+    return;
+  }
+
   MemoryMap::const_iterator mem_it = mem.begin();
   KernelMapIterator kernel_it(t);
   while (!kernel_it.at_end() && mem_it != mem.end()) {
@@ -1267,7 +1374,7 @@ AddressSpace::AddressSpace(Task* t, const string& exe, uint32_t exec_count)
   // https://github.com/mozilla/rr/issues/1113 .
   if (session_->done_initial_exec()) {
     populate_address_space(t);
-    assert(!vdso_start_addr.is_null());
+    DEBUG_ASSERT(!vdso_start_addr.is_null());
   } else {
     // Setup traced_syscall_ip_ now because we need to do AutoRemoteSyscalls
     // (for open_mem_fd) before the first exec. We rely on the fact that we
@@ -1286,6 +1393,7 @@ AddressSpace::AddressSpace(Session* session, const AddressSpace& o,
       brk_start(o.brk_start),
       brk_end(o.brk_end),
       mem(o.mem),
+      shm_sizes(o.shm_sizes),
       monitored_mem(o.monitored_mem),
       session_(session),
       vdso_start_addr(o.vdso_start_addr),
@@ -1294,8 +1402,6 @@ AddressSpace::AddressSpace(Session* session, const AddressSpace& o,
                             : nullptr),
       traced_syscall_ip_(o.traced_syscall_ip_),
       privileged_traced_syscall_ip_(o.privileged_traced_syscall_ip_),
-      syscallbuf_lib_start_(o.syscallbuf_lib_start_),
-      syscallbuf_lib_end_(o.syscallbuf_lib_end_),
       syscallbuf_enabled_(o.syscallbuf_enabled_),
       saved_auxv_(o.saved_auxv_),
       first_run_event_(0) {
@@ -1320,7 +1426,16 @@ AddressSpace::AddressSpace(Session* session, const AddressSpace& o,
 }
 
 void AddressSpace::post_vm_clone(Task* t) {
-  // Recreate preload_thread_locals mapping.
+  if (has_mapping(preload_thread_locals_start()) &&
+      (mapping_flags_of(preload_thread_locals_start()) &
+       AddressSpace::Mapping::IS_THREAD_LOCALS) == 0) {
+    // The tracee already has a mapping at this address that doesn't belong to
+    // us. Don't touch it.
+    return;
+  }
+
+  // Otherwise, the preload_thread_locals mapping is non-existent or ours.
+  // Recreate it.
   AutoRemoteSyscalls remote(t);
   t->session().create_shared_mmap(remote, PRELOAD_THREAD_LOCALS_SIZE,
                                   preload_thread_locals_start(),
@@ -1347,7 +1462,7 @@ static vector<MemoryRange> split_range(const MemoryRange& range) {
         !try_split_unaligned_range(r, 4, result) &&
         !try_split_unaligned_range(r, 2, result)) {
       bool ret = try_split_unaligned_range(r, 1, result);
-      assert(ret);
+      DEBUG_ASSERT(ret);
     }
   }
   return result;
@@ -1356,6 +1471,7 @@ static vector<MemoryRange> split_range(const MemoryRange& range) {
 static void configure_watch_registers(vector<WatchConfig>& regs,
                                       const MemoryRange& range, WatchType type,
                                       vector<int8_t>* assigned_regs) {
+  // Zero-sized WatchConfigs return no ranges here, so are ignored.
   auto split_ranges = split_range(range);
 
   if (type == WATCH_WRITE && range.size() > 1) {
@@ -1494,8 +1610,9 @@ static inline void assert_coalesceable(Task* t,
                                        const AddressSpace::Mapping& higher) {
   ASSERT(t, lower.emu_file == higher.emu_file);
   ASSERT(t, lower.flags == higher.flags);
-  ASSERT(t, (lower.local_addr == 0 && higher.local_addr == 0) ||
-                lower.local_addr + lower.map.size() == higher.local_addr);
+  ASSERT(t,
+         (lower.local_addr == 0 && higher.local_addr == 0) ||
+             lower.local_addr + lower.map.size() == higher.local_addr);
   ASSERT(t, !lower.monitored_shared_memory && !higher.monitored_shared_memory);
 }
 
@@ -1550,7 +1667,7 @@ void AddressSpace::coalesce_around(Task* t, MemoryMap::iterator it) {
   mem.erase(first_kv, ++last_kv);
 
   auto ins = mem.insert(MemoryMap::value_type(new_m.map, new_m));
-  assert(ins.second); // key didn't already exist
+  DEBUG_ASSERT(ins.second); // key didn't already exist
 }
 
 void AddressSpace::destroy_breakpoint(BreakpointMap::const_iterator it) {
@@ -1558,9 +1675,10 @@ void AddressSpace::destroy_breakpoint(BreakpointMap::const_iterator it) {
     return;
   }
   Task* t = *task_set().begin();
-  LOG(debug) << "Writing back " << std::hex << (int)it->second.overwritten_data
-             << std::dec;
-  t->write_mem(it->first.to_data_ptr<uint8_t>(), it->second.overwritten_data);
+  auto ptr = it->first.to_data_ptr<uint8_t>();
+  auto data = it->second.overwritten_data;
+  LOG(debug) << "Writing back " << HEX(data) << " at " << ptr;
+  t->write_mem(ptr, data, nullptr, Task::IS_BREAKPOINT_RELATED);
   breakpoints.erase(it);
 }
 
@@ -1705,7 +1823,8 @@ void AddressSpace::populate_address_space(Task* t) {
 static int random_addr_bits(SupportedArch arch) {
   switch (arch) {
     default:
-      assert(0 && "Unknown architecture");
+      DEBUG_ASSERT(0 && "Unknown architecture");
+      RR_FALLTHROUGH;
     case x86:
       return 32;
     // Current x86-64 systems have only 48 bits of virtual address space,
