@@ -4,13 +4,13 @@
 
 #include "Command.h"
 
-#include <assert.h>
 #include <stdio.h>
 #include <string.h>
 
 #include <algorithm>
 
 #include "TraceStream.h"
+#include "core.h"
 #include "main.h"
 
 using namespace std;
@@ -74,6 +74,7 @@ void Command::print_help_all(FILE* out) {
 void Command::print_help(FILE* out) {
   if (help) {
     fputs(help, out);
+    print_global_options(out);
   } else {
     print_usage(out);
   }
@@ -115,6 +116,10 @@ bool Command::parse_option(std::vector<std::string>& args,
           }
           return false;
         case HAS_PARAMETER:
+          if (args[0][2] == '=') {
+            assign_param(out, args[0].c_str() + 3);
+            return consume_args(args, 1);
+          }
           if (args[0][2] != 0) {
             assign_param(out, args[0].c_str() + 2);
             return consume_args(args, 1);
@@ -125,13 +130,12 @@ bool Command::parse_option(std::vector<std::string>& args,
           }
           return false;
         default:
-          assert(0 && "Unknown parameter type");
+          DEBUG_ASSERT(0 && "Unknown parameter type");
       }
     } else if (args[0][1] == '-') {
       size_t equals = args[0].find('=');
       if (strncmp(args[0].c_str() + 2, option_specs[i].long_name,
-                  (equals == string::npos ? args[0].length() : equals) - 2) ==
-          0) {
+                  (equals == string::npos ? 9999 : equals) - 2) == 0) {
         out->short_name = option_specs[i].short_name;
         switch (option_specs[i].param) {
           case NO_PARAMETER:
@@ -147,7 +151,7 @@ bool Command::parse_option(std::vector<std::string>& args,
             assign_param(out, args[0].c_str() + equals + 1);
             return consume_args(args, 1);
           default:
-            assert(0 && "Unknown parameter type");
+            DEBUG_ASSERT(0 && "Unknown parameter type");
         }
       }
     }
@@ -177,4 +181,13 @@ bool Command::parse_optional_trace_dir(vector<string>& args, string* out) {
   return true;
 }
 
+bool Command::parse_literal(std::vector<std::string>& args, const char* lit) {
+  if (args.size() > 0 && args[0] == lit) {
+    args.erase(args.begin());
+    return true;
+  } else {
+    return false;
+  }
+}
+
 } // namespace rr
