@@ -4,6 +4,7 @@
 
 #include "AutoRemoteSyscalls.h"
 #include "ReplaySession.h"
+#include "core.h"
 #include "kernel_metadata.h"
 #include "log.h"
 
@@ -19,8 +20,8 @@ DiversionSession::~DiversionSession() {
   // destroyed many times, and we don't want to temporarily hog
   // resources.
   kill_all_tasks();
-  assert(tasks().size() == 0 && vms().size() == 0);
-  assert(emu_fs->size() == 0);
+  DEBUG_ASSERT(tasks().size() == 0 && vms().size() == 0);
+  DEBUG_ASSERT(emu_fs->size() == 0);
 }
 
 static void finish_emulated_syscall_with_ret(Task* t, long ret) {
@@ -48,7 +49,7 @@ static void execute_syscall(Task* t) {
 
 template <typename Arch>
 static void process_syscall_arch(Task* t, int syscallno) {
-  LOG(debug) << "Processing " << t->syscall_name(syscallno);
+  LOG(debug) << "Processing " << syscall_name(syscallno, Arch::arch());
 
   if (syscallno == Arch::ioctl && t->is_desched_event_syscall()) {
     // The arm/disarm-desched ioctls are emulated as no-ops.
@@ -94,7 +95,7 @@ static void process_syscall_arch(Task* t, int syscallno) {
   return execute_syscall(t);
 }
 
-static void process_syscall(Task* t, int syscallno) {
+static void process_syscall(Task* t, int syscallno){
   RR_ARCH_FUNCTION(process_syscall_arch, t->arch(), t, syscallno)
 }
 
@@ -104,7 +105,7 @@ static void process_syscall(Task* t, int syscallno) {
  */
 DiversionSession::DiversionResult DiversionSession::diversion_step(
     Task* t, RunCommand command, int signal_to_deliver) {
-  assert(command != RUN_SINGLESTEP_FAST_FORWARD);
+  DEBUG_ASSERT(command != RUN_SINGLESTEP_FAST_FORWARD);
   assert_fully_initialized();
 
   DiversionResult result;
@@ -150,8 +151,9 @@ DiversionSession::DiversionResult DiversionSession::diversion_step(
                << "; break=" << result.break_status.breakpoint_hit
                << ", watch=" << !result.break_status.watchpoints_hit.empty()
                << ", singlestep=" << result.break_status.singlestep_complete;
-    ASSERT(t, !result.break_status.singlestep_complete ||
-                  command == RUN_SINGLESTEP);
+    ASSERT(t,
+           !result.break_status.singlestep_complete ||
+               command == RUN_SINGLESTEP);
     return result;
   }
 
