@@ -3,41 +3,46 @@
 #include "GdbCommand.h"
 
 #include "ReplayTask.h"
+#include "log.h"
 
 using namespace std;
 
 namespace rr {
 
-static SimpleGdbCommand when("when", [](GdbServer&, Task* t,
-                                        const vector<string>&) {
-  if (!t->session().is_replaying()) {
-    return string("Current event not known (diversion?)");
-  }
-  return string("Current event: ") +
-         to_string(static_cast<ReplayTask*>(t)->current_trace_frame().time());
-});
+static SimpleGdbCommand when(
+    "when", "Print the current rr event number.",
+    [](GdbServer&, Task* t, const vector<string>&) {
+      if (!t->session().is_replaying()) {
+        return GdbCommandHandler::cmd_end_diversion();
+      }
+      return string("Current event: ") +
+             to_string(
+                 static_cast<ReplayTask*>(t)->current_trace_frame().time());
+    });
 
-static SimpleGdbCommand when_ticks("when-ticks", [](GdbServer&, Task* t,
-                                                    const vector<string>&) {
-  if (!t->session().is_replaying()) {
-    return string("Current event not known (diversion?)");
-  }
-  return string("Current tick: ") + to_string(t->tick_count());
-});
+static SimpleGdbCommand when_ticks(
+    "when-ticks", "Print the current rr tick count for the current thread.",
+    [](GdbServer&, Task* t, const vector<string>&) {
+      if (!t->session().is_replaying()) {
+        return GdbCommandHandler::cmd_end_diversion();
+      }
+      return string("Current tick: ") + to_string(t->tick_count());
+    });
 
-static SimpleGdbCommand when_tid("when-tid", [](GdbServer&, Task* t,
-                                                const vector<string>&) {
-  if (!t->session().is_replaying()) {
-    return string("Current event not known (diversion?)");
-  }
-  return string("Current tid: ") + to_string(t->tid);
-});
+static SimpleGdbCommand when_tid(
+    "when-tid", "Print the real tid for the current thread.",
+    [](GdbServer&, Task* t, const vector<string>&) {
+      if (!t->session().is_replaying()) {
+        return GdbCommandHandler::cmd_end_diversion();
+      }
+      return string("Current tid: ") + to_string(t->tid);
+    });
 
 static std::vector<ReplayTimeline::Mark> back_stack;
 static ReplayTimeline::Mark current_history_cp;
 static std::vector<ReplayTimeline::Mark> forward_stack;
 static SimpleGdbCommand rr_history_push(
-    "rr-history-push",
+    "rr-history-push", "Push an entry into the rr history.",
     [](GdbServer& gdb_server, Task* t, const vector<string>&) {
       if (!t->session().is_replaying()) {
         // Don't create new history state inside a diversion
@@ -50,34 +55,36 @@ static SimpleGdbCommand rr_history_push(
       forward_stack.clear();
       return string();
     });
-static SimpleGdbCommand back("back", [](GdbServer& gdb_server, Task* t,
-                                        const vector<string>&) {
-  if (!t->session().is_replaying()) {
-    return GdbCommandHandler::cmd_end_diversion();
-  }
-  if (back_stack.size() == 0) {
-    return string("Can't go back. No more history entries.");
-  }
-  forward_stack.push_back(current_history_cp);
-  current_history_cp = back_stack.back();
-  back_stack.pop_back();
-  gdb_server.get_timeline().seek_to_mark(current_history_cp);
-  return string();
-});
-static SimpleGdbCommand forward("forward", [](GdbServer& gdb_server, Task* t,
-                                              const vector<string>&) {
-  if (!t->session().is_replaying()) {
-    return GdbCommandHandler::cmd_end_diversion();
-  }
-  if (forward_stack.size() == 0) {
-    return string("Can't go forward. No more history entries.");
-  }
-  back_stack.push_back(current_history_cp);
-  current_history_cp = forward_stack.back();
-  forward_stack.pop_back();
-  gdb_server.get_timeline().seek_to_mark(current_history_cp);
-  return string();
-});
+static SimpleGdbCommand back(
+    "back", "Go back one entry in the rr history.",
+    [](GdbServer& gdb_server, Task* t, const vector<string>&) {
+      if (!t->session().is_replaying()) {
+        return GdbCommandHandler::cmd_end_diversion();
+      }
+      if (back_stack.size() == 0) {
+        return string("Can't go back. No more history entries.");
+      }
+      forward_stack.push_back(current_history_cp);
+      current_history_cp = back_stack.back();
+      back_stack.pop_back();
+      gdb_server.get_timeline().seek_to_mark(current_history_cp);
+      return string();
+    });
+static SimpleGdbCommand forward(
+    "forward", "Go forward one entry in the rr history.",
+    [](GdbServer& gdb_server, Task* t, const vector<string>&) {
+      if (!t->session().is_replaying()) {
+        return GdbCommandHandler::cmd_end_diversion();
+      }
+      if (forward_stack.size() == 0) {
+        return string("Can't go forward. No more history entries.");
+      }
+      back_stack.push_back(current_history_cp);
+      current_history_cp = forward_stack.back();
+      forward_stack.pop_back();
+      gdb_server.get_timeline().seek_to_mark(current_history_cp);
+      return string();
+    });
 
 static int gNextCheckpointId = 0;
 
@@ -95,7 +102,7 @@ string invoke_checkpoint(GdbServer& gdb_server, Task*,
       gdb_server.timeline, gdb_server.last_continue_tuid, e, where);
   return string("Checkpoint ") + to_string(checkpoint_id) + " at " + where;
 }
-static SimpleGdbCommand checkpoint("checkpoint", invoke_checkpoint);
+static SimpleGdbCommand checkpoint("checkpoint", "", invoke_checkpoint);
 
 string invoke_delete_checkpoint(GdbServer& gdb_server, Task*,
                                 const vector<string>& args) {
@@ -111,7 +118,7 @@ string invoke_delete_checkpoint(GdbServer& gdb_server, Task*,
     return string("No checkpoint number ") + to_string(id) + ".";
   }
 }
-static SimpleGdbCommand delete_checkpoint("delete checkpoint",
+static SimpleGdbCommand delete_checkpoint("delete checkpoint", "",
                                           invoke_delete_checkpoint);
 
 string invoke_info_checkpoints(GdbServer& gdb_server, Task*,
@@ -126,7 +133,7 @@ string invoke_info_checkpoints(GdbServer& gdb_server, Task*,
   }
   return out;
 }
-static SimpleGdbCommand info_checkpoints("info checkpoints",
+static SimpleGdbCommand info_checkpoints("info checkpoints", "",
                                          invoke_info_checkpoints);
 
 /*static*/ void GdbCommand::init_auto_args() {
