@@ -134,6 +134,8 @@ enum GdbRegister {
   DREG_64_MXCSR,
   DREG_64_LAST_FXSAVE_REG = DREG_64_MXCSR,
   DREG_ORIG_RAX,
+  DREG_FS_BASE,
+  DREG_GS_BASE,
   DREG_64_YMM0H,
   DREG_64_YMM1H,
   DREG_64_YMM2H,
