@@ -2,9 +2,10 @@
 
 #include "GdbServer.h"
 
-#include <assert.h>
+#include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <sys/sysmacros.h>
 #include <sys/wait.h>
 #include <unistd.h>
 
@@ -19,8 +20,10 @@
 #include "GdbExpression.h"
 #include "ReplaySession.h"
 #include "ScopedFd.h"
+#include "StringVectorToCharArray.h"
 #include "Task.h"
-#include "TaskGroup.h"
+#include "ThreadGroup.h"
+#include "core.h"
 #include "kernel_metadata.h"
 #include "log.h"
 #include "util.h"
@@ -31,13 +34,14 @@ namespace rr {
 
 GdbServer::GdbServer(std::unique_ptr<GdbConnection>& dbg, Task* t)
     : dbg(std::move(dbg)),
-      debuggee_tguid(t->task_group()->tguid()),
+      debuggee_tguid(t->thread_group()->tguid()),
       last_continue_tuid(t->tuid()),
       last_query_tuid(t->tuid()),
       final_event(UINT32_MAX),
       stop_replaying_to_target(false),
       interrupt_pending(false),
-      emergency_debug_session(&t->session()) {
+      emergency_debug_session(&t->session()),
+      file_scope_pid(0) {
   memset(&stop_siginfo, 0, sizeof(stop_siginfo));
 }
 
@@ -106,7 +110,7 @@ static const string& gdb_rr_macros() {
        << GdbCommandHandler::gdb_macros()
        // Try both "set target-async" and "maint set target-async" since
        // that changed recently.
-       << "python-interactive\n"
+       << "python\n"
        << "import re\n"
        << "m = re.compile("
        << "'.* ([0-9]+)\\.([0-9]+)(\\.([0-9]+))?.*'"
@@ -122,7 +126,8 @@ static const string& gdb_rr_macros() {
        << "\n"
        << "if ver < 71101:\n"
        << "    gdb.execute('set target-async 0')\n"
-       << "    gdb.execute('maint set target-async 0')\n";
+       << "    gdb.execute('maint set target-async 0')\n"
+       << "end\n";
     s = ss.str();
   }
   return s;
@@ -237,7 +242,7 @@ void GdbServer::dispatch_regs_request(const Registers& regs,
       return;
   }
   vector<GdbRegisterValue> rs;
-  for (GdbRegister r = GdbRegister(0); r < end; r = GdbRegister(r + 1)) {
+  for (GdbRegister r = GdbRegister(0); r <= end; r = GdbRegister(r + 1)) {
     rs.push_back(get_reg(regs, extra_regs, r));
   }
   dbg->reply_get_regs(rs);
@@ -250,7 +255,7 @@ public:
       expressions.push_back(GdbExpression(b.data(), b.size()));
     }
   }
-  virtual bool evaluate(Task* t) const {
+  virtual bool evaluate(Task* t) const override {
     for (auto& e : expressions) {
       GdbExpression::Value v;
       // Break if evaluation fails or the result is nonzero
@@ -341,12 +346,12 @@ void GdbServer::maybe_intercept_mem_request(Task* target, const GdbRequest& req,
 void GdbServer::dispatch_debugger_request(Session& session,
                                           const GdbRequest& req,
                                           ReportState state) {
-  assert(!req.is_resume_request());
+  DEBUG_ASSERT(!req.is_resume_request());
 
   // These requests don't require a target task.
   switch (req.type) {
     case DREQ_RESTART:
-      assert(false);
+      DEBUG_ASSERT(false);
       return; // unreached
     case DREQ_GET_CURRENT_THREAD:
       dbg->reply_get_current_thread(get_threadid(session, last_continue_tuid));
@@ -369,7 +374,7 @@ void GdbServer::dispatch_debugger_request(Session& session,
       Task* t = session.find_task(last_continue_tuid);
       ASSERT(t, session.is_diversion())
           << "Replay interrupts should be handled at a higher level";
-      assert(!t || t->task_group()->tguid() == debuggee_tguid);
+      DEBUG_ASSERT(!t || t->thread_group()->tguid() == debuggee_tguid);
       dbg->notify_stop(t ? get_threadid(t) : GdbThreadId(), 0);
       memset(&stop_siginfo, 0, sizeof(stop_siginfo));
       if (t) {
@@ -377,6 +382,65 @@ void GdbServer::dispatch_debugger_request(Session& session,
       }
       return;
     }
+    case DREQ_GET_EXEC_FILE: {
+      // We shouldn't normally receive this since we try to pass the exe file
+      // name on gdb's command line, but the user might start gdb manually
+      // and this is easy to support in case some other debugger or
+      // configuration needs it.
+      Task* t = nullptr;
+      if (req.target.tid) {
+        ThreadGroup* tg = session.find_thread_group(req.target.tid);
+        if (tg) {
+          t = *tg->task_set().begin();
+        }
+      } else {
+        t = session.find_task(last_continue_tuid);
+      }
+      if (t) {
+        dbg->reply_get_exec_file(t->vm()->exe_image());
+      } else {
+        dbg->reply_get_exec_file(string());
+      }
+      return;
+    }
+    case DREQ_FILE_SETFS:
+      // Only the filesystem as seen by the remote stub is supported currently
+      file_scope_pid = req.file_setfs().pid;
+      dbg->reply_setfs(0);
+      return;
+    case DREQ_FILE_OPEN:
+      // We only support reading files
+      if (req.file_open().flags == O_RDONLY) {
+        int fd = open_file(session, req.file_open().file_name);
+        dbg->reply_open(fd, fd >= 0 ? 0 : ENOENT);
+      } else {
+        dbg->reply_open(-1, EACCES);
+      }
+      return;
+    case DREQ_FILE_PREAD: {
+      auto it = files.find(req.file_pread().fd);
+      if (it != files.end()) {
+        size_t size = min<uint64_t>(req.file_pread().size, 1024 * 1024);
+        vector<uint8_t> data;
+        data.resize(size);
+        ssize_t bytes =
+            read_to_end(it->second, req.file_pread().offset, data.data(), size);
+        dbg->reply_pread(data.data(), bytes, bytes >= 0 ? 0 : errno);
+      } else {
+        dbg->reply_pread(nullptr, 0, EBADF);
+      }
+      return;
+    }
+    case DREQ_FILE_CLOSE: {
+      auto it = files.find(req.file_close().fd);
+      if (it != files.end()) {
+        files.erase(it);
+        dbg->reply_close(0);
+      } else {
+        dbg->reply_close(EBADF);
+      }
+      return;
+    }
     default:
       /* fall through to next switch stmt */
       break;
@@ -602,14 +666,20 @@ void GdbServer::dispatch_debugger_request(Session& session,
       // copy.  When gdb sends a plain "qSymbol::" packet, because gdb
       // has detected some change in the inferior state that might
       // enable more symbol lookups, we restart the iterator.
+      if (!thread_db) {
+        thread_db =
+            std::unique_ptr<ThreadDb>(new ThreadDb(debuggee_tguid.tid()));
+      }
+
       const string& name = req.sym().name;
       if (req.sym().has_address) {
         // Got a response holding a previously-requested symbol's name
         // and address.
-        target->register_symbol(name, req.sym().address);
+        thread_db->register_symbol(name, req.sym().address);
       } else if (name == "") {
         // Plain "qSymbol::" request.
-        symbols = target->get_symbols_and_clear_map();
+        symbols =
+            thread_db->get_symbols_and_clear_map(target->thread_group().get());
         symbols_iter = symbols.begin();
       }
 
@@ -622,9 +692,14 @@ void GdbServer::dispatch_debugger_request(Session& session,
       return;
     }
     case DREQ_TLS: {
+      if (!thread_db) {
+        thread_db =
+            std::unique_ptr<ThreadDb>(new ThreadDb(debuggee_tguid.tid()));
+      }
       remote_ptr<void> address;
-      bool ok = target->get_tls_address(req.tls().offset, req.tls().load_module,
-                                        &address);
+      bool ok = thread_db->get_tls_address(target->thread_group().get(),
+                                           target->rec_tid, req.tls().offset,
+                                           req.tls().load_module, &address);
       dbg->reply_tls_addr(ok, address);
       return;
     }
@@ -672,7 +747,7 @@ bool GdbServer::diverter_process_debugger_requests(
 
       case DREQ_WRITE_SIGINFO:
         LOG(debug) << "Removing reference to diversion session ...";
-        assert(diversion_refcount > 0);
+        DEBUG_ASSERT(diversion_refcount > 0);
         --diversion_refcount;
         if (diversion_refcount == 0) {
           LOG(debug) << "  ... dying at next continue request";
@@ -681,7 +756,7 @@ bool GdbServer::diverter_process_debugger_requests(
         continue;
 
       case DREQ_RR_CMD: {
-        assert(req->type == DREQ_RR_CMD);
+        DEBUG_ASSERT(req->type == DREQ_RR_CMD);
         Task* task = diversion_session.find_task(last_continue_tuid);
         if (task) {
           std::string reply =
@@ -710,14 +785,16 @@ bool GdbServer::diverter_process_debugger_requests(
 
 static bool is_last_thread_exit(const BreakStatus& break_status) {
   return break_status.task_exit &&
-         break_status.task->task_group()->task_set().size() == 1;
+         break_status.task->thread_group()->task_set().size() == 1;
 }
 
 static Task* is_in_exec(ReplayTimeline& timeline) {
   Task* t = timeline.current_session().current_task();
-  return t &&
-                 timeline.current_session().next_step_is_syscall_exit(
-                     syscall_number_for_execve(t->arch()))
+  if (!t) {
+    return nullptr;
+  }
+  return timeline.current_session().next_step_is_successful_syscall_exit(
+             syscall_number_for_execve(t->arch()))
              ? t
              : nullptr;
 }
@@ -731,27 +808,36 @@ void GdbServer::maybe_notify_stop(const GdbRequest& req,
     memset(&stop_siginfo, 0, sizeof(stop_siginfo));
     stop_siginfo.si_signo = SIGTRAP;
     watch_addr = break_status.watchpoints_hit[0].addr;
+    LOG(debug) << "Stopping for watchpoint at " << watch_addr;
   }
   if (break_status.breakpoint_hit || break_status.singlestep_complete) {
     do_stop = true;
     memset(&stop_siginfo, 0, sizeof(stop_siginfo));
     stop_siginfo.si_signo = SIGTRAP;
+    if (break_status.breakpoint_hit) {
+      LOG(debug) << "Stopping for breakpoint";
+    } else {
+      LOG(debug) << "Stopping for singlestep";
+    }
   }
   if (break_status.signal) {
     do_stop = true;
     stop_siginfo = *break_status.signal;
+    LOG(debug) << "Stopping for signal " << stop_siginfo;
   }
   if (is_last_thread_exit(break_status) && dbg->features().reverse_execution) {
     do_stop = true;
     memset(&stop_siginfo, 0, sizeof(stop_siginfo));
     if (req.cont().run_direction == RUN_FORWARD) {
-      // The exit of the last task in a task group generates a fake SIGKILL,
+      // The exit of the last task in a thread group generates a fake SIGKILL,
       // when reverse-execution is enabled, because users often want to run
       // backwards from the end of the task.
       stop_siginfo.si_signo = SIGKILL;
+      LOG(debug) << "Stopping for synthetic SIGKILL";
     } else {
       // The start of the debuggee task-group should trigger a silent stop.
       stop_siginfo.si_signo = 0;
+      LOG(debug) << "Stopping at start of execution while running backwards";
     }
   }
   Task* t = break_status.task;
@@ -760,8 +846,9 @@ void GdbServer::maybe_notify_stop(const GdbRequest& req,
     do_stop = true;
     memset(&stop_siginfo, 0, sizeof(stop_siginfo));
     t = in_exec_task;
+    LOG(debug) << "Stopping at exec";
   }
-  if (do_stop && t->task_group()->tguid() == debuggee_tguid) {
+  if (do_stop && t->thread_group()->tguid() == debuggee_tguid) {
     /* Notify the debugger and process any new requests
      * that might have triggered before resuming. */
     dbg->notify_stop(get_threadid(t), stop_siginfo.si_signo,
@@ -795,8 +882,8 @@ struct AllowedTasks {
   RunCommand command;
 };
 static RunCommand compute_run_command_for_reverse_exec(
-    Session& session, const TaskGroupUid& debuggee_tguid, const GdbRequest& req,
-    vector<AllowedTasks>& allowed_tasks) {
+    Session& session, const ThreadGroupUid& debuggee_tguid,
+    const GdbRequest& req, vector<AllowedTasks>& allowed_tasks) {
   // Singlestep if any of the actions request singlestepping.
   RunCommand result = RUN_CONTINUE;
   for (auto& action : req.cont().actions) {
@@ -846,7 +933,7 @@ GdbRequest GdbServer::divert(ReplaySession& replay) {
 
   while (diverter_process_debugger_requests(*diversion_session,
                                             diversion_refcount, &req)) {
-    assert(req.is_resume_request());
+    DEBUG_ASSERT(req.is_resume_request());
 
     if (req.cont().run_direction == RUN_BACKWARD) {
       // We don't support reverse execution in a diversion. Just issue
@@ -876,13 +963,13 @@ GdbRequest GdbServer::divert(ReplaySession& replay) {
       break;
     }
 
-    assert(result.status == DiversionSession::DIVERSION_CONTINUE);
+    DEBUG_ASSERT(result.status == DiversionSession::DIVERSION_CONTINUE);
 
     maybe_notify_stop(req, result.break_status);
   }
 
   LOG(debug) << "... ending debugging diversion";
-  assert(diversion_refcount == 0);
+  DEBUG_ASSERT(diversion_refcount == 0);
 
   diversion_session->kill_all_tasks();
 
@@ -1043,7 +1130,9 @@ GdbServer::ContinueOrStop GdbServer::debug_one_step(
         return handle_exited_state(last_resume_request);
       }
     } else {
-      in_debuggee_end_state = false;
+      if (req.type != DREQ_DETACH) {
+        in_debuggee_end_state = false;
+      }
     }
     // Otherwise (e.g. detach, restart, interrupt or reverse-exec) process
     // the request as normal.
@@ -1062,15 +1151,15 @@ GdbServer::ContinueOrStop GdbServer::debug_one_step(
   if (req.is_resume_request()) {
     last_resume_request = req;
   } else {
-    assert(req.type == DREQ_INTERRUPT);
+    DEBUG_ASSERT(req.type == DREQ_INTERRUPT);
     interrupt_pending = true;
     req = last_resume_request;
-    assert(req.is_resume_request());
+    DEBUG_ASSERT(req.is_resume_request());
   }
 
   if (interrupt_pending) {
     Task* t = timeline.current_session().current_task();
-    if (t->task_group()->tguid() == debuggee_tguid) {
+    if (t->thread_group()->tguid() == debuggee_tguid) {
       interrupt_pending = false;
       dbg->notify_stop(get_threadid(t), in_debuggee_end_state ? SIGKILL : 0);
       memset(&stop_siginfo, 0, sizeof(stop_siginfo));
@@ -1080,7 +1169,7 @@ GdbServer::ContinueOrStop GdbServer::debug_one_step(
 
   if (req.cont().run_direction == RUN_FORWARD) {
     if (is_in_exec(timeline) &&
-        timeline.current_session().current_task()->task_group()->tguid() ==
+        timeline.current_session().current_task()->thread_group()->tguid() ==
             debuggee_tguid) {
       // Don't go any further forward. maybe_notify_stop will generate a
       // stop.
@@ -1102,7 +1191,7 @@ GdbServer::ContinueOrStop GdbServer::debug_one_step(
     RunCommand command = compute_run_command_for_reverse_exec(
         timeline.current_session(), debuggee_tguid, req, allowed_tasks);
     auto stop_filter = [&](Task* t) -> bool {
-      if (t->task_group()->tguid() != debuggee_tguid) {
+      if (t->thread_group()->tguid() != debuggee_tguid) {
         return false;
       }
       // If gdb's requested actions don't allow the task to run, we still
@@ -1123,13 +1212,13 @@ GdbServer::ContinueOrStop GdbServer::debug_one_step(
         break;
       case RUN_SINGLESTEP: {
         Task* t = timeline.current_session().find_task(last_continue_tuid);
-        assert(t);
+        DEBUG_ASSERT(t);
         result = timeline.reverse_singlestep(
             last_continue_tuid, t->tick_count(), stop_filter, interrupt_check);
         break;
       }
       default:
-        assert(0 && "Unknown RunCommand");
+        DEBUG_ASSERT(0 && "Unknown RunCommand");
     }
 
     if (result.status == REPLAY_EXITED) {
@@ -1141,7 +1230,7 @@ GdbServer::ContinueOrStop GdbServer::debug_one_step(
   }
   if (req.cont().run_direction == RUN_FORWARD &&
       is_last_thread_exit(result.break_status) &&
-      result.break_status.task->task_group()->tguid() == debuggee_tguid) {
+      result.break_status.task->thread_group()->tguid() == debuggee_tguid) {
     in_debuggee_end_state = true;
   }
   return CONTINUE_DEBUGGING;
@@ -1192,13 +1281,13 @@ bool GdbServer::at_target() {
  */
 void GdbServer::activate_debugger() {
   TraceFrame next_frame = timeline.current_session().current_trace_frame();
-  TraceFrame::Time event_now = next_frame.time();
+  FrameTime event_now = next_frame.time();
   if (!stop_replaying_to_target && (target.event > 0 || target.pid)) {
     fprintf(stderr, "\a\n"
                     "--------------------------------------------------\n"
-                    " ---> Reached target process %d at event %u.\n"
+                    " ---> Reached target process %d at event %llu.\n"
                     "--------------------------------------------------\n",
-            target.pid, event_now);
+            target.pid, (long long)event_now);
   }
 
   Task* t = timeline.current_session().current_task();
@@ -1228,8 +1317,8 @@ void GdbServer::activate_debugger() {
 }
 
 void GdbServer::restart_session(const GdbRequest& req) {
-  assert(req.type == DREQ_RESTART);
-  assert(dbg);
+  DEBUG_ASSERT(req.type == DREQ_RESTART);
+  DEBUG_ASSERT(dbg);
 
   in_debuggee_end_state = false;
   timeline.remove_breakpoints_and_watchpoints();
@@ -1270,7 +1359,7 @@ void GdbServer::restart_session(const GdbRequest& req) {
 
   stop_replaying_to_target = false;
 
-  assert(req.restart().type == RESTART_FROM_EVENT);
+  DEBUG_ASSERT(req.restart().type == RESTART_FROM_EVENT);
   // Note that we don't reset the target pid; we intentionally keep targeting
   // the same process no matter what is running when we hit the event.
   target.event = req.restart().param;
@@ -1281,9 +1370,9 @@ void GdbServer::restart_session(const GdbRequest& req) {
         timeline.replay_step_forward(RUN_CONTINUE, target.event);
     // We should never reach the end of the trace without hitting the stop
     // condition below.
-    assert(result.status != REPLAY_EXITED);
+    DEBUG_ASSERT(result.status != REPLAY_EXITED);
     if (is_last_thread_exit(result.break_status) &&
-        result.break_status.task->task_group()->tgid == target.pid) {
+        result.break_status.task->thread_group()->tgid == target.pid) {
       // Debuggee task is about to exit. Stop here.
       in_debuggee_end_state = true;
       break;
@@ -1318,17 +1407,90 @@ static uint32_t get_cpu_features(SupportedArch arch) {
   return cpu_features;
 }
 
+struct DebuggerParams {
+  char exe_image[PATH_MAX];
+  char host[16]; // INET_ADDRSTRLEN, omitted for header churn
+  short port;
+};
+
+static void push_default_gdb_options(vector<string>& vec) {
+  // The gdb protocol uses the "vRun" packet to reload
+  // remote targets.  The packet is specified to be like
+  // "vCont", in which gdb waits infinitely long for a
+  // stop reply packet.  But in practice, gdb client
+  // expects the vRun to complete within the remote-reply
+  // timeout, after which it issues vCont.  The timeout
+  // causes gdb<-->rr communication to go haywire.
+  //
+  // rr can take a very long time indeed to send the
+  // stop-reply to gdb after restarting replay; the time
+  // to reach a specified execution target is
+  // theoretically unbounded.  Timing out on vRun is
+  // technically a gdb bug, but because the rr replay and
+  // the gdb reload models don't quite match up, we'll
+  // work around it on the rr side by disabling the
+  // remote-reply timeout.
+  vec.push_back("-l");
+  vec.push_back("10000");
+  // For now, avoid requesting binary files through vFile. That is slow and
+  // hard to make work correctly, because gdb requests files based on the
+  // names it sees in memory and in ELF, and those names may be symlinks to
+  // the filenames in the trace, so it's hard to match those names to files in
+  // the trace.
+  vec.push_back("-ex");
+  vec.push_back("set sysroot /");
+}
+
+static void push_target_remote_cmd(vector<string>& vec, const string& host,
+                                   unsigned short port) {
+  vec.push_back("-ex");
+  stringstream ss;
+  // If we omit the address, then gdb can try to resolve "localhost" which
+  // in some broken environments may not actually resolve to the local host
+  ss << "target extended-remote " << host << ":" << port;
+  vec.push_back(ss.str());
+}
+
+/**
+ * Wait for exactly one gdb host to connect to this remote target on
+ * the specified IP address |host|, port |port|.  If |probe| is nonzero,
+ * a unique port based on |start_port| will be searched for.  Otherwise,
+ * if |port| is already bound, this function will fail.
+ *
+ * Pass the |tgid| of the task on which this debug-connection request
+ * is being made.  The remaining debugging session will be limited to
+ * traffic regarding |tgid|, but clients don't need to and shouldn't
+ * need to assume that.
+ *
+ * If we're opening this connection on behalf of a known client, pass
+ * an fd in |client_params_fd|; we'll write the allocated port and |exe_image|
+ * through the fd before waiting for a connection. |exe_image| is the
+ * process that will be debugged by client, or null ptr if there isn't
+ * a client.
+ *
+ * This function is infallible: either it will return a valid
+ * debugging context, or it won't return.
+ */
 static unique_ptr<GdbConnection> await_connection(
-    Task* t, unsigned short port, GdbConnection::ProbePort probe,
-    const GdbConnection::Features& features, const string& debugger_name,
-    ScopedFd* client_params_fd = nullptr) {
-  auto result = GdbConnection::await_client_connection(
-      port, probe, t->tgid(), debugger_name, t->vm()->exe_image(), features,
-      client_params_fd);
+    Task* t, ScopedFd& listen_fd, const GdbConnection::Features& features) {
+  auto dbg = unique_ptr<GdbConnection>(new GdbConnection(t->tgid(), features));
+  dbg->set_cpu_features(get_cpu_features(t->arch()));
+  dbg->await_debugger(listen_fd);
+  return dbg;
+}
 
-  result->set_cpu_features(get_cpu_features(t->arch()));
-
-  return result;
+static void print_debugger_launch_command(Task* t, const string& host,
+                                          unsigned short port,
+                                          const char* debugger_name,
+                                          FILE* out) {
+  vector<string> options;
+  push_default_gdb_options(options);
+  push_target_remote_cmd(options, host, port);
+  fprintf(out, "%s ", debugger_name);
+  for (auto& opt : options) {
+    fprintf(out, "'%s' ", opt.c_str());
+  }
+  fprintf(out, "%s\n", t->vm()->exe_image().c_str());
 }
 
 void GdbServer::serve_replay(const ConnectionFlags& flags) {
@@ -1347,36 +1509,57 @@ void GdbServer::serve_replay(const ConnectionFlags& flags) {
   // presumably break if a different port were to be selected by
   // rr (otherwise why would they specify a port in the first
   // place).  So fail with a clearer error message.
-  auto probe = flags.dbg_port > 0 ? GdbConnection::DONT_PROBE
-                                  : GdbConnection::PROBE_PORT;
+  auto probe = flags.dbg_port > 0 ? DONT_PROBE : PROBE_PORT;
   Task* t = timeline.current_session().current_task();
-  dbg = await_connection(t, port, probe, GdbConnection::Features(),
-                         flags.debugger_name, flags.debugger_params_write_pipe);
+  ScopedFd listen_fd = open_socket(flags.dbg_host.c_str(), &port, probe);
+  if (flags.debugger_params_write_pipe) {
+    DebuggerParams params;
+    memset(&params, 0, sizeof(params));
+    strncpy(params.exe_image, t->vm()->exe_image().c_str(),
+            sizeof(params.exe_image) - 1);
+    strncpy(params.host, flags.dbg_host.c_str(), sizeof(params.host) - 1);
+    params.port = port;
+
+    ssize_t nwritten =
+        write(*flags.debugger_params_write_pipe, &params, sizeof(params));
+    DEBUG_ASSERT(nwritten == sizeof(params));
+  } else {
+    fputs("Launch gdb with\n  ", stderr);
+    print_debugger_launch_command(t, flags.dbg_host, port,
+                                  flags.debugger_name.c_str(), stderr);
+  }
+
   if (flags.debugger_params_write_pipe) {
     flags.debugger_params_write_pipe->close();
   }
-  debuggee_tguid = t->task_group()->tguid();
+  debuggee_tguid = t->thread_group()->tguid();
 
-  TraceFrame::Time first_run_event = t->vm()->first_run_event();
+  FrameTime first_run_event = t->vm()->first_run_event();
   if (first_run_event) {
     timeline.set_reverse_execution_barrier_event(first_run_event);
   }
 
-  activate_debugger();
+  do {
+    LOG(debug) << "initializing debugger connection";
+    dbg = await_connection(t, listen_fd, GdbConnection::Features());
+    activate_debugger();
 
-  GdbRequest last_resume_request;
-  while (debug_one_step(last_resume_request) == CONTINUE_DEBUGGING) {
-  }
+    GdbRequest last_resume_request;
+    while (debug_one_step(last_resume_request) == CONTINUE_DEBUGGING) {
+    }
+
+    timeline.remove_breakpoints_and_watchpoints();
+  } while (flags.keep_listening);
 
   LOG(debug) << "debugger server exiting ...";
 }
 
 static string create_gdb_command_file(const string& macros) {
-  char tmp[] = "/tmp/rr-gdb-commands-XXXXXX";
+  TempFile file = create_temporary_file("rr-gdb-commands-XXXXXX");
   // This fd is just leaked. That's fine since we only call this once
   // per rr invocation at the moment.
-  int fd = mkstemp(tmp);
-  unlink(tmp);
+  int fd = file.fd.extract();
+  unlink(file.name.c_str());
 
   ssize_t len = macros.size();
   int written = write(fd, macros.c_str(), len);
@@ -1389,15 +1572,71 @@ static string create_gdb_command_file(const string& macros) {
   return procfile.str();
 }
 
+static string to_string(const vector<string>& args) {
+  stringstream ss;
+  for (auto& a : args) {
+    ss << "'" << a << "' ";
+  }
+  return ss.str();
+}
+
+static bool needs_target(const string& option) {
+  return !strncmp(option.c_str(), "continue", option.size());
+}
+
+/**
+ * Exec gdb using the params that were written to
+ * |params_pipe_fd|.  Optionally, pre-define in the gdb client the set
+ * of macros defined in |macros| if nonnull.
+ */
 void GdbServer::launch_gdb(ScopedFd& params_pipe_fd,
                            const string& gdb_binary_file_path,
                            const vector<string>& gdb_options) {
   auto macros = gdb_rr_macros();
   string gdb_command_file = create_gdb_command_file(macros);
-  vector<string> options = gdb_options;
-  options.push_back("-x");
-  options.push_back(gdb_command_file);
-  GdbConnection::launch_gdb(params_pipe_fd, gdb_binary_file_path, options);
+
+  DebuggerParams params;
+  ssize_t nread;
+  while (true) {
+    nread = read(params_pipe_fd, &params, sizeof(params));
+    if (nread == 0) {
+      // pipe was closed. Probably rr failed/died.
+      return;
+    }
+    if (nread != -1 || errno != EINTR) {
+      break;
+    }
+  }
+  DEBUG_ASSERT(nread == sizeof(params));
+
+  vector<string> args;
+  args.push_back(gdb_binary_file_path);
+  push_default_gdb_options(args);
+  args.push_back("-x");
+  args.push_back(gdb_command_file);
+  bool did_set_remote = false;
+  for (size_t i = 0; i < gdb_options.size(); ++i) {
+    if (!did_set_remote && gdb_options[i] == "-ex" &&
+        i + 1 < gdb_options.size() && needs_target(gdb_options[i + 1])) {
+      push_target_remote_cmd(args, string(params.host), params.port);
+      did_set_remote = true;
+    }
+    args.push_back(gdb_options[i]);
+  }
+  if (!did_set_remote) {
+    push_target_remote_cmd(args, string(params.host), params.port);
+  }
+  args.push_back(params.exe_image);
+
+  vector<string> env = current_env();
+  env.push_back("GDB_UNDER_RR=1");
+
+  LOG(debug) << "launching " << to_string(args);
+
+  StringVectorToCharArray c_args(args);
+  StringVectorToCharArray c_env(env);
+  execvpe(gdb_binary_file_path.c_str(), c_args.get(), c_env.get());
+  FATAL() << "Failed to exec gdb.";
 }
 
 void GdbServer::emergency_debug(Task* t) {
@@ -1418,12 +1657,118 @@ void GdbServer::emergency_debug(Task* t) {
   // b) some gdb versions will fail if the user doesn't turn off async
   // mode (and we don't want to require users to do that)
   features.reverse_execution = false;
-  unique_ptr<GdbConnection> dbg =
-      await_connection(t, t->tid, GdbConnection::PROBE_PORT, features, "gdb");
+  unsigned short port = t->tid;
+  ScopedFd listen_fd = open_socket(localhost_addr.c_str(), &port, PROBE_PORT);
+
+  char* test_monitor_pid = getenv("RUNNING_UNDER_TEST_MONITOR");
+  if (test_monitor_pid) {
+    pid_t pid = atoi(test_monitor_pid);
+    // Tell test-monitor to wake up and take a snapshot. It will also
+    // connect the emergency debugger so let that happen.
+    FILE* gdb_cmd = fopen("gdb_cmd", "w");
+    if (gdb_cmd) {
+      print_debugger_launch_command(t, localhost_addr, port, "gdb", gdb_cmd);
+      fclose(gdb_cmd);
+    }
+    kill(pid, SIGURG);
+  } else {
+    dump_rr_stack();
+    fputs("Launch gdb with\n  ", stderr);
+    print_debugger_launch_command(t, localhost_addr, port, "gdb", stderr);
+  }
+  unique_ptr<GdbConnection> dbg = await_connection(t, listen_fd, features);
 
   GdbServer(dbg, t).process_debugger_requests();
 }
 
 string GdbServer::init_script() { return gdb_rr_macros(); }
 
+static ScopedFd generate_fake_proc_maps(Task* t) {
+  TempFile file = create_temporary_file("rr-fake-proc-maps-XXXXXX");
+  unlink(file.name.c_str());
+
+  int fd = dup(file.fd);
+  if (fd < 0) {
+    FATAL() << "Cannot dup";
+  }
+  FILE* f = fdopen(fd, "w");
+
+  int addr_min_width = word_size(t->arch()) == 8 ? 10 : 8;
+  for (auto& m : t->vm()->maps()) {
+    int len =
+        fprintf(f, "%0*llx-%0*llx %s%s%s%s %08llx %02x:%02x %lld",
+                addr_min_width, (long long)m.recorded_map.start().as_int(),
+                addr_min_width, (long long)m.recorded_map.end().as_int(),
+                (m.recorded_map.prot() & PROT_READ) ? "r" : "-",
+                (m.recorded_map.prot() & PROT_WRITE) ? "w" : "-",
+                (m.recorded_map.prot() & PROT_EXEC) ? "x" : "-",
+                (m.recorded_map.flags() & MAP_SHARED) ? "s" : "p",
+                (long long)m.recorded_map.file_offset_bytes(),
+                major(m.recorded_map.device()), minor(m.recorded_map.device()),
+                (long long)m.recorded_map.inode());
+    while (len < 72) {
+      fputc(' ', f);
+      ++len;
+    }
+    fputc(' ', f);
+
+    string name;
+    const string& fsname = m.recorded_map.fsname();
+    for (size_t i = 0; i < fsname.size(); ++i) {
+      if (fsname[i] == '\n') {
+        name.append("\\012");
+      } else {
+        name.push_back(fsname[i]);
+      }
+    }
+    fputs(name.c_str(), f);
+    fputc('\n', f);
+  }
+  if (ferror(f) || fclose(f)) {
+    FATAL() << "Can't write";
+  }
+
+  return move(file.fd);
+}
+
+int GdbServer::open_file(Session& session, const std::string& file_name) {
+  // XXX should we require file_scope_pid == 0 here?
+  ScopedFd contents;
+
+  LOG(debug) << "Trying to open " << file_name;
+
+  if (file_name.substr(0, 6) == "/proc/") {
+    char* tid_end;
+    long tid = strtol(file_name.c_str() + 6, &tid_end, 10);
+    if (*tid_end != '/') {
+      return -1;
+    }
+    if (!strncmp(tid_end, "/task/", 6)) {
+      tid = strtol(tid_end + 6, &tid_end, 10);
+      if (*tid_end != '/') {
+        return -1;
+      }
+    }
+    if (tid != (pid_t)tid) {
+      return -1;
+    }
+    Task* t = session.find_task(tid);
+    if (!t) {
+      return -1;
+    }
+    if (!strcmp(tid_end, "/maps")) {
+      contents = generate_fake_proc_maps(t);
+    } else {
+      return -1;
+    }
+  }
+
+  int ret_fd = 0;
+  while (files.find(ret_fd) != files.end()) {
+    ++ret_fd;
+  }
+  files.insert(make_pair(ret_fd, move(contents)));
+  return ret_fd;
+}
+
 } // namespace rr
