@@ -80,7 +80,7 @@ public:
   void patch_after_mmap(RecordTask* t, remote_ptr<void> start, size_t size,
                         size_t offset_pages, int child_fd);
 
-  remote_ptr<void> x86_sysenter_vsyscall;
+  remote_ptr<void> x86_vsyscall;
   /**
    * The list of pages we've allocated to hold our extended jumps.
    */
@@ -91,7 +91,13 @@ public:
   };
   std::vector<ExtendedJumpPage> extended_jump_pages;
 
-  bool is_jump_stub_instruction(remote_ptr<void> p);
+  bool is_jump_stub_instruction(remote_code_ptr p);
+
+  /**
+   * Syscalls in the VDSO that we patched to be direct syscalls. These can
+   * always be safely patched to jump to the syscallbuf.
+   */
+  std::unordered_set<remote_code_ptr> patched_vdso_syscalls;
 
 private:
   /**
