@@ -2,7 +2,6 @@
 
 #include "PerfCounters.h"
 
-#include <assert.h>
 #include <err.h>
 #include <fcntl.h>
 #include <linux/perf_event.h>
@@ -19,6 +18,9 @@
 #include <string>
 
 #include "Flags.h"
+#include "Session.h"
+#include "Task.h"
+#include "core.h"
 #include "kernel_metadata.h"
 #include "log.h"
 #include "util.h"
@@ -30,9 +32,14 @@ namespace rr {
 static bool attributes_initialized;
 static struct perf_event_attr ticks_attr;
 static struct perf_event_attr cycles_attr;
-static struct perf_event_attr page_faults_attr;
 static struct perf_event_attr hw_interrupts_attr;
-static struct perf_event_attr instructions_retired_attr;
+static uint32_t skid_size;
+static bool has_ioc_period_bug;
+static bool has_kvm_in_txcp_bug;
+static bool has_xen_pmi_bug;
+static bool supports_txcp;
+static bool only_one_counter;
+static bool activate_useless_counter;
 
 /*
  * Find out the cpu model using the cpuid instruction.
@@ -52,15 +59,16 @@ enum CpuMicroarch {
   IntelBroadwell,
   IntelSkylake,
   IntelSilvermont,
-  IntelKabylake
+  IntelKabylake,
+  AMDRyzen,
 };
 
 struct PmuConfig {
   CpuMicroarch uarch;
   const char* name;
   unsigned rcb_cntr_event;
-  unsigned rinsn_cntr_event;
   unsigned hw_intr_cntr_event;
+  uint32_t skid_size;
   bool supported;
   /*
    * Some CPUs turn off the whole PMU when there are no remaining events
@@ -76,23 +84,19 @@ struct PmuConfig {
 
 // XXX please only edit this if you really know what you're doing.
 static const PmuConfig pmu_configs[] = {
-  { IntelKabylake, "Intel Kabylake", 0x5101c4, 0x5100c0, 0x5301cb, true,
+  { IntelKabylake, "Intel Kabylake", 0x5101c4, 0x5301cb, 100, true, false },
+  { IntelSilvermont, "Intel Silvermont", 0x517ec4, 0x5301cb, 100, true, true },
+  { IntelSkylake, "Intel Skylake", 0x5101c4, 0x5301cb, 100, true, false },
+  { IntelBroadwell, "Intel Broadwell", 0x5101c4, 0x5301cb, 100, true, false },
+  { IntelHaswell, "Intel Haswell", 0x5101c4, 0x5301cb, 100, true, false },
+  { IntelIvyBridge, "Intel Ivy Bridge", 0x5101c4, 0x5301cb, 100, true, false },
+  { IntelSandyBridge, "Intel Sandy Bridge", 0x5101c4, 0x5301cb, 100, true,
     false },
-  { IntelSilvermont, "Intel Silvermont", 0x517ec4, 0x5100c0, 0x5301cb, true,
-    true },
-  { IntelSkylake, "Intel Skylake", 0x5101c4, 0x5100c0, 0x5301cb, true, false },
-  { IntelBroadwell, "Intel Broadwell", 0x5101c4, 0x5100c0, 0x5301cb, true,
-    false },
-  { IntelHaswell, "Intel Haswell", 0x5101c4, 0x5100c0, 0x5301cb, true, false },
-  { IntelIvyBridge, "Intel Ivy Bridge", 0x5101c4, 0x5100c0, 0x5301cb, true,
-    false },
-  { IntelSandyBridge, "Intel Sandy Bridge", 0x5101c4, 0x5100c0, 0x5301cb, true,
-    false },
-  { IntelNehalem, "Intel Nehalem", 0x5101c4, 0x5100c0, 0x50011d, true, false },
-  { IntelWestmere, "Intel Westmere", 0x5101c4, 0x5100c0, 0x50011d, true,
-    false },
-  { IntelPenryn, "Intel Penryn", 0, 0, 0, false, false },
-  { IntelMerom, "Intel Merom", 0, 0, 0, false, false },
+  { IntelNehalem, "Intel Nehalem", 0x5101c4, 0x50011d, 100, true, false },
+  { IntelWestmere, "Intel Westmere", 0x5101c4, 0x50011d, 100, true, false },
+  { IntelPenryn, "Intel Penryn", 0, 0, 100, false, false },
+  { IntelMerom, "Intel Merom", 0, 0, 100, false, false },
+  { AMDRyzen, "AMD Ryzen", 0x5100d1, 0, 1000, true, false },
 };
 
 static string lowercase(const string& s) {
@@ -116,11 +120,23 @@ static CpuMicroarch get_cpu_microarch() {
         return pmu.uarch;
       }
     }
-    FATAL() << "Forced uarch " << Flags::get().forced_uarch << " isn't known.";
+    CLEAN_FATAL() << "Forced uarch " << Flags::get().forced_uarch
+                  << " isn't known.";
+  }
+
+  auto cpuid_vendor = cpuid(CPUID_GETVENDORSTRING, 0);
+  char vendor[13];
+  memcpy(&vendor[0], &cpuid_vendor.ebx, 4);
+  memcpy(&vendor[4], &cpuid_vendor.edx, 4);
+  memcpy(&vendor[8], &cpuid_vendor.ecx, 4);
+  vendor[12] = 0;
+  if (strcmp(vendor, "GenuineIntel") && strcmp(vendor, "AuthenticAMD")) {
+    CLEAN_FATAL() << "Unknown CPU vendor '" << vendor << "'";
   }
 
   auto cpuid_data = cpuid(CPUID_GETFEATURES, 0);
-  unsigned int cpu_type = (cpuid_data.eax & 0xF0FF0);
+  unsigned int cpu_type = cpuid_data.eax & 0xF0FF0;
+  unsigned int ext_family = (cpuid_data.eax >> 20) & 0xff;
   switch (cpu_type) {
     case 0x006F0:
     case 0x10660:
@@ -152,16 +168,39 @@ static CpuMicroarch get_cpu_microarch() {
     case 0x50660:
       return IntelBroadwell;
     case 0x406e0:
+    case 0x50650:
     case 0x506e0:
       return IntelSkylake;
     case 0x50670:
       return IntelSilvermont;
     case 0x806e0:
+    case 0x906e0:
       return IntelKabylake;
+    case 0x00f10:
+      if (ext_family == 8) {
+        if (!Flags::get().suppress_environment_warnings) {
+          fprintf(stderr, "You have a Ryzen CPU. The Ryzen "
+                          "retired-conditional-branches hardware\n"
+                          "performance counter is not accurate enough; rr will "
+                          "be unreliable.\n"
+                          "See https://github.com/mozilla/rr/issues/2034.\n");
+        }
+        return AMDRyzen;
+      }
+      break;
     default:
-      FATAL() << "CPU " << HEX(cpu_type) << " unknown.";
-      return UnknownCpu; // not reached
+      break;
   }
+
+  if (!strcmp(vendor, "AuthenticAMD")) {
+    CLEAN_FATAL()
+        << "AMD CPUs not supported.\n"
+        << "For Ryzen, see https://github.com/mozilla/rr/issues/2034.\n"
+        << "For post-Ryzen CPUs, please file a Github issue.";
+  } else {
+    CLEAN_FATAL() << "Intel CPU type " << HEX(cpu_type) << " unknown";
+  }
+  return UnknownCpu; // not reached
 }
 
 static void init_perf_event_attr(struct perf_event_attr* attr,
@@ -176,72 +215,22 @@ static void init_perf_event_attr(struct perf_event_attr* attr,
   attr->exclude_guest = 1;
 }
 
-static bool activate_useless_counter;
-static bool has_ioc_period_bug();
-static void init_attributes() {
-  if (attributes_initialized) {
-    return;
-  }
-  attributes_initialized = true;
-
-  CpuMicroarch uarch = get_cpu_microarch();
-  const PmuConfig* pmu = nullptr;
-  for (size_t i = 0; i < array_length(pmu_configs); ++i) {
-    if (uarch == pmu_configs[i].uarch) {
-      pmu = &pmu_configs[i];
-      break;
-    }
-  }
-  assert(pmu);
-
-  if (!pmu->supported) {
-    FATAL() << "Microarchitecture `" << pmu->name << "' currently unsupported.";
-  }
-
-  /*
-   * For maintainability, and since it doesn't impact performance when not
-   * needed, we always activate this. If it ever turns out to be a problem,
-   * this can be set to pmu->benefits_from_useless_counter, instead.
-   *
-   * We also disable this counter when running under rr. Even though it's the
-   * same event for the same task as the outer rr, the linux kernel does not
-   * coalesce them and tries to schedule the new one on a general purpose PMC.
-   * On CPUs with only 2 general PMCs (e.g. KNL), we'd run out.
-   */
-  activate_useless_counter = has_ioc_period_bug() && !running_under_rr();
-
-  init_perf_event_attr(&ticks_attr, PERF_TYPE_RAW, pmu->rcb_cntr_event);
-  init_perf_event_attr(&cycles_attr, PERF_TYPE_HARDWARE,
-                       PERF_COUNT_HW_CPU_CYCLES);
-  init_perf_event_attr(&instructions_retired_attr, PERF_TYPE_RAW,
-                       pmu->rinsn_cntr_event);
-  init_perf_event_attr(&hw_interrupts_attr, PERF_TYPE_RAW,
-                       pmu->hw_intr_cntr_event);
-  // libpfm encodes the event with this bit set, so we'll do the
-  // same thing.  Unclear if necessary.
-  hw_interrupts_attr.exclude_hv = 1;
-  init_perf_event_attr(&page_faults_attr, PERF_TYPE_SOFTWARE,
-                       PERF_COUNT_SW_PAGE_FAULTS);
-}
-
 static const uint64_t IN_TX = 1ULL << 32;
 static const uint64_t IN_TXCP = 1ULL << 33;
 
-bool PerfCounters::is_ticks_attr(const perf_event_attr& attr) {
-  init_attributes();
-  perf_event_attr tmp_attr = attr;
-  tmp_attr.sample_period = 0;
-  tmp_attr.config &= ~IN_TXCP;
-  return memcmp(&ticks_attr, &tmp_attr, sizeof(attr)) == 0;
-}
-
-PerfCounters::PerfCounters(pid_t tid)
-    : counting(false), tid(tid), started(false) {
-  init_attributes();
+static int64_t read_counter(ScopedFd& fd) {
+  int64_t val;
+  ssize_t nread = read(fd, &val, sizeof(val));
+  DEBUG_ASSERT(nread == sizeof(val));
+  return val;
 }
 
 static ScopedFd start_counter(pid_t tid, int group_fd,
-                              struct perf_event_attr* attr) {
+                              struct perf_event_attr* attr,
+                              bool* disabled_txcp = nullptr) {
+  if (disabled_txcp) {
+    *disabled_txcp = false;
+  }
   int fd = syscall(__NR_perf_event_open, attr, tid, -1, group_fd, 0);
   if (0 > fd && errno == EINVAL && attr->type == PERF_TYPE_RAW &&
       (attr->config & IN_TXCP)) {
@@ -250,15 +239,17 @@ static ScopedFd start_counter(pid_t tid, int group_fd,
     tmp_attr.config &= ~IN_TXCP;
     fd = syscall(__NR_perf_event_open, &tmp_attr, tid, -1, group_fd, 0);
     if (fd >= 0) {
+      if (disabled_txcp) {
+        *disabled_txcp = true;
+      }
       LOG(warn) << "kernel does not support IN_TXCP";
-      if (cpuid(CPUID_GETEXTENDEDFEATURES, 0).ebx & HLE_FEATURE_FLAG) {
-        if (!Flags::get().suppress_environment_warnings) {
-          fprintf(
-              stderr,
-              "Your CPU supports Hardware Lock Elision but your kernel does not\n"
-              "support setting the IN_TXCP PMU flag. Record and replay of code\n"
-              "that uses HLE will fail unless you update your kernel.\n");
-        }
+      if ((cpuid(CPUID_GETEXTENDEDFEATURES, 0).ebx & HLE_FEATURE_FLAG) &&
+          !Flags::get().suppress_environment_warnings) {
+        fprintf(stderr,
+                "Your CPU supports Hardware Lock Elision but your kernel does\n"
+                "not support setting the IN_TXCP PMU flag. Record and replay\n"
+                "of code that uses HLE will fail unless you update your\n"
+                "kernel.\n");
       }
     }
   }
@@ -276,73 +267,7 @@ static ScopedFd start_counter(pid_t tid, int group_fd,
   return fd;
 }
 
-static int64_t read_counter(ScopedFd& fd) {
-  int64_t val;
-  ssize_t nread = read(fd, &val, sizeof(val));
-  assert(nread == sizeof(val));
-  return val;
-}
-
-static const int NUM_BRANCHES = 500;
-
-static bool has_kvm_in_txcp_bug() {
-  static bool did_test = false;
-  static bool bug_detected = true;
-  if (did_test) {
-    return bug_detected;
-  }
-
-  did_test = true;
-  if (running_under_rr()) {
-    // Under rr we emulate an idealized performance counter, so the result of
-    // this test doesn't matter. Just say we don't have the bug.
-    bug_detected = false;
-    return bug_detected;
-  }
-
-  init_attributes();
-
-  struct perf_event_attr attr = rr::ticks_attr;
-  attr.config |= IN_TXCP;
-  attr.sample_period = 0;
-  ScopedFd fd = syscall(__NR_perf_event_open, &attr, 0, -1, -1, 0);
-  ioctl(fd, PERF_EVENT_IOC_DISABLE, 0);
-  ioctl(fd, PERF_EVENT_IOC_ENABLE, 0);
-  // Do NUM_BRANCHES conditional branches that can't be optimized out.
-  // 'accumulator' is always odd and can't be zero
-  int accumulator = rand() * 2 + 1;
-  for (int i = 0; i < 500 && accumulator; ++i) {
-    accumulator = ((accumulator * 7) + 2) & 0xffffff;
-  }
-  srand(accumulator);
-  int64_t count = read_counter(fd);
-
-  bug_detected = count < NUM_BRANCHES;
-  LOG(debug) << "has_kvm_in_txcp_bug count=" << count
-             << " detected=" << bug_detected;
-  return bug_detected;
-}
-
-static bool has_ioc_period_bug() {
-  static bool did_test = false;
-  static bool bug_detected = true;
-  if (did_test) {
-    return bug_detected;
-  }
-
-  did_test = true;
-  if (running_under_rr()) {
-    // Under rr we emulate an idealized performance counter, so the result of
-    // this test doesn't matter. Further, since this is not exactly the same as
-    // our ticks_attr, it can take up an extra PMC when recording rr replay,
-    // which we don't have available on some architectures. Just say we don't
-    // have the bug.
-    bug_detected = false;
-    return bug_detected;
-  }
-
-  init_attributes();
-
+static void check_for_ioc_period_bug() {
   // Start a cycles counter
   struct perf_event_attr attr = rr::ticks_attr;
   attr.sample_period = 0xffffffff;
@@ -357,9 +282,303 @@ static bool has_ioc_period_bug() {
   struct pollfd poll_bug_fd = {.fd = bug_fd, .events = POLL_IN, .revents = 0 };
   poll(&poll_bug_fd, 1, 0);
 
-  bug_detected = poll_bug_fd.revents == 0;
-  LOG(debug) << "has_ioc_period_bug detected=" << bug_detected;
-  return bug_detected;
+  has_ioc_period_bug = poll_bug_fd.revents == 0;
+  LOG(debug) << "has_ioc_period_bug=" << has_ioc_period_bug;
+}
+
+static const int NUM_BRANCHES = 500;
+
+static void do_branches() {
+  // Do NUM_BRANCHES conditional branches that can't be optimized out.
+  // 'accumulator' is always odd and can't be zero
+  int accumulator = rand() * 2 + 1;
+  for (int i = 0; i < NUM_BRANCHES && accumulator; ++i) {
+    accumulator = ((accumulator * 7) + 2) & 0xffffff;
+  }
+  srand(accumulator);
+}
+
+static void check_for_kvm_in_txcp_bug() {
+  int64_t count = 0;
+  struct perf_event_attr attr = rr::ticks_attr;
+  attr.config |= IN_TXCP;
+  attr.sample_period = 0;
+  bool disabled_txcp;
+  ScopedFd fd = start_counter(0, -1, &attr, &disabled_txcp);
+  if (fd.is_open() && !disabled_txcp) {
+    ioctl(fd, PERF_EVENT_IOC_DISABLE, 0);
+    ioctl(fd, PERF_EVENT_IOC_ENABLE, 0);
+    do_branches();
+    count = read_counter(fd);
+  }
+
+  supports_txcp = count > 0;
+  has_kvm_in_txcp_bug = supports_txcp && count < NUM_BRANCHES;
+  LOG(debug) << "supports txcp=" << supports_txcp;
+  LOG(debug) << "has_kvm_in_txcp_bug=" << has_kvm_in_txcp_bug
+             << " count=" << count;
+}
+
+static void check_for_xen_pmi_bug() {
+  int32_t count = -1;
+  struct perf_event_attr attr = rr::ticks_attr;
+  attr.sample_period = NUM_BRANCHES - 1;
+  ScopedFd fd = start_counter(0, -1, &attr);
+  if (fd.is_open()) {
+    // Do NUM_BRANCHES conditional branches that can't be optimized out.
+    // 'accumulator' is always odd and can't be zero
+    uint32_t accumulator = rand() * 2 + 1;
+    int raw_fd = fd;
+    asm volatile(
+#if defined(__x86_64__)
+        "mov %[_SYS_ioctl], %%rax;"
+        "mov %[raw_fd], %%edi;"
+        "xor %%rdx, %%rdx;"
+        "mov %[_PERF_EVENT_IOC_ENABLE], %%rsi;"
+        "syscall;"
+        "cmp $-4095, %%rax;"
+        "jae 2f;"
+        "mov %[_SYS_ioctl], %%rax;"
+        "mov %[_PERF_EVENT_IOC_RESET], %%rsi;"
+        "syscall;"
+        // From this point on all conditional branches count!
+        "cmp $-4095, %%rax;"
+        "jae 2f;"
+        // Reset the counter period to the desired value.
+        "mov %[_SYS_ioctl], %%rax;"
+        "mov %[_PERF_EVENT_IOC_PERIOD], %%rsi;"
+        "mov %[period], %%rdx;"
+        "syscall;"
+        "cmp $-4095, %%rax;"
+        "jae 2f;"
+        "mov %[_iterations], %%rax;"
+        "1: dec %%rax;"
+        // Multiply by 7.
+        "mov %[accumulator], %%edx;"
+        "shl $3, %[accumulator];"
+        "sub %%edx, %[accumulator];"
+        // Add 2.
+        "add $2, %[accumulator];"
+        // Mask off bits.
+        "and $0xffffff, %[accumulator];"
+        // And loop.
+        "test %%rax, %%rax;"
+        "jnz 1b;"
+        "mov %[_PERF_EVENT_IOC_DISABLE], %%rsi;"
+        "mov %[_SYS_ioctl], %%rax;"
+        "xor %%rdx, %%rdx;"
+        // We didn't touch rdi.
+        "syscall;"
+        "cmp $-4095, %%rax;"
+        "jae 2f;"
+        "movl $0, %[count];"
+        "2: nop;"
+#elif defined(__i386__)
+        "mov %[_SYS_ioctl], %%eax;"
+        "mov %[raw_fd], %%ebx;"
+        "xor %%edx, %%edx;"
+        "mov %[_PERF_EVENT_IOC_ENABLE], %%ecx;"
+        "int $0x80;"
+        "cmp $-4095, %%eax;"
+        "jae 2f;"
+        "mov %[_SYS_ioctl], %%eax;"
+        "mov %[_PERF_EVENT_IOC_RESET], %%ecx;"
+        "int $0x80;"
+        // From this point on all conditional branches count!
+        "cmp $-4095, %%eax;"
+        "jae 2f;"
+        // Reset the counter period to the desired value.
+        "mov %[_SYS_ioctl], %%eax;"
+        "mov %[_PERF_EVENT_IOC_PERIOD], %%ecx;"
+        "mov %[period], %%edx;"
+        "int $0x80;"
+        "cmp $-4095, %%eax;"
+        "jae 2f;"
+        "mov %[_iterations], %%eax;"
+        "1: dec %%eax;"
+        // Multiply by 7.
+        "mov %[accumulator], %%edx;"
+        "shll $3, %[accumulator];"
+        "sub %%edx, %[accumulator];"
+        // Add 2.
+        "add $2, %[accumulator];"
+        // Mask off bits.
+        "andl $0xffffff, %[accumulator];"
+        // And loop.
+        "test %%eax, %%eax;"
+        "jnz 1b;"
+        "mov %[_PERF_EVENT_IOC_DISABLE], %%ecx;"
+        "mov %[_SYS_ioctl], %%eax;"
+        "xor %%edx, %%edx;"
+        // We didn't touch rdi.
+        "int $0x80;"
+        "cmp $-4095, %%eax;"
+        "jae 2f;"
+        "movl $0, %[count];"
+        "2: nop;"
+#else
+#error unknown CPU architecture
+#endif
+        : [accumulator] "+rm"(accumulator), [count] "=rm"(count)
+        : [_SYS_ioctl] "i"(SYS_ioctl),
+          [_PERF_EVENT_IOC_DISABLE] "i"(PERF_EVENT_IOC_DISABLE),
+          [_PERF_EVENT_IOC_ENABLE] "i"(PERF_EVENT_IOC_ENABLE),
+          [_PERF_EVENT_IOC_PERIOD] "i"(PERF_EVENT_IOC_PERIOD),
+          [_PERF_EVENT_IOC_RESET] "i"(PERF_EVENT_IOC_RESET),
+          // The check for the failure of some of our ioctls is in
+          // the measured region, so account for that when looping.
+          [_iterations] "i"(NUM_BRANCHES - 2),
+          [period] "rm"(&attr.sample_period), [raw_fd] "rm"(raw_fd)
+        :
+#if defined(__x86_64__)
+        "rax", "rdx", "rdi", "rsi"
+        // `syscall` clobbers rcx and r11.
+        ,
+        "rcx", "r11"
+#elif defined(__i386__)
+        "eax", "ebx", "ecx", "edx"
+#else
+#error unknown CPU architecture
+#endif
+        );
+    // If things worked above, `count` should have been set to 0.
+    if (count == 0) {
+      count = read_counter(fd);
+    }
+    srand(accumulator);
+  }
+
+  has_xen_pmi_bug = count > NUM_BRANCHES || count == -1;
+  if (has_xen_pmi_bug) {
+    LOG(debug) << "has_xen_pmi_bug=" << has_xen_pmi_bug << " count=" << count;
+    if (!Flags::get().force_things) {
+      FATAL()
+          << "Overcount triggered by PMU interrupts detected due to Xen PMU "
+             "virtualization bug.\n"
+             "Aborting. Retry with -F to override, but it will probably\n"
+             "fail.";
+    }
+  }
+}
+
+static void check_working_counters() {
+  struct perf_event_attr attr = rr::ticks_attr;
+  attr.sample_period = 0;
+  struct perf_event_attr attr2 = rr::cycles_attr;
+  attr.sample_period = 0;
+  ScopedFd fd = start_counter(0, -1, &attr);
+  ScopedFd fd2 = start_counter(0, -1, &attr2);
+  do_branches();
+  int64_t events = read_counter(fd);
+  int64_t events2 = read_counter(fd2);
+
+  if (events < NUM_BRANCHES) {
+    char config[100];
+    sprintf(config, "%llx", (long long)ticks_attr.config);
+    FATAL()
+        << "\nGot " << events << " branch events, expected at least "
+        << NUM_BRANCHES
+        << ".\n"
+           "\nThe hardware performance counter seems to not be working. Check\n"
+           "that hardware performance counters are working by running\n"
+           "  perf stat -e r"
+        << config
+        << " true\n"
+           "and checking that it reports a nonzero number of events.\n"
+           "If performance counters seem to be working with 'perf', file an\n"
+           "rr issue, otherwise check your hardware/OS/VM configuration. Also\n"
+           "check that other software is not using performance counters on\n"
+           "this CPU.";
+  }
+
+  only_one_counter = events2 == 0;
+  LOG(debug) << "only_one_counter=" << only_one_counter;
+
+  if (only_one_counter &&
+      (cpuid(CPUID_GETEXTENDEDFEATURES, 0).ebx & HLE_FEATURE_FLAG) &&
+      !Flags::get().suppress_environment_warnings) {
+    fprintf(stderr,
+            "Your CPU supports Hardware Lock Elision but you only have one\n"
+            "hardware performance counter available. Record and replay\n"
+            "of code that uses HLE will fail unless you alter your\n"
+            "configuration to make more than one hardware performance counter\n"
+            "available.\n");
+  }
+}
+
+static void check_for_bugs() {
+  if (running_under_rr()) {
+    // Under rr we emulate idealized performance counters, so we can assume
+    // none of the bugs apply.
+    return;
+  }
+
+  check_for_ioc_period_bug();
+  check_for_kvm_in_txcp_bug();
+  check_for_xen_pmi_bug();
+  check_working_counters();
+}
+
+static void init_attributes() {
+  if (attributes_initialized) {
+    return;
+  }
+  attributes_initialized = true;
+
+  CpuMicroarch uarch = get_cpu_microarch();
+  const PmuConfig* pmu = nullptr;
+  for (size_t i = 0; i < array_length(pmu_configs); ++i) {
+    if (uarch == pmu_configs[i].uarch) {
+      pmu = &pmu_configs[i];
+      break;
+    }
+  }
+  DEBUG_ASSERT(pmu);
+
+  if (!pmu->supported) {
+    FATAL() << "Microarchitecture `" << pmu->name << "' currently unsupported.";
+  }
+
+  skid_size = pmu->skid_size;
+  init_perf_event_attr(&ticks_attr, PERF_TYPE_RAW, pmu->rcb_cntr_event);
+  init_perf_event_attr(&cycles_attr, PERF_TYPE_HARDWARE,
+                       PERF_COUNT_HW_CPU_CYCLES);
+  init_perf_event_attr(&hw_interrupts_attr, PERF_TYPE_RAW,
+                       pmu->hw_intr_cntr_event);
+  // libpfm encodes the event with this bit set, so we'll do the
+  // same thing.  Unclear if necessary.
+  hw_interrupts_attr.exclude_hv = 1;
+
+  check_for_bugs();
+  /*
+   * For maintainability, and since it doesn't impact performance when not
+   * needed, we always activate this. If it ever turns out to be a problem,
+   * this can be set to pmu->benefits_from_useless_counter, instead.
+   *
+   * We also disable this counter when running under rr. Even though it's the
+   * same event for the same task as the outer rr, the linux kernel does not
+   * coalesce them and tries to schedule the new one on a general purpose PMC.
+   * On CPUs with only 2 general PMCs (e.g. KNL), we'd run out.
+   */
+  activate_useless_counter = has_ioc_period_bug && !running_under_rr();
+}
+
+bool PerfCounters::is_ticks_attr(const perf_event_attr& attr) {
+  init_attributes();
+  perf_event_attr tmp_attr = attr;
+  tmp_attr.sample_period = 0;
+  tmp_attr.config &= ~IN_TXCP;
+  return memcmp(&ticks_attr, &tmp_attr, sizeof(attr)) == 0;
+}
+
+uint32_t PerfCounters::skid_size() {
+  init_attributes();
+  return rr::skid_size;
+}
+
+PerfCounters::PerfCounters(pid_t tid)
+    : tid(tid), started(false), counting(false) {
+  init_attributes();
 }
 
 static void make_counter_async(ScopedFd& fd, int signal) {
@@ -372,11 +591,11 @@ static void make_counter_async(ScopedFd& fd, int signal) {
 static bool always_recreate_counters() {
   // When we have the KVM IN_TXCP bug, reenabling the TXCP counter after
   // disabling it does not work.
-  return has_ioc_period_bug() || has_kvm_in_txcp_bug();
+  return has_ioc_period_bug || has_kvm_in_txcp_bug;
 }
 
 void PerfCounters::reset(Ticks ticks_period) {
-  assert(ticks_period >= 0);
+  DEBUG_ASSERT(ticks_period >= 0);
 
   if (ticks_period == 0 && !always_recreate_counters()) {
     // We can't switch a counter between sampling and non-sampling via
@@ -391,28 +610,32 @@ void PerfCounters::reset(Ticks ticks_period) {
     attr.sample_period = ticks_period;
     fd_ticks_interrupt = start_counter(tid, -1, &attr);
 
-    if (has_kvm_in_txcp_bug()) {
-      // IN_TXCP isn't going to work reliably. Assume that HLE/RTM are not used,
-      // and check that.
-      attr.sample_period = 0;
-      attr.config |= IN_TX;
-      fd_ticks_in_transaction = start_counter(tid, fd_ticks_interrupt, &attr);
-    } else {
-      // Set up a separate counter for measuring ticks, which does not have
-      // a sample period and does not count events during aborted transactions.
-      // We have to use two separate counters here because the kernel does
-      // not support setting a sample_period with IN_TXCP, apparently for
-      // reasons related to this Intel note on IA32_PERFEVTSEL2:
-      // ``When IN_TXCP=1 & IN_TX=1 and in sampling, spurious PMI may
-      // occur and transactions may continuously abort near overflow
-      // conditions. Software should favor using IN_TXCP for counting over
-      // sampling. If sampling, software should use large “sample-after“
-      // value after clearing the counter configured to use IN_TXCP and
-      // also always reset the counter even when no overflow condition
-      // was reported.''
-      attr.sample_period = 0;
-      attr.config |= IN_TXCP;
-      fd_ticks_measure = start_counter(tid, fd_ticks_interrupt, &attr);
+    if (!only_one_counter && supports_txcp) {
+      if (has_kvm_in_txcp_bug) {
+        // IN_TXCP isn't going to work reliably. Assume that HLE/RTM are not
+        // used,
+        // and check that.
+        attr.sample_period = 0;
+        attr.config |= IN_TX;
+        fd_ticks_in_transaction = start_counter(tid, fd_ticks_interrupt, &attr);
+      } else {
+        // Set up a separate counter for measuring ticks, which does not have
+        // a sample period and does not count events during aborted
+        // transactions.
+        // We have to use two separate counters here because the kernel does
+        // not support setting a sample_period with IN_TXCP, apparently for
+        // reasons related to this Intel note on IA32_PERFEVTSEL2:
+        // ``When IN_TXCP=1 & IN_TX=1 and in sampling, spurious PMI may
+        // occur and transactions may continuously abort near overflow
+        // conditions. Software should favor using IN_TXCP for counting over
+        // sampling. If sampling, software should use large “sample-after“
+        // value after clearing the counter configured to use IN_TXCP and
+        // also always reset the counter even when no overflow condition
+        // was reported.''
+        attr.sample_period = 0;
+        attr.config |= IN_TXCP;
+        fd_ticks_measure = start_counter(tid, fd_ticks_interrupt, &attr);
+      }
     }
 
     if (activate_useless_counter && !fd_useless_counter.is_open()) {
@@ -428,14 +651,6 @@ void PerfCounters::reset(Ticks ticks_period) {
       FATAL() << "Failed to SETOWN_EX ticks event fd";
     }
     make_counter_async(fd_ticks_interrupt, PerfCounters::TIME_SLICE_SIGNAL);
-
-    if (extra_perf_counters_enabled()) {
-      int group_leader = fd_ticks_interrupt;
-      fd_hw_interrupts = start_counter(tid, group_leader, &hw_interrupts_attr);
-      fd_instructions_retired =
-          start_counter(tid, group_leader, &instructions_retired_attr);
-      fd_page_faults = start_counter(tid, group_leader, &page_faults_attr);
-    }
   } else {
     LOG(debug) << "Resetting counters with period " << ticks_period;
 
@@ -469,6 +684,7 @@ void PerfCounters::reset(Ticks ticks_period) {
 
   started = true;
   counting = true;
+  counting_period = ticks_period;
 }
 
 void PerfCounters::set_tid(pid_t tid) {
@@ -484,14 +700,14 @@ void PerfCounters::stop() {
 
   fd_ticks_interrupt.close();
   fd_ticks_measure.close();
-  fd_page_faults.close();
-  fd_hw_interrupts.close();
-  fd_instructions_retired.close();
   fd_useless_counter.close();
   fd_ticks_in_transaction.close();
 }
 
 void PerfCounters::stop_counting() {
+  if (!counting) {
+    return;
+  }
   counting = false;
   if (always_recreate_counters()) {
     stop();
@@ -527,8 +743,14 @@ Ticks PerfCounters::read_ticks(Task* t) {
     }
   }
 
+  uint64_t adjusted_counting_period =
+      counting_period +
+      (t->session().is_recording() ? recording_skid_size() : skid_size());
   uint64_t interrupt_val = read_counter(fd_ticks_interrupt);
   if (!fd_ticks_measure.is_open()) {
+    ASSERT(t, !counting_period || interrupt_val <= adjusted_counting_period)
+        << "Detected " << interrupt_val << " ticks, expected no more than "
+        << adjusted_counting_period;
     return interrupt_val;
   }
 
@@ -542,23 +764,15 @@ Ticks PerfCounters::read_ticks(Task* t) {
     // being used).
     LOG(debug) << "Measured too many ticks; measure=" << measure_val
                << ", interrupt=" << interrupt_val;
+    ASSERT(t, !counting_period || interrupt_val <= adjusted_counting_period)
+        << "Detected " << interrupt_val << " ticks, expected no more than "
+        << adjusted_counting_period;
     return interrupt_val;
   }
+  ASSERT(t, !counting_period || measure_val <= adjusted_counting_period)
+      << "Detected " << measure_val << " ticks, expected no more than "
+      << adjusted_counting_period;
   return measure_val;
 }
 
-PerfCounters::Extra PerfCounters::read_extra() {
-  assert(extra_perf_counters_enabled());
-
-  Extra extra;
-  if (started) {
-    extra.page_faults = read_counter(fd_page_faults);
-    extra.hw_interrupts = read_counter(fd_hw_interrupts);
-    extra.instructions_retired = read_counter(fd_instructions_retired);
-  } else {
-    memset(&extra, 0, sizeof(extra));
-  }
-  return extra;
-}
-
 } // namespace rr
