@@ -42,10 +42,6 @@ public:
 
   void set_tid(pid_t tid);
 
-  // Change this to 'true' to enable perf counters that may be interesting
-  // for experimentation, but aren't necessary for core functionality.
-  static bool extra_perf_counters_enabled() { return false; }
-
   /**
    * Reset all counter values to 0 and program the counters to send
    * TIME_SLICE_SIGNAL when 'ticks_period' tick events have elapsed. (In reality
@@ -85,20 +81,24 @@ public:
    * hope that tracees don't either. */
   enum { TIME_SLICE_SIGNAL = SIGSTKFLT };
 
-  struct Extra {
-    Extra() : page_faults(0), hw_interrupts(0), instructions_retired(0) {}
-
-    int64_t page_faults;
-    int64_t hw_interrupts;
-    int64_t instructions_retired;
-  };
-  Extra read_extra();
-
   static bool is_ticks_attr(const perf_event_attr& attr);
 
-  bool counting;
+  /**
+   * When an interrupt is requested, at most this many ticks may elapse before
+   * the interrupt is delivered.
+   */
+  static uint32_t skid_size();
+
+  /**
+   * Use a separate skid_size for recording since we seem to see more skid
+   * in practice during recording, in particular during the
+   * async_signal_syscalls tests
+   */
+  static uint32_t recording_skid_size() { return skid_size() * 5; }
 
 private:
+  // Only valid while 'counting' is true
+  Ticks counting_period;
   pid_t tid;
   // We use separate fds for counting ticks and for generating interrupts. The
   // former ignores ticks in aborted transactions, and does not support
@@ -107,11 +107,9 @@ private:
   ScopedFd fd_ticks_measure;
   ScopedFd fd_ticks_interrupt;
   ScopedFd fd_ticks_in_transaction;
-  ScopedFd fd_page_faults;
-  ScopedFd fd_hw_interrupts;
-  ScopedFd fd_instructions_retired;
   ScopedFd fd_useless_counter;
   bool started;
+  bool counting;
 };
 
 } // namespace rr
