@@ -5,6 +5,7 @@
 #include "Command.h"
 #include "TraceStream.h"
 #include "TraceTaskEvent.h"
+#include "core.h"
 #include "main.h"
 
 using namespace std;
@@ -13,7 +14,7 @@ namespace rr {
 
 class PsCommand : public Command {
 public:
-  virtual int run(vector<string>& args);
+  virtual int run(vector<string>& args) override;
 
 protected:
   PsCommand(const char* name, const char* help) : Command(name, help) {}
@@ -88,7 +89,7 @@ static int find_exit_code(pid_t pid, const vector<TraceTaskEvent>& events,
       if (status.type() == WaitStatus::EXIT) {
         return status.exit_code();
       }
-      assert(status.type() == WaitStatus::FATAL_SIGNAL);
+      DEBUG_ASSERT(status.type() == WaitStatus::FATAL_SIGNAL);
       return -status.fatal_sig();
     }
     update_tid_to_pid_map(tid_to_pid, e);
@@ -102,8 +103,12 @@ static int ps(const string& trace_dir, FILE* out) {
   fprintf(out, "PID\tPPID\tEXIT\tCMD\n");
 
   vector<TraceTaskEvent> events;
-  while (trace.good()) {
-    events.push_back(trace.read_task_event());
+  while (true) {
+    TraceTaskEvent r = trace.read_task_event();
+    if (r.type() == TraceTaskEvent::NONE) {
+      break;
+    }
+    events.push_back(r);
   }
 
   if (events.empty() || events[0].type() != TraceTaskEvent::EXEC) {
