@@ -2,7 +2,6 @@
 
 #include "RecordCommand.h"
 
-#include <assert.h>
 #include <linux/capability.h>
 #include <sys/prctl.h>
 #include <sysexits.h>
@@ -13,6 +12,7 @@
 #include "RecordSession.h"
 #include "StringVectorToCharArray.h"
 #include "WaitStatus.h"
+#include "core.h"
 #include "kernel_metadata.h"
 #include "log.h"
 #include "main.h"
@@ -102,7 +102,7 @@ struct RecordFlags {
 
   /* Whether tracee processes in record and replay are allowed
    * to run on any logical CPU. */
-  int bind_cpu;
+  BindCPU bind_cpu;
 
   /* True if we should context switch after every rr event */
   bool always_switch;
@@ -130,7 +130,7 @@ struct RecordFlags {
         print_trace_dir(-1),
         use_file_cloning(true),
         use_read_cloning(true),
-        bind_cpu(RecordSession::BIND_CPU),
+        bind_cpu(BIND_CPU),
         always_switch(false),
         chaos(false),
         wait_for_all(false),
@@ -150,7 +150,7 @@ static void parse_signal_name(ParsedOption& opt) {
       opt.int_value = i;
       return;
     }
-    assert(signame[0] == 'S' && signame[1] == 'I' && signame[2] == 'G');
+    DEBUG_ASSERT(signame[0] == 'S' && signame[1] == 'I' && signame[2] == 'G');
     if (signame.substr(3) == opt.value) {
       opt.int_value = i;
       return;
@@ -255,10 +255,10 @@ static bool parse_record_arg(vector<string>& args, RecordFlags& flags) {
       if (!opt.verify_valid_int(0, INT32_MAX)) {
         return false;
       }
-      flags.bind_cpu = opt.int_value;
+      flags.bind_cpu = BindCPU(opt.int_value);
       break;
     case 'u':
-      flags.bind_cpu = RecordSession::UNBOUND_CPU;
+      flags.bind_cpu = UNBOUND_CPU;
       break;
     case 'v':
       flags.extra_env.push_back(opt.value);
@@ -267,7 +267,7 @@ static bool parse_record_arg(vector<string>& args, RecordFlags& flags) {
       flags.wait_for_all = true;
       break;
     default:
-      assert(0 && "Unknown option");
+      DEBUG_ASSERT(0 && "Unknown option");
   }
 
   args = args_copy;
@@ -290,8 +290,8 @@ static void handle_SIGTERM(__attribute__((unused)) int sig) {
     static const char msg[] =
         "Received SIGTERM while an earlier one was pending.  We're "
         "probably wedged.\n";
-    write(STDERR_FILENO, msg, sizeof(msg) - 1);
-    abort();
+    write_all(STDERR_FILENO, msg, sizeof(msg) - 1);
+    notifying_abort();
   }
   term_request = true;
 }
@@ -327,6 +327,16 @@ static void setup_session_from_flags(RecordSession& session,
   }
 }
 
+static RecordSession* static_session;
+
+// This can be called during debugging to close the trace so it can be used
+// later.
+void force_close_record_session() {
+  if (static_session) {
+    static_session->terminate_recording();
+  }
+}
+
 static WaitStatus record(const vector<string>& args, const RecordFlags& flags) {
   LOG(info) << "Start recording...";
 
@@ -334,10 +344,12 @@ static WaitStatus record(const vector<string>& args, const RecordFlags& flags) {
       args, flags.extra_env, flags.use_syscall_buffer, flags.bind_cpu);
   setup_session_from_flags(*session, flags);
 
+  static_session = session.get();
+
   if (flags.print_trace_dir >= 0) {
     const string& dir = session->trace_writer().dir();
-    write(flags.print_trace_dir, dir.c_str(), dir.size());
-    write(flags.print_trace_dir, "\n", 1);
+    write_all(flags.print_trace_dir, dir.c_str(), dir.size());
+    write_all(flags.print_trace_dir, "\n", 1);
   }
 
   // Install signal handlers after creating the session, to ensure they're not
@@ -354,6 +366,7 @@ static WaitStatus record(const vector<string>& args, const RecordFlags& flags) {
   } while (step_result.status == RecordSession::STEP_CONTINUE && !term_request);
 
   session->terminate_recording();
+  static_session = nullptr;
 
   switch (step_result.status) {
     case RecordSession::STEP_CONTINUE:
@@ -368,7 +381,7 @@ static WaitStatus record(const vector<string>& args, const RecordFlags& flags) {
       return WaitStatus::for_exit_code(EX_UNAVAILABLE);
 
     default:
-      assert(0 && "Unknown exit status");
+      DEBUG_ASSERT(0 && "Unknown exit status");
       return WaitStatus();
   }
 }
@@ -397,7 +410,7 @@ static void reset_uid_sudo() {
   // and the capabilities in the child
   std::string sudo_uid = getenv("SUDO_UID");
   std::string sudo_gid = getenv("SUDO_GID");
-  assert(!sudo_uid.empty() && !sudo_gid.empty());
+  DEBUG_ASSERT(!sudo_uid.empty() && !sudo_gid.empty());
   uid_t tracee_uid = stoi(sudo_uid);
   gid_t tracee_gid = stoi(sudo_gid);
   // Setuid will drop effective capabilities. Save them now and set them
@@ -446,7 +459,6 @@ int RecordCommand::run(vector<string>& args) {
   }
 
   assert_prerequisites(flags.use_syscall_buffer);
-  check_performance_settings();
 
   if (flags.setuid_sudo) {
     if (geteuid() != 0 || getenv("SUDO_UID") == NULL) {
@@ -460,6 +472,10 @@ int RecordCommand::run(vector<string>& args) {
   }
 
   WaitStatus status = record(args, flags);
+
+  // Everything should have been cleaned up by now.
+  check_for_leaks();
+
   switch (status.type()) {
     case WaitStatus::EXIT:
       return status.exit_code();
