@@ -9,7 +9,7 @@
 #include "Scheduler.h"
 #include "SeccompFilterRewriter.h"
 #include "Session.h"
-#include "TaskGroup.h"
+#include "ThreadGroup.h"
 #include "TraceFrame.h"
 #include "WaitStatus.h"
 
@@ -26,17 +26,11 @@ public:
    * Create a recording session for the initial command line |argv|.
    */
   enum SyscallBuffering { ENABLE_SYSCALL_BUF, DISABLE_SYSCALL_BUF };
-  /**
-   * BIND_CPU means binding to a randomly chosen CPU.
-   * UNBOUND_CPU means not binding to a particular CPU.
-   * A non-negative value means binding to the specific CPU number.
-   */
-  enum BindCPU { BIND_CPU = -2, UNBOUND_CPU = -1 };
   static shr_ptr create(
       const std::vector<std::string>& argv,
       const std::vector<std::string>& extra_env = std::vector<std::string>(),
       SyscallBuffering syscallbuf = ENABLE_SYSCALL_BUF,
-      int bind_cpu = BIND_CPU);
+      BindCPU bind_cpu = BIND_CPU);
 
   bool use_syscall_buffer() const { return use_syscall_buffer_; }
   size_t syscall_buffer_size() const { return syscall_buffer_size_; }
@@ -81,11 +75,11 @@ public:
    */
   void terminate_recording();
 
-  virtual RecordSession* as_record() { return this; }
+  virtual RecordSession* as_record() override { return this; }
 
   TraceWriter& trace_writer() { return trace_out; }
 
-  virtual void on_destroy(Task* t);
+  virtual void on_destroy(Task* t) override;
 
   Scheduler& scheduler() { return scheduler_; }
 
@@ -116,7 +110,7 @@ public:
   }
 
   virtual Task* new_task(pid_t tid, pid_t rec_tid, uint32_t serial,
-                         SupportedArch a);
+                         SupportedArch a) override;
 
   RecordTask* find_task(pid_t rec_tid) const;
   RecordTask* find_task(const TaskUid& tuid) const;
@@ -136,18 +130,20 @@ private:
   RecordSession(const std::string& exe_path,
                 const std::vector<std::string>& argv,
                 const std::vector<std::string>& envp,
-                SyscallBuffering syscallbuf, int bind_cpu);
+                SyscallBuffering syscallbuf, BindCPU bind_cpu);
 
-  virtual void on_create(Task* t);
+  virtual void on_create(Task* t) override;
 
   void handle_seccomp_traced_syscall(RecordTask* t,
                                      RecordSession::StepState* step_state,
                                      RecordResult* result,
                                      bool* did_enter_syscall);
-  void process_syscall_entry(RecordTask* t, StepState* step_state,
-                             RecordResult* step_result);
+  // Returns false if the task exits during processing
+  bool process_syscall_entry(RecordTask* t, StepState* step_state,
+                             RecordResult* step_result,
+                             SupportedArch syscall_arch);
   void check_initial_task_syscalls(RecordTask* t, RecordResult* step_result);
-  bool handle_ptrace_event(RecordTask* t, StepState* step_state,
+  bool handle_ptrace_event(RecordTask** t_ptr, StepState* step_state,
                            RecordResult* result, bool* did_enter_syscall);
   bool handle_signal_event(RecordTask* t, StepState* step_state);
   void runnable_state_changed(RecordTask* t, StepState* step_state,
@@ -162,7 +158,7 @@ private:
 
   TraceWriter trace_out;
   Scheduler scheduler_;
-  TaskGroup::shr_ptr initial_task_group;
+  ThreadGroup::shr_ptr initial_thread_group;
   SeccompFilterRewriter seccomp_filter_rewriter_;
 
   int ignore_sig;
