@@ -1,6 +1,8 @@
 /* -*- Mode: C++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include <assert.h>
+#include "ReplayCommand.h"
+
+#include <sys/prctl.h>
 #include <sys/time.h>
 #include <sys/wait.h>
 #include <unistd.h>
@@ -12,6 +14,7 @@
 #include "GdbServer.h"
 #include "ReplaySession.h"
 #include "ScopedFd.h"
+#include "core.h"
 #include "kernel_metadata.h"
 #include "log.h"
 #include "main.h"
@@ -22,19 +25,9 @@ namespace rr {
 
 static int DUMP_STATS_PERIOD = 0;
 
-class ReplayCommand : public Command {
-public:
-  virtual int run(vector<string>& args);
-
-protected:
-  ReplayCommand(const char* name, const char* help) : Command(name, help) {}
-
-  static ReplayCommand singleton;
-};
-
 ReplayCommand ReplayCommand::singleton(
     "replay",
-    " rr replay [OPTION]... [<trace-dir>]\n"
+    " rr replay [OPTION]... [<trace-dir>] [-- <debugger-options>]\n"
     "  -a, --autopilot            replay without debugger server\n"
     "  -f, --onfork=<PID>         start a debug server when <PID> has been\n"
     "                             fork()d, AND the target event has been\n"
@@ -51,11 +44,22 @@ ReplayCommand ReplayCommand::singleton(
     "                             has been exec()d, AND the target event has "
     "been\n"
     "                             reached.\n"
+    "  --fullname\n"
+    "  -i=<X>\n"
+    "  --interpreter=<X>          These are passed directly to gdb. They are\n"
+    "                             here for convenience to support 'gdb -i=mi'\n"
+    "                             and 'gdb --fullname' as suggested by GNU "
+    "Emacs\n"
     "  -d, --debugger=<FILE>      use <FILE> as the gdb command\n"
     "  -q, --no-redirect-output   don't replay writes to stdout/stderr\n"
-    "  -s, --dbgport=<PORT>       only start a debug server on <PORT>;\n"
+    "  -h, --dbghost=<HOST>       listen address for the debug server.\n"
+    "                             default listen address is set to localhost.\n"
+    "  -s, --dbgport=<PORT>       only start a debug server on <PORT>,\n"
     "                             don't automatically launch the debugger\n"
-    "                             client too.\n"
+    "                             client; set PORT to 0 to automatically\n"
+    "                             probe a port\n"
+    "  -k, --keep-listening       keep listening after detaching when using \n"
+    "                             --dbgport (-s) mode\n"
     "  -t, --trace=<EVENT>        singlestep instructions and dump register\n"
     "                             states when replaying towards <EVENT> or\n"
     "                             later\n"
@@ -65,9 +69,9 @@ struct ReplayFlags {
   // Start a debug server for the task scheduled at the first
   // event at which reached this event AND target_process has
   // been "created".
-  TraceFrame::Time goto_event;
+  FrameTime goto_event;
 
-  TraceFrame::Time singlestep_to_event;
+  FrameTime singlestep_to_event;
 
   pid_t target_process;
 
@@ -88,6 +92,13 @@ struct ReplayFlags {
   // IP port to listen on for debug connections.
   int dbg_port;
 
+  // IP host to listen on for debug connections.
+  string dbg_host;
+
+  // Whether to keep listening with a new server after the existing server
+  // detaches
+  bool keep_listening;
+
   // Pass these options to gdb
   vector<string> gdb_options;
 
@@ -108,6 +119,8 @@ struct ReplayFlags {
         process_created_how(CREATED_NONE),
         dont_launch_debugger(false),
         dbg_port(-1),
+        dbg_host(localhost_addr),
+        keep_listening(false),
         gdb_binary_file_path("gdb"),
         redirect(true),
         share_private_mappings(false) {}
@@ -121,15 +134,19 @@ static bool parse_replay_arg(vector<string>& args, ReplayFlags& flags) {
   static const OptionSpec options[] = {
     { 'a', "autopilot", NO_PARAMETER },
     { 'd', "debugger", HAS_PARAMETER },
+    { 'k', "keep-listening", NO_PARAMETER },
     { 'f', "onfork", HAS_PARAMETER },
     { 'g', "goto", HAS_PARAMETER },
     { 'o', "debugger-option", HAS_PARAMETER },
     { 'p', "onprocess", HAS_PARAMETER },
     { 'q', "no-redirect-output", NO_PARAMETER },
+    { 'h', "dbghost", HAS_PARAMETER },
     { 's', "dbgport", HAS_PARAMETER },
     { 't', "trace", HAS_PARAMETER },
     { 'x', "gdb-x", HAS_PARAMETER },
-    { 0, "share-private-mappings", NO_PARAMETER }
+    { 0, "share-private-mappings", NO_PARAMETER },
+    { 1, "fullname", NO_PARAMETER },
+    { 'i', "interpreter", HAS_PARAMETER }
   };
   ParsedOption opt;
   if (!Command::parse_option(args, options, &opt)) {
@@ -157,6 +174,9 @@ static bool parse_replay_arg(vector<string>& args, ReplayFlags& flags) {
       }
       flags.goto_event = opt.int_value;
       break;
+    case 'k':
+      flags.keep_listening = true;
+      break;
     case 'o':
       flags.gdb_options.push_back(opt.value);
       break;
@@ -174,8 +194,12 @@ static bool parse_replay_arg(vector<string>& args, ReplayFlags& flags) {
     case 'q':
       flags.redirect = false;
       break;
+    case 'h':
+      flags.dbg_host = opt.value;
+      flags.dont_launch_debugger = true;
+      break;
     case 's':
-      if (!opt.verify_valid_int(1, INT32_MAX)) {
+      if (!opt.verify_valid_int(0, INT32_MAX)) {
         return false;
       }
       flags.dbg_port = opt.int_value;
@@ -194,8 +218,15 @@ static bool parse_replay_arg(vector<string>& args, ReplayFlags& flags) {
     case 0:
       flags.share_private_mappings = true;
       break;
+    case 1:
+      flags.gdb_options.push_back("--fullname");
+      break;
+    case 'i':
+      flags.gdb_options.push_back("-i");
+      flags.gdb_options.push_back(opt.value);
+      break;
     default:
-      assert(0 && "Unknown option");
+      DEBUG_ASSERT(0 && "Unknown option");
   }
   return true;
 }
@@ -204,8 +235,11 @@ static int find_pid_for_command(const string& trace_dir,
                                 const string& command) {
   TraceReader trace(trace_dir);
 
-  while (trace.good()) {
-    auto e = trace.read_task_event();
+  while (true) {
+    TraceTaskEvent e = trace.read_task_event();
+    if (e.type() == TraceTaskEvent::NONE) {
+      return -1;
+    }
     if (e.type() != TraceTaskEvent::EXEC) {
       continue;
     }
@@ -219,31 +253,34 @@ static int find_pid_for_command(const string& trace_dir,
       return e.tid();
     }
   }
-  return -1;
 }
 
 static bool pid_exists(const string& trace_dir, pid_t pid) {
   TraceReader trace(trace_dir);
 
-  while (trace.good()) {
+  while (true) {
     auto e = trace.read_task_event();
+    if (e.type() == TraceTaskEvent::NONE) {
+      return false;
+    }
     if (e.tid() == pid) {
       return true;
     }
   }
-  return false;
 }
 
 static bool pid_execs(const string& trace_dir, pid_t pid) {
   TraceReader trace(trace_dir);
 
-  while (trace.good()) {
+  while (true) {
     auto e = trace.read_task_event();
+    if (e.type() == TraceTaskEvent::NONE) {
+      return false;
+    }
     if (e.tid() == pid && e.type() == TraceTaskEvent::EXEC) {
       return true;
     }
   }
-  return false;
 }
 
 // The parent process waits until the server, |waiting_for_child|, creates a
@@ -277,18 +314,18 @@ static void serve_replay_no_debugger(const string& trace_dir,
     if (flags.singlestep_to_event > 0 &&
         replay_session->trace_reader().time() >= flags.singlestep_to_event) {
       cmd = RUN_SINGLESTEP;
-      fputs("Stepping from:", stderr);
+      fputs("Stepping from: ", stderr);
       Task* t = replay_session->current_task();
       t->regs().print_register_file_compact(stderr);
-      fputs(" ", stderr);
+      fputc(' ', stderr);
       t->extra_regs().print_register_file_compact(stderr);
       fprintf(stderr, " ticks:%" PRId64 "\n", t->tick_count());
     }
 
-    TraceFrame::Time before_time = replay_session->trace_reader().time();
+    FrameTime before_time = replay_session->trace_reader().time();
     auto result = replay_session->replay_step(cmd);
-    TraceFrame::Time after_time = replay_session->trace_reader().time();
-    assert(after_time >= before_time && after_time <= before_time + 1);
+    FrameTime after_time = replay_session->trace_reader().time();
+    DEBUG_ASSERT(after_time >= before_time && after_time <= before_time + 1);
 
     ++step_count;
     if (DUMP_STATS_PERIOD > 0 && step_count % DUMP_STATS_PERIOD == 0) {
@@ -309,10 +346,11 @@ static void serve_replay_no_debugger(const string& trace_dir,
     if (result.status == REPLAY_EXITED) {
       break;
     }
-    assert(result.status == REPLAY_CONTINUE);
-    assert(result.break_status.watchpoints_hit.empty());
-    assert(!result.break_status.breakpoint_hit);
-    assert(cmd == RUN_SINGLESTEP || !result.break_status.singlestep_complete);
+    DEBUG_ASSERT(result.status == REPLAY_CONTINUE);
+    DEBUG_ASSERT(result.break_status.watchpoints_hit.empty());
+    DEBUG_ASSERT(!result.break_status.breakpoint_hit);
+    DEBUG_ASSERT(cmd == RUN_SINGLESTEP ||
+                 !result.break_status.singlestep_complete);
   }
 
   LOG(info) << "Replayer successfully finished";
@@ -332,14 +370,14 @@ static void serve_replay_no_debugger(const string& trace_dir,
  * why we use a signal handler instead of SIG_IGN).
  */
 static void handle_SIGINT_in_parent(int sig) {
-  assert(sig == SIGINT);
+  DEBUG_ASSERT(sig == SIGINT);
   // Just ignore it.
 }
 
 static GdbServer* server_ptr = nullptr;
 
 static void handle_SIGINT_in_child(int sig) {
-  assert(sig == SIGINT);
+  DEBUG_ASSERT(sig == SIGINT);
   if (server_ptr) {
     server_ptr->interrupt_replay_to_target();
   }
@@ -371,9 +409,14 @@ static int replay(const string& trace_dir, const ReplayFlags& flags) {
       auto session = ReplaySession::create(trace_dir);
       GdbServer::ConnectionFlags conn_flags;
       conn_flags.dbg_port = flags.dbg_port;
+      conn_flags.dbg_host = flags.dbg_host;
       conn_flags.debugger_name = flags.gdb_binary_file_path;
+      conn_flags.keep_listening = flags.keep_listening;
       GdbServer(session, session_flags(flags), target).serve_replay(conn_flags);
     }
+
+    // Everything should have been cleaned up by now.
+    check_for_leaks();
     return 0;
   }
 
@@ -386,23 +429,30 @@ static int replay(const string& trace_dir, const ReplayFlags& flags) {
     // the parent dies, our writes to the pipe will error out.
     close(debugger_params_pipe[0]);
 
-    ScopedFd debugger_params_write_pipe(debugger_params_pipe[1]);
-    auto session = ReplaySession::create(trace_dir);
-    GdbServer::ConnectionFlags conn_flags;
-    conn_flags.dbg_port = flags.dbg_port;
-    conn_flags.debugger_params_write_pipe = &debugger_params_write_pipe;
-    GdbServer server(session, session_flags(flags), target);
+    {
+      prctl(PR_SET_PDEATHSIG, SIGTERM, 0, 0, 0);
 
-    server_ptr = &server;
-    struct sigaction sa;
-    memset(&sa, 0, sizeof(sa));
-    sa.sa_flags = SA_RESTART;
-    sa.sa_handler = handle_SIGINT_in_child;
-    if (sigaction(SIGINT, &sa, nullptr)) {
-      FATAL() << "Couldn't set sigaction for SIGINT.";
+      ScopedFd debugger_params_write_pipe(debugger_params_pipe[1]);
+      auto session = ReplaySession::create(trace_dir);
+      GdbServer::ConnectionFlags conn_flags;
+      conn_flags.dbg_port = flags.dbg_port;
+      conn_flags.dbg_host = flags.dbg_host;
+      conn_flags.debugger_params_write_pipe = &debugger_params_write_pipe;
+      GdbServer server(session, session_flags(flags), target);
+
+      server_ptr = &server;
+      struct sigaction sa;
+      memset(&sa, 0, sizeof(sa));
+      sa.sa_flags = SA_RESTART;
+      sa.sa_handler = handle_SIGINT_in_child;
+      if (sigaction(SIGINT, &sa, nullptr)) {
+        FATAL() << "Couldn't set sigaction for SIGINT.";
+      }
+
+      server.serve_replay(conn_flags);
     }
-
-    server.serve_replay(conn_flags);
+    // Everything should have been cleaned up by now.
+    check_for_leaks();
     return 0;
   }
   // Ensure only the child has the write end of the pipe open. Then if
@@ -456,6 +506,11 @@ int ReplayCommand::run(vector<string>& args) {
     if (parse_replay_arg(args, flags)) {
       continue;
     }
+    if (parse_literal(args, "--")) {
+      flags.gdb_options.insert(flags.gdb_options.end(), args.begin(),
+                               args.end());
+      break;
+    }
     if (!found_dir && parse_optional_trace_dir(args, &trace_dir)) {
       found_dir = true;
       continue;
@@ -489,7 +544,6 @@ int ReplayCommand::run(vector<string>& args) {
   }
 
   assert_prerequisites();
-  check_performance_settings();
 
   if (running_under_rr()) {
     if (!Flags::get().suppress_environment_warnings) {
@@ -504,6 +558,12 @@ int ReplayCommand::run(vector<string>& args) {
     }
   }
 
+  if (flags.keep_listening && flags.dbg_port == -1) {
+    fprintf(stderr,
+            "Cannot use --keep-listening (-k) without --dbgport (-s).\n");
+    return 4;
+  }
+
   return replay(trace_dir, flags);
 }
 
