@@ -110,7 +110,7 @@ struct ReplayResult {
 
 /**
  * An indicator of how much progress the ReplaySession has made within a given
- * (TraceFrame::Time, Ticks) pair. These can only be used for comparisons, to
+ * (FrameTime, Ticks) pair. These can only be used for comparisons, to
  * check whether two ReplaySessions are in the same state and to help
  * order their states temporally.
  */
@@ -145,7 +145,7 @@ public:
   ~ReplaySession();
 
   virtual Task* new_task(pid_t tid, pid_t rec_tid, uint32_t serial,
-                         SupportedArch a);
+                         SupportedArch a) override;
 
   using Session::clone;
   /**
@@ -187,6 +187,10 @@ public:
    * for replay to reach.
    */
   const TraceFrame& current_trace_frame() const { return trace_frame; }
+  /**
+   * Time of the current frame
+   */
+  FrameTime current_frame_time() const { return trace_frame.time(); }
 
   /**
    * The Task for the current trace record.
@@ -203,7 +207,7 @@ public:
    * Returns true if the next step for this session is to exit a syscall with
    * the given number.
    */
-  bool next_step_is_syscall_exit(int syscallno);
+  bool next_step_is_successful_syscall_exit(int syscallno);
 
   /**
    * The current ReplayStepKey.
@@ -226,7 +230,7 @@ public:
     explicit StepConstraints(RunCommand command)
         : command(command), stop_at_time(0), ticks_target(0) {}
     RunCommand command;
-    TraceFrame::Time stop_at_time;
+    FrameTime stop_at_time;
     Ticks ticks_target;
     // When the RunCommand is RUN_SINGLESTEP_FAST_FORWARD, stop if the next
     // singlestep would enter one of the register states in this list.
@@ -258,7 +262,7 @@ public:
     return replay_step(StepConstraints(command));
   }
 
-  virtual ReplaySession* as_replay() { return this; }
+  virtual ReplaySession* as_replay() override { return this; }
 
   /**
    * Return true if |sig| is a signal that may be generated during
@@ -304,11 +308,13 @@ private:
   Completion enter_syscall(ReplayTask* t, const StepConstraints& constraints);
   Completion exit_syscall(ReplayTask* t);
   Completion exit_task(ReplayTask* t);
+  bool handle_unrecorded_cpuid_fault(ReplayTask* t,
+                                     const StepConstraints& constraints);
   void check_ticks_consistency(ReplayTask* t, const Event& ev);
   void check_pending_sig(ReplayTask* t);
-  void continue_or_step(ReplayTask* t, const StepConstraints& constraints,
-                        TicksRequest tick_request,
-                        ResumeRequest resume_how = RESUME_SYSEMU);
+  Completion continue_or_step(ReplayTask* t, const StepConstraints& constraints,
+                              TicksRequest tick_request,
+                              ResumeRequest resume_how = RESUME_SYSEMU);
   Completion advance_to_ticks_target(ReplayTask* t,
                                      const StepConstraints& constraints);
   Completion emulate_deterministic_signal(ReplayTask* t, int sig,
