@@ -4,6 +4,7 @@
 
 #include <math.h>
 
+#include "core.h"
 #include "fast_forward.h"
 #include "log.h"
 
@@ -38,7 +39,7 @@ ostream& operator<<(ostream& s, const ReplayTimeline::ProtoMark& o) {
 }
 
 bool ReplayTimeline::less_than(const Mark& m1, const Mark& m2) {
-  assert(m1.ptr->owner == m2.ptr->owner);
+  DEBUG_ASSERT(m1.ptr->owner == m2.ptr->owner);
   if (m1.ptr->proto.key < m2.ptr->proto.key) {
     return true;
   }
@@ -56,7 +57,7 @@ bool ReplayTimeline::less_than(const Mark& m1, const Mark& m2) {
       return true;
     }
   }
-  assert(0 && "Marks missing from vector, invariants broken!");
+  DEBUG_ASSERT(0 && "Marks missing from vector, invariants broken!");
   return false;
 }
 
@@ -99,7 +100,11 @@ bool ReplayTimeline::ProtoMark::equal_states(ReplaySession& session) const {
 }
 
 ReplayTimeline::ProtoMark ReplayTimeline::proto_mark() const {
-  return ProtoMark(current_mark_key(), current->current_task());
+  ReplayTask* t = current->current_task();
+  if (t) {
+    return ProtoMark(current_mark_key(), t);
+  }
+  return ProtoMark(current_mark_key());
 }
 
 shared_ptr<ReplayTimeline::InternalMark> ReplayTimeline::current_mark() {
@@ -146,7 +151,7 @@ ReplayTimeline::Mark ReplayTimeline::mark() {
     vector<shared_ptr<InternalMark>> new_marks;
     new_marks.push_back(m);
 
-    LOG(debug) << "mark() replaying to find mark location";
+    LOG(debug) << "mark() replaying to find mark location for " << *m;
 
     // Allow coalescing of multiple repetitions of a single x86 string
     // instruction (as long as we don't reach one of our mark_vector states).
@@ -191,8 +196,6 @@ ReplayTimeline::Mark ReplayTimeline::mark() {
       new_marks.push_back(make_shared<InternalMark>(this, *tmp_session, key));
     }
 
-    LOG(debug) << "Mark location found";
-
     // mark_index is the current index of the next mark after 'current'. So
     // insert our new marks at mark_index.
     mark_vector.insert(mark_index, new_marks.begin(), new_marks.end());
@@ -204,7 +207,7 @@ ReplayTimeline::Mark ReplayTimeline::mark() {
 
 void ReplayTimeline::mark_after_singlestep(const Mark& from,
                                            const ReplayResult& result) {
-  assert(result.break_status.singlestep_complete);
+  DEBUG_ASSERT(result.break_status.singlestep_complete);
   Mark m = mark();
   if (!result.did_fast_forward && m.ptr->proto.key == from.ptr->proto.key &&
       !result.break_status.signal) {
@@ -215,8 +218,8 @@ void ReplayTimeline::mark_after_singlestep(const Mark& from,
           for (size_t j = 0; j < mark_vector.size(); ++j) {
             LOG(debug) << "  mark_vector[" << j << "] " << *mark_vector[j];
           }
-          ASSERT(result.break_status.task, false) << " expected to find " << m
-                                                  << " at index " << i + 1;
+          ASSERT(result.break_status.task, false)
+              << " expected to find " << m << " at index " << i + 1;
         }
         break;
       }
@@ -233,7 +236,7 @@ ReplayTimeline::Mark ReplayTimeline::find_singlestep_before(const Mark& mark) {
       break;
     }
   }
-  assert(i >= 0 && "Mark not in vector???");
+  DEBUG_ASSERT(i >= 0 && "Mark not in vector???");
 
   Mark m;
   if (i == 0) {
@@ -261,7 +264,7 @@ ReplayTimeline::Mark ReplayTimeline::lazy_reverse_singlestep(const Mark& from,
 }
 
 ReplayTimeline::Mark ReplayTimeline::add_explicit_checkpoint() {
-  assert(current->can_clone());
+  DEBUG_ASSERT(current->can_clone());
 
   Mark m = mark();
   if (!m.ptr->checkpoint) {
@@ -279,14 +282,14 @@ ReplayTimeline::Mark ReplayTimeline::add_explicit_checkpoint() {
 }
 
 void ReplayTimeline::remove_mark_with_checkpoint(const MarkKey& key) {
-  assert(marks_with_checkpoints[key] > 0);
+  DEBUG_ASSERT(marks_with_checkpoints[key] > 0);
   if (--marks_with_checkpoints[key] == 0) {
     marks_with_checkpoints.erase(key);
   }
 }
 
 void ReplayTimeline::remove_explicit_checkpoint(const Mark& mark) {
-  assert(mark.ptr->checkpoint_refcount > 0);
+  DEBUG_ASSERT(mark.ptr->checkpoint_refcount > 0);
   if (--mark.ptr->checkpoint_refcount == 0) {
     mark.ptr->checkpoint = nullptr;
     remove_mark_with_checkpoint(mark.ptr->proto.key);
@@ -327,7 +330,7 @@ void ReplayTimeline::seek_to_before_key(const MarkKey& key) {
           break;
         }
       }
-      assert(current);
+      DEBUG_ASSERT(current);
       breakpoints_applied = false;
       current_at_or_after_mark = nullptr;
     }
@@ -396,7 +399,10 @@ void ReplayTimeline::update_strategy_and_fix_watchpoint_quirk(
 
 ReplayResult ReplayTimeline::replay_step_to_mark(
     const Mark& mark, ReplayStepToMarkStrategy& strategy) {
+  ReplayTask* t = current->current_task();
   ProtoMark before = proto_mark();
+  ASSERT(t, before.key <= mark.ptr->proto.key)
+      << "Current mark " << before << " is already after target " << mark;
   ReplayResult result;
   if (current->trace_reader().time() < mark.ptr->proto.key.trace_time) {
     // Easy case: each RUN_CONTINUE can only advance by at most one
@@ -411,7 +417,6 @@ ReplayResult ReplayTimeline::replay_step_to_mark(
     return result;
   }
 
-  ReplayTask* t = current->current_task();
   ASSERT(t, current->trace_reader().time() == mark.ptr->proto.key.trace_time);
   // t must remain valid through here since t can only die when we complete
   // an event, and we're not going to complete another event before
@@ -752,15 +757,15 @@ bool ReplayTimeline::run_forward_to_intermediate_point(const Mark& end,
              << current_mark_key() << " and " << end
              << (force == FORCE_PROGRESS ? " (forced)" : "");
 
-  TraceFrame::Time now = current->trace_reader().time();
-  TraceFrame::Time mid = (now + end.ptr->proto.key.trace_time) / 2;
+  FrameTime now = current->trace_reader().time();
+  FrameTime mid = (now + end.ptr->proto.key.trace_time) / 2;
   if (now < mid && mid < end.ptr->proto.key.trace_time) {
     ReplaySession::StepConstraints constraints(RUN_CONTINUE);
     constraints.stop_at_time = mid;
     while (current->trace_reader().time() < mid) {
       current->replay_step(constraints);
     }
-    assert(current->trace_reader().time() == mid);
+    DEBUG_ASSERT(current->trace_reader().time() == mid);
     LOG(debug) << "Ran forward to mid event " << current_mark_key();
     return true;
   }
@@ -772,7 +777,8 @@ bool ReplayTimeline::run_forward_to_intermediate_point(const Mark& end,
     while (current->trace_reader().time() < end.ptr->proto.key.trace_time) {
       current->replay_step(constraints);
     }
-    assert(current->trace_reader().time() == end.ptr->proto.key.trace_time);
+    DEBUG_ASSERT(current->trace_reader().time() ==
+                 end.ptr->proto.key.trace_time);
     LOG(debug) << "Ran forward to event " << current_mark_key();
     return true;
   }
@@ -804,8 +810,8 @@ bool ReplayTimeline::run_forward_to_intermediate_point(const Mark& end,
       LOG(debug) << "Ran forward to " << current_mark_key();
       return true;
     }
-    assert(result.break_status.approaching_ticks_target);
-    assert(t->tick_count() == start_ticks);
+    DEBUG_ASSERT(result.break_status.approaching_ticks_target);
+    DEBUG_ASSERT(t->tick_count() == start_ticks);
   }
 
   // We didn't make any progress that way.
@@ -832,7 +838,7 @@ bool ReplayTimeline::run_forward_to_intermediate_point(const Mark& end,
     constraints.stop_before_states.push_back(&end.ptr->proto.regs);
     ReplayResult result = current->replay_step(constraints);
     if (at_mark(end)) {
-      assert(tmp_session);
+      DEBUG_ASSERT(tmp_session);
       current = move(tmp_session);
       LOG(debug) << "Singlestepping arrived at |end|, restoring session";
     } else if (!m.equal_states(*current)) {
@@ -899,6 +905,9 @@ ReplayResult ReplayTimeline::reverse_continue(
     }
     maybe_add_reverse_exec_checkpoint(EXPECT_SHORT_REVERSE_EXECUTION);
 
+    LOG(debug) << "reverse-continue continuing forward from " << start
+               << " up to " << end;
+
     bool at_breakpoint = false;
     ReplayStepToMarkStrategy strategy;
     int stop_count = 0;
@@ -946,9 +955,12 @@ ReplayResult ReplayTimeline::reverse_continue(
       if (!result.break_status.watchpoints_hit.empty() ||
           result.break_status.signal) {
         dest = mark();
-        LOG(debug) << "Found "
-                   << (result.break_status.signal ? "signal" : "watch")
-                   << " break at " << dest;
+        if (result.break_status.signal) {
+          LOG(debug) << "Found signal break at " << dest;
+        } else {
+          LOG(debug) << "Found watch break at " << dest << ", addr="
+                     << result.break_status.watchpoints_hit[0].addr;
+        }
         final_result = result;
         final_tuid = result.break_status.task ? result.break_status.task->tuid()
                                               : TaskUid();
@@ -957,7 +969,7 @@ ReplayResult ReplayTimeline::reverse_continue(
                           : 0;
         last_stop_is_watch_or_signal = true;
       }
-      assert(result.status == REPLAY_CONTINUE);
+      DEBUG_ASSERT(result.status == REPLAY_CONTINUE);
 
       if (is_start_of_reverse_execution_barrier_event()) {
         dest = mark();
@@ -1001,10 +1013,11 @@ ReplayResult ReplayTimeline::reverse_continue(
         ++stop_count;
         if (stop_count > stop_count_limit) {
           Mark before_running = mark();
-          if (run_forward_to_intermediate_point(end, made_progress_between_stops
-                                                         ? DONT_FORCE_PROGRESS
-                                                         : FORCE_PROGRESS)) {
-            assert(!at_mark(end));
+          if (run_forward_to_intermediate_point(end,
+                                                made_progress_between_stops
+                                                    ? DONT_FORCE_PROGRESS
+                                                    : FORCE_PROGRESS)) {
+            DEBUG_ASSERT(!at_mark(end));
             // We made some progress towards |end| with breakpoints/watchpoints
             // disabled, without reaching |end|. Continuing running forward from
             // here with breakpoints/watchpoints enabled. If we need to seek
@@ -1149,7 +1162,7 @@ ReplayResult ReplayTimeline::reverse_singlestep(
       }
       end = start;
     }
-    assert(stop_filter(current->current_task()) || seen_barrier);
+    DEBUG_ASSERT(stop_filter(current->current_task()) || seen_barrier);
 
     Mark destination_candidate;
     Mark step_start = set_short_checkpoint();
@@ -1191,6 +1204,8 @@ ReplayResult ReplayTimeline::reverse_singlestep(
               break;
             }
             destination_candidate = step_start;
+            LOG(debug) << "Setting candidate after step: "
+                       << destination_candidate;
             destination_candidate_result = result;
             destination_candidate_tuid = result.break_status.task->tuid();
             seen_other_task_break = false;
@@ -1220,6 +1235,7 @@ ReplayResult ReplayTimeline::reverse_singlestep(
 
       if (is_start_of_reverse_execution_barrier_event()) {
         destination_candidate = mark();
+        LOG(debug) << "Setting candidate to barrier " << destination_candidate;
         destination_candidate_result = result;
         destination_candidate_result.break_status.task_exit = true;
         destination_candidate_tuid = current->current_task()->tuid();
@@ -1227,6 +1243,7 @@ ReplayResult ReplayTimeline::reverse_singlestep(
       }
 
       if (now >= end) {
+        LOG(debug) << "Stepped to " << now << " (>= " << end << "), stopping";
         break;
       }
       maybe_add_reverse_exec_checkpoint(EXPECT_SHORT_REVERSE_EXECUTION);
@@ -1246,7 +1263,7 @@ ReplayResult ReplayTimeline::reverse_singlestep(
       seek_to_mark(destination_candidate);
       destination_candidate_result.break_status.task =
           current->find_task(destination_candidate_tuid);
-      assert(destination_candidate_result.break_status.task);
+      DEBUG_ASSERT(destination_candidate_result.break_status.task);
       evaluate_conditions(destination_candidate_result);
       return destination_candidate_result;
     }
@@ -1306,9 +1323,9 @@ void ReplayTimeline::evaluate_conditions(ReplayResult& result) {
   }
 }
 
-ReplayResult ReplayTimeline::replay_step_forward(
-    RunCommand command, TraceFrame::Time stop_at_time) {
-  assert(command != RUN_SINGLESTEP_FAST_FORWARD);
+ReplayResult ReplayTimeline::replay_step_forward(RunCommand command,
+                                                 FrameTime stop_at_time) {
+  DEBUG_ASSERT(command != RUN_SINGLESTEP_FAST_FORWARD);
 
   ReplayResult result;
   apply_breakpoints_and_watchpoints();
