@@ -1,6 +1,5 @@
 /* -*- Mode: C++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include <assert.h>
 #include <sys/time.h>
 #include <sys/wait.h>
 #include <unistd.h>
@@ -12,6 +11,7 @@
 #include "GdbServer.h"
 #include "ReplaySession.h"
 #include "ScopedFd.h"
+#include "core.h"
 #include "kernel_metadata.h"
 #include "log.h"
 #include "main.h"
@@ -28,7 +28,7 @@ namespace rr {
  */
 class RerunCommand : public Command {
 public:
-  virtual int run(vector<string>& args);
+  virtual int run(vector<string>& args) override;
 
 protected:
   RerunCommand(const char* name, const char* help) : Command(name, help) {}
@@ -40,8 +40,10 @@ RerunCommand RerunCommand::singleton(
     "rerun",
     " rr rerun [OPTION]... [<trace-dir>]\n"
     "  -e, --trace-end=<EVENT>    end tracing at <EVENT>\n"
-    "  -r, --singlestep-registers=<REGS>\n"
-    "                             dump registers <REGS> after each singlestep\n"
+    "  -f, --function=<ADDR>      when starting tracing, push sentinel return\n"
+    "                             address and jump to <ADDR> to fake call\n"
+    "  --singlestep=<REGS>        dump <REGS> after each singlestep\n"
+    "  -r, --raw                  dump registers in raw format\n"
     "  -s, --trace-start=<EVENT>  start tracing at <EVENT>\n"
     "\n"
     "<REGS> is a comma-separated sequence of 'event','icount','ip','flags',\n"
@@ -58,7 +60,12 @@ enum TraceFieldKind {
   TRACE_EVENT_NUMBER,      // outputs 64-bit value
   TRACE_INSTRUCTION_COUNT, // outputs 64-bit value
   TRACE_IP,                // outputs 64-bit value
+  TRACE_FSBASE,            // outputs 64-bit value
+  TRACE_GSBASE,            // outputs 64-bit value
   TRACE_FLAGS,             // outputs 64-bit value
+  TRACE_ORIG_AX,           // outputs 64-bit value
+  TRACE_SEG_REG,           // outputs 64-bit value
+  TRACE_XINUSE,            // outputs 64-bit value
   TRACE_GP_REG,            // outputs 64-bit value
   TRACE_XMM_REG,           // outputs 128-bit value
   TRACE_YMM_REG,           // outputs 256-bit value
@@ -68,6 +75,19 @@ struct TraceField {
   uint8_t reg_num;
 };
 
+struct RerunFlags {
+  FrameTime trace_start;
+  FrameTime trace_end;
+  remote_code_ptr function;
+  vector<TraceField> singlestep_trace;
+  bool raw;
+
+  RerunFlags()
+      : trace_start(0),
+        trace_end(numeric_limits<decltype(trace_end)>::max()),
+        raw(false) {}
+};
+
 #ifdef __x86_64__
 static uint8_t user_regs_fields[16] = {
   offsetof(user_regs_struct, rax), offsetof(user_regs_struct, rcx),
@@ -90,33 +110,119 @@ static uint8_t user_regs_fields[16] = {
 #error Unsupported architecture
 #endif
 
-static void print_regs_raw(Task* t, TraceFrame::Time event,
-                           uint64_t instruction_count,
-                           const vector<TraceField>& fields, FILE* out) {
+static void print_hex(uint8_t* value, size_t size, FILE* out) {
+  bool any_printed = false;
+  for (ssize_t i = size - 1; i >= 0; --i) {
+    if (value[i] || any_printed || i == 0) {
+      fprintf(out, any_printed ? "%02x" : "%x", value[i]);
+      any_printed = true;
+    }
+  }
+}
+
+static void print_value(const char* name, void* value, size_t size,
+                        const RerunFlags& flags, FILE* out) {
+  if (flags.raw) {
+    fwrite(value, size, 1, out);
+  } else {
+    fprintf(out, "%s:0x", name);
+    print_hex(static_cast<uint8_t*>(value), size, out);
+  }
+}
+
+static const char gp_reg_names[16][4] = { "rax", "rcx", "rdx", "rbx",
+                                          "rsp", "rbp", "rsi", "rdi",
+                                          "r8",  "r9",  "r10", "r11",
+                                          "r12", "r13", "r14", "r15" };
+static const char gp_reg_names_32[8][4] = { "eax", "ecx", "edx", "ebx",
+                                            "esp", "ebp", "esi", "edi" };
+
+static const char seg_reg_names[6][3] = { "es", "cs", "ss", "ds", "fs", "gs" };
+
+static uint64_t seg_reg(const Registers& regs, uint8_t index) {
+  switch (index) {
+    case 0:
+      return regs.es();
+    case 1:
+      return regs.cs();
+    case 2:
+      return regs.ss();
+    case 3:
+      return regs.ds();
+    case 4:
+      return regs.fs();
+    case 5:
+      return regs.gs();
+    default:
+      FATAL() << "Unknown seg reg";
+      return 0;
+  }
+}
+
+static void print_regs(Task* t, FrameTime event, uint64_t instruction_count,
+                       const RerunFlags& flags, FILE* out) {
   union {
     struct user_regs_struct gp_regs;
     uintptr_t regs_values[sizeof(struct user_regs_struct) / sizeof(uintptr_t)];
   };
   bool got_gp_regs = false;
+  const vector<TraceField>& fields = flags.singlestep_trace;
+  bool first = true;
 
   for (auto& field : fields) {
+    if (first) {
+      first = false;
+    } else if (!flags.raw) {
+      fputc(' ', out);
+    }
     switch (field.kind) {
       case TRACE_EVENT_NUMBER: {
         uint64_t value = event;
-        fwrite(&value, sizeof(value), 1, out);
+        print_value("event", &value, sizeof(value), flags, out);
         break;
       }
       case TRACE_INSTRUCTION_COUNT:
-        fwrite(&instruction_count, sizeof(instruction_count), 1, out);
+        print_value("icount", &instruction_count, sizeof(instruction_count),
+                    flags, out);
         break;
       case TRACE_IP: {
         uint64_t value = t->regs().ip().register_value();
-        fwrite(&value, sizeof(value), 1, out);
+        print_value(t->arch() == x86 ? "eip" : "rip", &value, sizeof(value),
+                    flags, out);
+        break;
+      }
+      case TRACE_FSBASE: {
+        uint64_t value = t->regs().fs_base();
+        print_value("fsbase", &value, sizeof(value), flags, out);
+        break;
+      }
+      case TRACE_GSBASE: {
+        uint64_t value = t->regs().gs_base();
+        print_value("gsbase", &value, sizeof(value), flags, out);
         break;
       }
       case TRACE_FLAGS: {
         uint64_t value = t->regs().flags();
-        fwrite(&value, sizeof(value), 1, out);
+        print_value(t->arch() == x86 ? "eflags" : "rflags", &value,
+                    sizeof(value), flags, out);
+        break;
+      }
+      case TRACE_ORIG_AX: {
+        uint64_t value = t->regs().original_syscallno();
+        print_value(t->arch() == x86 ? "orig_eax" : "orig_rax", &value,
+                    sizeof(value), flags, out);
+        break;
+      }
+      case TRACE_SEG_REG: {
+        uint64_t value = seg_reg(t->regs(), field.reg_num);
+        print_value(seg_reg_names[field.reg_num], &value, sizeof(value), flags,
+                    out);
+        break;
+      }
+      case TRACE_XINUSE: {
+        bool defined;
+        uint64_t value = t->extra_regs().read_xinuse(&defined);
+        print_value("xinuse", &value, sizeof(value), flags, out);
         break;
       }
       case TRACE_GP_REG: {
@@ -131,7 +237,10 @@ static void print_regs_raw(Task* t, TraceFrame::Time event,
           // EAX->RAX is sign-extended, so undo that.
           value = (uint32_t)value;
         }
-        fwrite(&value, sizeof(value), 1, out);
+        const char* name = (t->arch() == x86 && field.reg_num < 8)
+                               ? gp_reg_names_32[field.reg_num]
+                               : gp_reg_names[field.reg_num];
+        print_value(name, &value, sizeof(value), flags, out);
         break;
       }
       case TRACE_XMM_REG: {
@@ -155,7 +264,9 @@ static void print_regs_raw(Task* t, TraceFrame::Time event,
             }
             break;
         }
-        fwrite(&value, sizeof(value), 1, out);
+        char buf[8];
+        sprintf(buf, "xmm%d", field.reg_num);
+        print_value(buf, value, sizeof(value), flags, out);
         break;
       }
       case TRACE_YMM_REG: {
@@ -185,29 +296,29 @@ static void print_regs_raw(Task* t, TraceFrame::Time event,
             }
             break;
         }
-        fwrite(&value, sizeof(value), 1, out);
+        char buf[8];
+        sprintf(buf, "ymm%d", field.reg_num);
+        print_value(buf, value, sizeof(value), flags, out);
         break;
       }
     }
   }
+
+  fputc('\n', out);
 }
 
-struct RerunFlags {
-  TraceFrame::Time trace_start;
-  TraceFrame::Time trace_end;
-
-  vector<TraceField> singlestep_trace;
-
-  RerunFlags()
-      : trace_start(0), trace_end(numeric_limits<decltype(trace_end)>::max()) {}
-};
-
 static int find_gp_reg(const string& reg) {
-  static const char regs[16][4] = { "rax", "rcx", "rdx", "rbx", "rsp", "rbp",
-                                    "rsi", "rdi", "r8",  "r9",  "r10", "r11",
-                                    "r12", "r13", "r14", "r15" };
   for (int i = 0; i < 16; ++i) {
-    if (reg == regs[i]) {
+    if (reg == gp_reg_names[i] || (i < 8 && reg == gp_reg_names_32[i])) {
+      return i;
+    }
+  }
+  return -1;
+}
+
+static int find_seg_reg(const string& reg) {
+  for (int i = 0; i < 6; ++i) {
+    if (reg == seg_reg_names[i]) {
       return i;
     }
   }
@@ -236,8 +347,14 @@ static bool parse_regs(const string& value, vector<TraceField>* out) {
       out->push_back({ TRACE_INSTRUCTION_COUNT, 0 });
     } else if (reg == "ip" || reg == "rip") {
       out->push_back({ TRACE_IP, 0 });
+    } else if (reg == "fsbase") {
+      out->push_back({ TRACE_FSBASE, 0 });
+    } else if (reg == "gsbase") {
+      out->push_back({ TRACE_GSBASE, 0 });
     } else if (reg == "flags" || reg == "rflags") {
       out->push_back({ TRACE_FLAGS, 0 });
+    } else if (reg == "orig_rax" || reg == "orig_eax") {
+      out->push_back({ TRACE_ORIG_AX, 0 });
     } else if (reg == "gp_x16") {
       for (uint8_t i = 0; i < 16; ++i) {
         out->push_back({ TRACE_GP_REG, i });
@@ -252,6 +369,10 @@ static bool parse_regs(const string& value, vector<TraceField>* out) {
       }
     } else if (find_gp_reg(reg) >= 0) {
       out->push_back({ TRACE_GP_REG, (uint8_t)find_gp_reg(reg) });
+    } else if (find_seg_reg(reg) >= 0) {
+      out->push_back({ TRACE_SEG_REG, (uint8_t)find_seg_reg(reg) });
+    } else if (reg == "xinuse") {
+      out->push_back({ TRACE_XINUSE, 0 });
     } else {
       fprintf(stderr, "Unknown register '%s'\n", reg.c_str());
       return false;
@@ -266,8 +387,8 @@ static bool parse_rerun_arg(vector<string>& args, RerunFlags& flags) {
   }
 
   static const OptionSpec options[] = {
-    { 'e', "trace-end", NO_PARAMETER },
-    { 'r', "singlestep-registers", HAS_PARAMETER },
+    { 1, "singlestep", HAS_PARAMETER },    { 'e', "trace-end", HAS_PARAMETER },
+    { 'f', "function", HAS_PARAMETER },    { 'r', "raw", NO_PARAMETER },
     { 's', "trace-start", HAS_PARAMETER },
   };
   ParsedOption opt;
@@ -276,17 +397,29 @@ static bool parse_rerun_arg(vector<string>& args, RerunFlags& flags) {
   }
 
   switch (opt.short_name) {
+    case 1:
+      if (!parse_regs(opt.value, &flags.singlestep_trace)) {
+        return false;
+      }
+      break;
     case 'e':
       if (!opt.verify_valid_int(1, UINT32_MAX)) {
         return false;
       }
       flags.trace_end = opt.int_value;
       break;
-    case 'r':
-      if (!parse_regs(opt.value, &flags.singlestep_trace)) {
+    case 'f': {
+      char* endptr;
+      flags.function = strtoul(opt.value.c_str(), &endptr, 0);
+      if (*endptr) {
+        fprintf(stderr, "Invalid function address %s\n", opt.value.c_str());
         return false;
       }
       break;
+    }
+    case 'r':
+      flags.raw = true;
+      break;
     case 's':
       if (!opt.verify_valid_int(1, UINT32_MAX)) {
         return false;
@@ -294,7 +427,7 @@ static bool parse_rerun_arg(vector<string>& args, RerunFlags& flags) {
       flags.trace_start = opt.int_value;
       break;
     default:
-      assert(0 && "Unknown option");
+      DEBUG_ASSERT(0 && "Unknown option");
   }
   return true;
 }
@@ -302,7 +435,7 @@ static bool parse_rerun_arg(vector<string>& args, RerunFlags& flags) {
 static bool treat_event_completion_as_singlestep_complete(const Event& ev) {
   switch (ev.type()) {
     case EV_PATCH_SYSCALL:
-    case EV_SEGV_RDTSC:
+    case EV_INSTRUCTION_TRAP:
     case EV_SYSCALL:
       return true;
     default:
@@ -310,6 +443,11 @@ static bool treat_event_completion_as_singlestep_complete(const Event& ev) {
   }
 }
 
+/**
+ * Return true if the final "event" state change doesn't really change any
+ * user-visible state and is therefore not to be considered a singlestep for
+ * our purposes.
+ */
 static bool ignore_singlestep_for_event(const Event& ev) {
   switch (ev.type()) {
     // These don't actually change user-visible state, so we skip them.
@@ -343,22 +481,73 @@ static void clear_breakpoint_after_cpuid(Task* t) {
   t->vm()->remove_breakpoint(t->ip(), BKPT_USER);
 }
 
+static const uint64_t sentinel_ret_address = 9;
+
+static void run_diversion_function(ReplaySession& replay, Task* task,
+                                   const RerunFlags& flags) {
+  DiversionSession::shr_ptr diversion_session = replay.clone_diversion();
+  Task* t = diversion_session->find_task(task->tuid());
+  Registers regs = t->regs();
+  // align stack
+  auto sp = remote_ptr<uint64_t>(regs.sp().as_int() & ~uintptr_t(0xf)) - 1;
+  t->write_mem(sp.cast<uint64_t>(), sentinel_ret_address);
+  regs.set_sp(sp);
+  regs.set_ip(flags.function);
+  regs.set_di(0);
+  regs.set_si(0);
+  t->set_regs(regs);
+  RunCommand cmd =
+      flags.singlestep_trace.empty() ? RUN_CONTINUE : RUN_SINGLESTEP;
+
+  while (true) {
+    DiversionSession::DiversionResult result =
+        diversion_session->diversion_step(t, cmd);
+    print_regs(t, 0, 0, flags, stdout);
+    if (result.break_status.signal) {
+      if (result.break_status.signal->si_signo == SIGSEGV &&
+          result.break_status.signal->si_addr == (void*)sentinel_ret_address) {
+        return;
+      }
+      ASSERT(task, false) << "Unexpected signal "
+                          << *result.break_status.signal;
+    }
+  }
+}
+
 static int rerun(const string& trace_dir, const RerunFlags& flags) {
   ReplaySession::shr_ptr replay_session = ReplaySession::create(trace_dir);
   uint64_t instruction_count_within_event = 0;
+  bool done_first_step = false;
+
+  // Now that we've spawned the replay, raise our resource limits if
+  // possible.
+  raise_resource_limits();
 
   while (replay_session->trace_reader().time() < flags.trace_end) {
     RunCommand cmd = RUN_CONTINUE;
+
+    Task* old_task = replay_session->current_task();
+    remote_code_ptr old_ip = old_task ? old_task->ip() : remote_code_ptr();
+    FrameTime before_time = replay_session->trace_reader().time();
     if (replay_session->done_initial_exec() &&
-        !flags.singlestep_trace.empty() &&
-        replay_session->trace_reader().time() >= flags.trace_start) {
+        before_time >= flags.trace_start) {
+      if (!done_first_step) {
+        if (!flags.function.is_null()) {
+          run_diversion_function(*replay_session, old_task, flags);
+          return 0;
+        }
+
+        if (!flags.singlestep_trace.empty()) {
+          done_first_step = true;
+          print_regs(old_task, before_time - 1, instruction_count_within_event,
+                     flags, stdout);
+        }
+      }
+
       cmd = RUN_SINGLESTEP_FAST_FORWARD;
     }
 
-    TraceFrame::Time before_time = replay_session->trace_reader().time();
     Event replayed_event = replay_session->current_trace_frame().event();
-    Task* old_task = replay_session->current_task();
-    remote_code_ptr old_ip = old_task ? old_task->ip() : remote_code_ptr();
 
     bool set_breakpoint = maybe_set_breakpoint_after_cpuid(old_task);
     auto result = replay_session->replay_step(cmd);
@@ -372,39 +561,47 @@ static int rerun(const string& trace_dir, const RerunFlags& flags) {
       }
     }
 
-    TraceFrame::Time after_time = replay_session->trace_reader().time();
-    remote_code_ptr after_ip = old_task ? old_task->ip() : remote_code_ptr();
-    assert(after_time >= before_time && after_time <= before_time + 1);
-
     if (result.status == REPLAY_EXITED) {
       break;
     }
-    assert(result.status == REPLAY_CONTINUE);
-    assert(result.break_status.watchpoints_hit.empty());
-    assert(!result.break_status.breakpoint_hit);
-    assert(cmd == RUN_SINGLESTEP_FAST_FORWARD ||
-           !result.break_status.singlestep_complete);
 
-    // Treat singlesteps that partially executed a string instruction (that
-    // was not interrupted) as not really singlestepping.
-    bool singlestep_really_complete =
-        result.break_status.singlestep_complete &&
-        !ignore_singlestep_for_event(replayed_event) &&
-        (!result.did_fast_forward || old_ip != after_ip ||
-         before_time < after_time);
-    if (!flags.singlestep_trace.empty() && cmd == RUN_SINGLESTEP_FAST_FORWARD &&
-        (singlestep_really_complete ||
-         (before_time < after_time &&
-          treat_event_completion_as_singlestep_complete(replayed_event)))) {
-      print_regs_raw(old_task, before_time, instruction_count_within_event,
-                     flags.singlestep_trace, stdout);
-      fputc('\n', stdout);
-    }
+    FrameTime after_time = replay_session->trace_reader().time();
+    if (cmd != RUN_CONTINUE) {
+      remote_code_ptr after_ip = old_task ? old_task->ip() : remote_code_ptr();
+      DEBUG_ASSERT(after_time >= before_time && after_time <= before_time + 1);
 
-    if (singlestep_really_complete) {
-      instruction_count_within_event += 1;
+      DEBUG_ASSERT(result.status == REPLAY_CONTINUE);
+      DEBUG_ASSERT(result.break_status.watchpoints_hit.empty());
+      DEBUG_ASSERT(!result.break_status.breakpoint_hit);
+      DEBUG_ASSERT(cmd == RUN_SINGLESTEP_FAST_FORWARD ||
+                   !result.break_status.singlestep_complete);
+
+      // Treat singlesteps that partially executed a string instruction (that
+      // was not interrupted) as not really singlestepping.
+      bool singlestep_really_complete =
+          result.break_status.singlestep_complete &&
+          // ignore_singlestep_for_event only matters if we really completed the
+          // event
+          (!ignore_singlestep_for_event(replayed_event) ||
+           before_time == after_time) &&
+          (!result.did_fast_forward || old_ip != after_ip ||
+           before_time < after_time);
+      if (!flags.singlestep_trace.empty() &&
+          cmd == RUN_SINGLESTEP_FAST_FORWARD &&
+          (singlestep_really_complete ||
+           (before_time < after_time &&
+            treat_event_completion_as_singlestep_complete(replayed_event)))) {
+        print_regs(old_task, before_time, instruction_count_within_event, flags,
+                   stdout);
+      }
+
+      if (singlestep_really_complete) {
+        instruction_count_within_event += 1;
+      }
     }
     if (before_time < after_time) {
+      LOG(debug) << "Completed event " << before_time
+                 << " instruction_count=" << instruction_count_within_event;
       instruction_count_within_event = 1;
     }
   }
@@ -431,7 +628,6 @@ int RerunCommand::run(vector<string>& args) {
   }
 
   assert_prerequisites();
-  check_performance_settings();
 
   if (running_under_rr()) {
     if (!Flags::get().suppress_environment_warnings) {
