@@ -15,8 +15,10 @@
 
 #include "AutoRemoteSyscalls.h"
 #include "EmuFs.h"
+#include "Flags.h"
 #include "Task.h"
-#include "TaskGroup.h"
+#include "ThreadGroup.h"
+#include "core.h"
 #include "kernel_metadata.h"
 #include "log.h"
 #include "util.h"
@@ -26,20 +28,22 @@ using namespace std;
 namespace rr {
 
 struct Session::CloneCompletion {
-  struct TaskGroup {
+  struct AddressSpaceClone {
     Task* clone_leader;
     Task::CapturedState clone_leader_state;
     vector<Task::CapturedState> member_states;
     vector<pair<remote_ptr<void>, vector<uint8_t>>> captured_memory;
   };
-  vector<TaskGroup> task_groups;
+  vector<AddressSpaceClone> address_spaces;
 };
 
 Session::Session()
     : next_task_serial_(1),
       syscall_seccomp_ordering_(PTRACE_SYSCALL_BEFORE_SECCOMP_UNKNOWN),
       done_initial_exec_(false),
-      visible_execution_(true) {
+      visible_execution_(true),
+      has_cpuid_faulting_(!Flags::get().disable_cpuid_faulting &&
+                          cpuid_faulting_works()) {
   LOG(debug) << "Session " << this << " created";
 }
 
@@ -47,7 +51,7 @@ Session::~Session() {
   kill_all_tasks();
   LOG(debug) << "Session " << this << " destroyed";
 
-  for (auto tg : task_group_map) {
+  for (auto tg : thread_group_map) {
     tg.second->forget_session();
   }
 }
@@ -57,19 +61,27 @@ Session::Session(const Session& other) {
   next_task_serial_ = other.next_task_serial_;
   done_initial_exec_ = other.done_initial_exec_;
   visible_execution_ = other.visible_execution_;
+  has_cpuid_faulting_ = other.has_cpuid_faulting_;
 }
 
-void Session::on_create(TaskGroup* tg) { task_group_map[tg->tguid()] = tg; }
-void Session::on_destroy(TaskGroup* tg) { task_group_map.erase(tg->tguid()); }
+void Session::on_create(ThreadGroup* tg) { thread_group_map[tg->tguid()] = tg; }
+void Session::on_destroy(ThreadGroup* tg) {
+  thread_group_map.erase(tg->tguid());
+}
 
 void Session::post_exec() {
+  /* We just saw a successful exec(), so from now on we know
+   * that the address space layout for the replay tasks will
+   * (should!) be the same as for the recorded tasks.  So we can
+   * start validating registers at events. */
   assert_fully_initialized();
   if (done_initial_exec_) {
     return;
   }
   done_initial_exec_ = true;
-  assert(tasks().size() == 1);
-  tasks().begin()->second->flush_inconsistent_state();
+  DEBUG_ASSERT(tasks().size() == 1);
+  Task* t = tasks().begin()->second;
+  t->flush_inconsistent_state();
   spawned_task_error_fd_.close();
 }
 
@@ -99,25 +111,25 @@ AddressSpace::shr_ptr Session::clone(Task* t, AddressSpace::shr_ptr vm) {
   return as;
 }
 
-TaskGroup::shr_ptr Session::create_tg(Task* t) {
-  TaskGroup::shr_ptr tg(
-      new TaskGroup(this, nullptr, t->rec_tid, t->tid, t->tuid().serial()));
+ThreadGroup::shr_ptr Session::create_tg(Task* t) {
+  ThreadGroup::shr_ptr tg(
+      new ThreadGroup(this, nullptr, t->rec_tid, t->tid, t->tuid().serial()));
   tg->insert_task(t);
   return tg;
 }
 
-TaskGroup::shr_ptr Session::clone(Task* t, TaskGroup::shr_ptr tg) {
+ThreadGroup::shr_ptr Session::clone(Task* t, ThreadGroup::shr_ptr tg) {
   assert_fully_initialized();
   // If tg already belongs to our session this is a fork to create a new
   // taskgroup, otherwise it's a session-clone of an existing taskgroup
   if (this == tg->session()) {
-    return TaskGroup::shr_ptr(
-        new TaskGroup(this, tg.get(), t->rec_tid, t->tid, t->tuid().serial()));
+    return ThreadGroup::shr_ptr(new ThreadGroup(this, tg.get(), t->rec_tid,
+                                                t->tid, t->tuid().serial()));
   }
-  TaskGroup* parent =
-      tg->parent() ? find_task_group(tg->parent()->tguid()) : nullptr;
-  return TaskGroup::shr_ptr(
-      new TaskGroup(this, parent, tg->tgid, t->tid, tg->tguid().serial()));
+  ThreadGroup* parent =
+      tg->parent() ? find_thread_group(tg->parent()->tguid()) : nullptr;
+  return ThreadGroup::shr_ptr(
+      new ThreadGroup(this, parent, tg->tgid, t->tid, tg->tguid().serial()));
 }
 
 Task* Session::new_task(pid_t tid, pid_t rec_tid, uint32_t serial,
@@ -154,15 +166,25 @@ Task* Session::find_task(const TaskUid& tuid) const {
   return t && t->tuid() == tuid ? t : nullptr;
 }
 
-TaskGroup* Session::find_task_group(const TaskGroupUid& tguid) const {
+ThreadGroup* Session::find_thread_group(const ThreadGroupUid& tguid) const {
   finish_initializing();
-  auto it = task_group_map.find(tguid);
-  if (task_group_map.end() == it) {
+  auto it = thread_group_map.find(tguid);
+  if (thread_group_map.end() == it) {
     return nullptr;
   }
   return it->second;
 }
 
+ThreadGroup* Session::find_thread_group(pid_t pid) const {
+  finish_initializing();
+  for (auto& tg : thread_group_map) {
+    if (tg.first.tid() == pid) {
+      return tg.second;
+    }
+  }
+  return nullptr;
+}
+
 AddressSpace* Session::find_address_space(const AddressSpaceUid& vmuid) const {
   finish_initializing();
   auto it = vm_map.find(vmuid);
@@ -215,6 +237,7 @@ void Session::kill_all_tasks() {
     r.set_syscallno(syscall_number_for_exit(r.arch()));
     r.set_arg1(0);
     t->set_regs(r);
+    t->flush_regs();
     long result;
     do {
       // We have observed this failing with an ESRCH when the thread clearly
@@ -222,6 +245,10 @@ void Session::kill_all_tasks() {
       // work around it.
       result = t->fallible_ptrace(PTRACE_DETACH, nullptr, nullptr);
       ASSERT(t, result >= 0 || errno == ESRCH);
+      // But we it might get ESRCH because it really doesn't exist.
+      if (errno == ESRCH && is_zombie_process(t->tid)) {
+        break;
+      }
     } while (result < 0);
   }
 
@@ -236,7 +263,7 @@ void Session::kill_all_tasks() {
        */
       LOG(debug) << "sending SIGKILL to " << t->tid << " ...";
       // If we haven't already done a stable exit via syscall,
-      // kill the task and note that the entire task group is unstable.
+      // kill the task and note that the entire thread group is unstable.
       // The task may already have exited due to the preparation above,
       // so we might accidentally shoot down the wrong task :-(, but we
       // have to do this because the task might be in a state where it's not
@@ -244,7 +271,7 @@ void Session::kill_all_tasks() {
       // Linux doesn't seem to give us a reliable way to detach and kill
       // the tracee without races.
       syscall(SYS_tgkill, t->real_tgid(), t->tid, SIGKILL);
-      t->task_group()->destabilize();
+      t->thread_group()->destabilize();
     }
 
     t->destroy();
@@ -252,13 +279,13 @@ void Session::kill_all_tasks() {
 }
 
 void Session::on_destroy(AddressSpace* vm) {
-  assert(vm->task_set().size() == 0);
-  assert(vm_map.count(vm->uid()) == 1);
+  DEBUG_ASSERT(vm->task_set().size() == 0);
+  DEBUG_ASSERT(vm_map.count(vm->uid()) == 1);
   vm_map.erase(vm->uid());
 }
 
 void Session::on_destroy(Task* t) {
-  assert(task_map.count(t->rec_tid) == 1);
+  DEBUG_ASSERT(task_map.count(t->rec_tid) == 1);
   task_map.erase(t->rec_tid);
 }
 
@@ -358,7 +385,7 @@ void Session::check_for_watchpoint_changes(Task* t, BreakStatus& break_status) {
 }
 
 void Session::assert_fully_initialized() const {
-  assert(!clone_completion && "Session not fully initialized");
+  DEBUG_ASSERT(!clone_completion && "Session not fully initialized");
 }
 
 void Session::finish_initializing() const {
@@ -367,23 +394,24 @@ void Session::finish_initializing() const {
   }
 
   Session* self = const_cast<Session*>(this);
-  for (auto& tgleader : clone_completion->task_groups) {
-    AutoRemoteSyscalls remote(tgleader.clone_leader);
-    for (auto m : tgleader.clone_leader->vm()->maps()) {
-      // Creating this mapping was delayed in capture_state for performance
-      if (m.flags & AddressSpace::Mapping::IS_SYSCALLBUF) {
-        self->recreate_shared_mmap(remote, m);
+  for (auto& tgleader : clone_completion->address_spaces) {
+    {
+      AutoRemoteSyscalls remote(tgleader.clone_leader);
+      for (const auto& m : tgleader.clone_leader->vm()->maps()) {
+        // Creating this mapping was delayed in capture_state for performance
+        if (m.flags & AddressSpace::Mapping::IS_SYSCALLBUF) {
+          self->recreate_shared_mmap(remote, m);
+        }
+      }
+      for (auto& mem : tgleader.captured_memory) {
+        tgleader.clone_leader->write_bytes_helper(mem.first, mem.second.size(),
+                                                  mem.second.data());
+      }
+      for (auto& tgmember : tgleader.member_states) {
+        Task* t_clone = Task::os_clone_into(tgmember, remote);
+        self->on_create(t_clone);
+        t_clone->copy_state(tgmember);
       }
-    }
-    for (auto& mem : tgleader.captured_memory) {
-      tgleader.clone_leader->write_bytes_helper(mem.first, mem.second.size(),
-                                                mem.second.data());
-    }
-    for (auto& tgmember : tgleader.member_states) {
-      Task* t_clone =
-          Task::os_clone_into(tgmember, tgleader.clone_leader, remote);
-      self->on_create(t_clone);
-      t_clone->copy_state(tgmember);
     }
     tgleader.clone_leader->copy_state(tgleader.clone_leader_state);
   }
@@ -446,7 +474,8 @@ static void remap_shared_mmap(AutoRemoteSyscalls& remote, EmuFs& emu_fs,
   remote.infallible_syscall(syscall_number_for_close(remote.arch()), remote_fd);
 }
 
-#define RR_MAPPING_PREFIX "/tmp/rr-shared-"
+/*static*/ const char* Session::rr_mapping_prefix() { return "/rr-shared-"; }
+
 KernelMapping Session::create_shared_mmap(
     AutoRemoteSyscalls& remote, size_t size, remote_ptr<void> map_hint,
     const char* name, int tracee_prot, int tracee_flags,
@@ -454,8 +483,8 @@ KernelMapping Session::create_shared_mmap(
   static int nonce = 0;
   // Create the segment we'll share with the tracee.
   char path[PATH_MAX];
-  snprintf(path, sizeof(path) - 1, RR_MAPPING_PREFIX "%s-%d-%d", name,
-           remote.task()->real_tgid(), nonce++);
+  snprintf(path, sizeof(path) - 1, "%s%s%s-%d-%d", tmp_dir(),
+           rr_mapping_prefix(), name, remote.task()->real_tgid(), nonce++);
 
   // Let the child create the shmem block and then send the fd back to us.
   // This lets us avoid having to make the file world-writeable so that
@@ -506,35 +535,47 @@ KernelMapping Session::create_shared_mmap(
 
 static char* extract_name(char* name_buffer, size_t buffer_size) {
   // Recover the name that was originally chosen by finding the part of the
-  // name between RR_MAPPING_PREFIX and the -%d-%d at the end.
+  // name between rr_mapping_prefix and the -%d-%d at the end.
+  char* path_start = strstr(name_buffer, Session::rr_mapping_prefix());
+  DEBUG_ASSERT(path_start &&
+               "Passed something to create_shared_mmap that"
+               " wasn't a mapping shared between rr and the tracee?");
+  size_t prefix_len = path_start - name_buffer;
+  buffer_size -= prefix_len;
+  name_buffer += prefix_len;
+
   char* name_end = name_buffer + strnlen(name_buffer, buffer_size);
-  char* name_start = name_buffer + strlen(RR_MAPPING_PREFIX);
+  char* name_start = name_buffer + strlen(Session::rr_mapping_prefix());
   int hyphens_seen = 0;
   while (name_end > name_start) {
     --name_end;
     if (*name_end == '-') {
       ++hyphens_seen;
     } else if (*name_end == '/') {
-      assert(false && "Passed something to create_shared_mmap that"
-                      " wasn't a mapping shared between rr and the tracee?");
+      DEBUG_ASSERT(false &&
+                   "Passed something to create_shared_mmap that"
+                   " wasn't a mapping shared between rr and the tracee?");
     }
     if (hyphens_seen == 2) {
       break;
     }
   }
-  assert(hyphens_seen == 2);
+  DEBUG_ASSERT(hyphens_seen == 2);
   *name_end = '\0';
   return name_start;
 }
 
-// Recreate an mmap region that is shared between rr and the tracee. The caller
-// is responsible for recreating the data in the new mmap.
 const AddressSpace::Mapping& Session::recreate_shared_mmap(
     AutoRemoteSyscalls& remote, const AddressSpace::Mapping& m,
-    MonitoredSharedMemory::shr_ptr&& monitored) {
+    PreserveContents preserve, MonitoredSharedMemory::shr_ptr&& monitored) {
   char name[PATH_MAX];
   strncpy(name, m.map.fsname().c_str(), sizeof(name));
   uint32_t flags = m.flags;
+  size_t size = m.map.size();
+  void* preserved_data = preserve == PRESERVE_CONTENTS ? m.local_addr : nullptr;
+  if (preserved_data) {
+    remote.task()->vm()->detach_local_mapping(m.map.start());
+  }
   remote_ptr<void> new_addr =
       create_shared_mmap(remote, m.map.size(), m.map.start(),
                          extract_name(name, sizeof(name)), m.map.prot(), 0,
@@ -542,7 +583,12 @@ const AddressSpace::Mapping& Session::recreate_shared_mmap(
           .start();
   // m may be invalid now
   remote.task()->vm()->mapping_flags_of(new_addr) = flags;
-  return remote.task()->vm()->mapping_of(new_addr);
+  auto& new_map = remote.task()->vm()->mapping_of(new_addr);
+  if (preserved_data) {
+    memcpy(new_map.local_addr, preserved_data, size);
+    munmap(preserved_data, size);
+  }
+  return new_map;
 }
 
 const AddressSpace::Mapping& Session::steal_mapping(
@@ -625,7 +671,7 @@ static vector<uint8_t> capture_syscallbuf(const AddressSpace::Mapping& m,
 
 void Session::copy_state_to(Session& dest, EmuFs& emu_fs, EmuFs& dest_emu_fs) {
   assert_fully_initialized();
-  assert(!dest.clone_completion);
+  DEBUG_ASSERT(!dest.clone_completion);
 
   auto completion = unique_ptr<CloneCompletion>(new CloneCompletion());
 
@@ -636,8 +682,8 @@ void Session::copy_state_to(Session& dest, EmuFs& emu_fs, EmuFs& dest_emu_fs) {
     LOG(debug) << "  forking tg " << group_leader->tgid()
                << " (real: " << group_leader->real_tgid() << ")";
 
-    completion->task_groups.push_back(CloneCompletion::TaskGroup());
-    auto& group = completion->task_groups.back();
+    completion->address_spaces.push_back(CloneCompletion::AddressSpaceClone());
+    auto& group = completion->address_spaces.back();
 
     group.clone_leader = group_leader->os_fork_into(&dest);
     dest.on_create(group.clone_leader);
@@ -645,7 +691,8 @@ void Session::copy_state_to(Session& dest, EmuFs& emu_fs, EmuFs& dest_emu_fs) {
 
     {
       AutoRemoteSyscalls remote(group.clone_leader);
-      for (auto m : group.clone_leader->vm()->maps()) {
+      vector<AddressSpace::Mapping> shared_maps_to_clone;
+      for (const auto& m : group.clone_leader->vm()->maps()) {
         // Special case the syscallbuf as a performance optimization. The amount
         // of data we need to capture is usually significantly smaller than the
         // size of the mapping, so allocating the whole mapping here would be
@@ -654,15 +701,19 @@ void Session::copy_state_to(Session& dest, EmuFs& emu_fs, EmuFs& dest_emu_fs) {
           group.captured_memory.push_back(make_pair(
               m.map.start(), capture_syscallbuf(m, group.clone_leader)));
         } else if (m.local_addr != nullptr) {
-          memcpy(recreate_shared_mmap(remote, m).local_addr, m.local_addr,
-                 m.map.size());
+          ASSERT(group.clone_leader,
+                 m.map.start() == AddressSpace::preload_thread_locals_start());
         } else if ((m.recorded_map.flags() & MAP_SHARED) &&
                    emu_fs.has_file_for(m.recorded_map)) {
-          remap_shared_mmap(remote, emu_fs, dest_emu_fs, m);
+          shared_maps_to_clone.push_back(m);
         }
       }
+      // Do this in a separate loop to avoid iteration invalidation issues
+      for (const auto& m : shared_maps_to_clone) {
+        remap_shared_mmap(remote, emu_fs, dest_emu_fs, m);
+      }
 
-      for (auto t : group_leader->task_group()->task_set()) {
+      for (auto t : vm.second->task_set()) {
         if (group_leader == t) {
           continue;
         }
@@ -676,7 +727,7 @@ void Session::copy_state_to(Session& dest, EmuFs& emu_fs, EmuFs& dest_emu_fs) {
   }
   dest.clone_completion = move(completion);
 
-  assert(dest.vms().size() > 0);
+  DEBUG_ASSERT(dest.vms().size() > 0);
 }
 
 } // namespace rr
