@@ -6,6 +6,8 @@
 #include <string>
 #include <vector>
 
+namespace rr {
+
 /**
  * Converts a vector of strings to a POSIX-style array of char*s terminated
  * by a nullptr.
@@ -24,4 +26,6 @@ private:
   std::vector<char*> array;
 };
 
+} // namespace rr
+
 #endif
