@@ -2,6 +2,7 @@
 
 #include "fast_forward.h"
 
+#include "core.h"
 #include "log.h"
 
 using namespace std;
@@ -105,8 +106,8 @@ static bool decode_x86_string_instruction(const InstructionBuf& code,
 
 static bool mem_intersect(remote_ptr<void> a1, int s1, remote_ptr<void> a2,
                           int s2) {
-  assert(a1 + s1 > a1);
-  assert(a2 + s2 > a2);
+  DEBUG_ASSERT(a1 + s1 > a1);
+  DEBUG_ASSERT(a2 + s2 > a2);
   return max(a1, a2) < min(a1 + s1, a2 + s2);
 }
 
@@ -114,6 +115,11 @@ static void bound_iterations_for_watchpoint(Task* t, remote_ptr<void> reg,
                                             const DecodedInstruction& decoded,
                                             const WatchConfig& watch,
                                             uintptr_t* iterations) {
+  if (watch.num_bytes == 0) {
+    // Ignore zero-sized watch. It can't ever trigger.
+    return;
+  }
+
   // Compute how many iterations it will take before we hit the watchpoint.
   // 0 means the first iteration will hit the watchpoint.
   int size = decoded.operand_size;
@@ -151,7 +157,7 @@ static bool is_x86ish(Task* t) {
 
 bool fast_forward_through_instruction(Task* t, ResumeRequest how,
                                       const vector<const Registers*>& states) {
-  assert(how == RESUME_SINGLESTEP || how == RESUME_SYSEMU_SINGLESTEP);
+  DEBUG_ASSERT(how == RESUME_SINGLESTEP || how == RESUME_SYSEMU_SINGLESTEP);
 
   remote_code_ptr ip = t->ip();
 
@@ -325,8 +331,9 @@ bool fast_forward_through_instruction(Task* t, ResumeRequest how,
       // instructions that don't modify flags should not terminate too early.
       // We can terminate prematurely when the watchpoint we set relative
       // to DI is triggered by a read via SI.
-      ASSERT(t, decoded.modifies_flags || iterations <= BYTES_COALESCED ||
-                    triggered_watchpoint);
+      ASSERT(t,
+             decoded.modifies_flags || iterations <= BYTES_COALESCED ||
+                 triggered_watchpoint);
 
       if (!triggered_watchpoint) {
         // watchpoint didn't fire. We must have exited the loop early and
