@@ -3,11 +3,11 @@
 #ifndef RR_KERNEL_ABI_H
 #define RR_KERNEL_ABI_H
 
-#include <assert.h>
 #include <signal.h>
 
 #include <vector>
 
+#include "core.h"
 #include "remote_ptr.h"
 
 namespace rr {
@@ -322,7 +322,7 @@ struct BaseArch : public wordsize,
     template <typename U> ptr<T>& operator=(remote_ptr<U> p) {
       remote_ptr<T> pt = p;
       val = pt.as_int();
-      assert(val == pt.as_int());
+      DEBUG_ASSERT(val == pt.as_int());
       return *this;
     }
     operator bool() const { return val; }
@@ -573,7 +573,8 @@ struct BaseArch : public wordsize,
   };
   RR_VERIFY_TYPE(msginfo);
 
-  struct shmid64_ds {
+  /* Don't align for the 64-bit values on 32-bit x86 */
+  struct __attribute__((packed)) shmid64_ds {
     ipc64_perm shm_perm;
     size_t shm_segsz;
     uint64_t shm_atime_only_little_endian;
@@ -917,6 +918,12 @@ struct BaseArch : public wordsize,
     signed_long optlen;
   };
 
+  struct connect_args {
+    signed_long sockfd;
+    ptr<void> addr;
+    socklen_t addrlen;
+  };
+
   struct recv_args {
     signed_int sockfd;
     char __pad[sizeof(ptr<void>) - sizeof(int)];
@@ -1036,7 +1043,7 @@ struct BaseArch : public wordsize,
     ptr<size_t> oldlenp;
     ptr<void> newval;
     ptr<size_t> newlen;
-    unsigned_long __unused[4];
+    unsigned_long __rr_unused[4];
   };
   RR_VERIFY_TYPE(__sysctl_args);
 
@@ -1367,6 +1374,98 @@ struct BaseArch : public wordsize,
     uint32_t permitted;
     uint32_t inheritable;
   };
+
+  struct hci_dev_req {
+    uint16_t dev_id;
+    uint32_t dev_opt;
+  };
+
+  struct hci_dev_list_req {
+    uint16_t dev_num;
+    struct hci_dev_req dev_req[0];
+  };
+
+  typedef struct { uint8_t b[6]; } __attribute__((__packed__)) bdaddr_t;
+
+  struct hci_dev_stats {
+    uint32_t err_rx;
+    uint32_t err_tx;
+    uint32_t cmd_tx;
+    uint32_t evt_rx;
+    uint32_t acl_tx;
+    uint32_t acl_rx;
+    uint32_t sco_tx;
+    uint32_t sco_rx;
+    uint32_t byte_rx;
+    uint32_t byte_tx;
+  };
+
+  struct hci_dev_info {
+    uint16_t dev_id;
+    char name[8];
+
+    bdaddr_t bdaddr;
+
+    uint32_t flags;
+    uint8_t type;
+
+    uint8_t features[8];
+
+    uint32_t pkt_type;
+    uint32_t link_policy;
+    uint32_t link_mode;
+
+    uint16_t acl_mtu;
+    uint16_t acl_pkts;
+    uint16_t sco_mtu;
+    uint16_t sco_pkts;
+
+    struct hci_dev_stats stat;
+  };
+
+  typedef struct ifbond {
+    int32_t bond_mode;
+    int32_t num_slaves;
+    int32_t miimon;
+  } ifbond;
+  RR_VERIFY_TYPE(ifbond);
+
+  typedef struct timex {
+    unsigned int modes;
+    __kernel_long_t offset;
+    __kernel_long_t freq;
+    __kernel_long_t maxerror;
+    __kernel_long_t esterror;
+    int status;
+    __kernel_long_t constant;
+    __kernel_long_t precision;
+    __kernel_long_t tolerance;
+    timeval time;
+    __kernel_long_t tick;
+    __kernel_long_t ppsfreq;
+    __kernel_long_t jitter;
+    int shift;
+    __kernel_long_t stabil;
+    __kernel_long_t jitcnt;
+    __kernel_long_t calcnt;
+    __kernel_long_t errcnt;
+    __kernel_long_t stbcnt;
+    int tai;
+
+    // Further padding bytes to allow for future expansion.
+    int : 32;
+    int : 32;
+    int : 32;
+    int : 32;
+    int : 32;
+    int : 32;
+    int : 32;
+    int : 32;
+    int : 32;
+    int : 32;
+    int : 32;
+  } timex;
+  RR_VERIFY_TYPE(timex);
 };
 
 struct X86Arch : public BaseArch<SupportedArch::x86, WordSize32Defs> {
@@ -1472,7 +1571,7 @@ struct X86Arch : public BaseArch<SupportedArch::x86, WordSize32Defs> {
   };
   RR_VERIFY_TYPE_ARCH(SupportedArch::x86, ::sigcontext, sigcontext);
 
-  struct ucontext {
+  struct ucontext_t {
     uint32_t uc_flags;
     uint32_t uc_link;
     stack_t uc_stack;
@@ -1486,7 +1585,7 @@ struct X86Arch : public BaseArch<SupportedArch::x86, WordSize32Defs> {
     uint32_t pinfo;
     uint32_t puc;
     siginfo_t info;
-    struct ucontext uc;
+    struct ucontext_t uc;
   };
 
   struct _fpstate_32 {
@@ -1548,7 +1647,7 @@ struct X86Arch : public BaseArch<SupportedArch::x86, WordSize32Defs> {
   };
   RR_VERIFY_TYPE_ARCH(SupportedArch::x86, struct ::stat, struct stat);
 
-  struct stat64 {
+  struct __attribute__((packed)) stat64 {
     dev_t st_dev;
     unsigned_int __pad1;
     ino_t __st_ino;
@@ -1585,12 +1684,6 @@ struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
 
 #include "SyscallEnumsX64.generated"
 
-  // The kernel defines the segment registers and eflags as 64-bit quantities,
-  // even though the segment registers are really 16-bit and eflags is
-  // architecturally defined as 32-bit.  GDB wants the segment registers and
-  // eflags to appear as 32-bit quantities.  From the perspective of providing
-  // registers to GDB, it's easier if we define these registers as uint32_t
-  // with extra padding.
   struct user_regs_struct {
     uint64_t r15;
     uint64_t r14;
@@ -1611,25 +1704,18 @@ struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
     // signed in practice.
     uint64_t orig_rax;
     uint64_t rip;
-    uint32_t cs;
-    uint32_t cs_upper;
-    uint32_t eflags;
-    uint32_t eflags_upper;
+    uint64_t cs;
+    uint64_t eflags;
     uint64_t rsp;
-    uint32_t ss;
-    uint32_t ss_upper;
+    uint64_t ss;
     // These _base registers are architecturally defined MSRs and really do
     // need to be 64-bit.
     uint64_t fs_base;
     uint64_t gs_base;
-    uint32_t ds;
-    uint32_t ds_upper;
-    uint32_t es;
-    uint32_t es_upper;
-    uint32_t fs;
-    uint32_t fs_upper;
-    uint32_t gs;
-    uint32_t gs_upper;
+    uint64_t ds;
+    uint64_t es;
+    uint64_t fs;
+    uint64_t gs;
   };
   RR_VERIFY_TYPE_ARCH(SupportedArch::x86_64, ::user_regs_struct,
                       user_regs_struct);
@@ -1682,19 +1768,19 @@ struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
   RR_VERIFY_TYPE_ARCH(SupportedArch::x86_64, ::user_fpregs_struct,
                       user_fpregs_struct);
 
-  struct ucontext {
+  struct ucontext_t {
     uint64_t ucflags;
-    ptr<struct ucontext> uc_link;
+    ptr<struct ucontext_t> uc_link;
     stack_t uc_stack;
     struct sigcontext uc_mcontext;
     sigset_t uc_sigmask;
     user_fpregs_struct uc_fpregs;
   };
-  RR_VERIFY_TYPE_ARCH(SupportedArch::x86_64, ::ucontext, ucontext);
+  RR_VERIFY_TYPE_ARCH(SupportedArch::x86_64, ::ucontext_t, ucontext_t);
 
   struct rt_sigframe {
     ptr<char> pretcode;
-    struct ucontext uc;
+    struct ucontext_t uc;
     siginfo_t info;
   };
 
@@ -1738,7 +1824,7 @@ struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
     struct timespec st_atim;
     struct timespec st_mtim;
     struct timespec st_ctim;
-    syscall_slong_t __unused[3];
+    syscall_slong_t __rr_unused[3];
   };
   RR_VERIFY_TYPE_ARCH(SupportedArch::x86_64, struct ::stat, struct stat);
 
@@ -1757,7 +1843,7 @@ struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
     struct timespec st_atim;
     struct timespec st_mtim;
     struct timespec st_ctim;
-    syscall_slong_t __unused[3];
+    syscall_slong_t __rr_unused[3];
   };
   RR_VERIFY_TYPE_ARCH(SupportedArch::x86_64, struct ::stat64, struct stat64);
 };
@@ -1765,7 +1851,8 @@ struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
 #define RR_ARCH_FUNCTION(f, arch, args...)                                     \
   switch (arch) {                                                              \
     default:                                                                   \
-      assert(0 && "Unknown architecture");                                     \
+      DEBUG_ASSERT(0 && "Unknown architecture");                               \
+      RR_FALLTHROUGH;                                                          \
     case x86:                                                                  \
       return f<rr::X86Arch>(args);                                             \
     case x86_64:                                                               \
@@ -1811,9 +1898,4 @@ typedef X64Arch NativeArch;
 
 } // namespace rr
 
-// New in the 4.6 kernel.
-#ifndef CLONE_NEWCGROUP
-#define CLONE_NEWCGROUP 0x02000000
-#endif
-
 #endif /* RR_KERNEL_ABI_H */
