@@ -3,6 +3,7 @@
 #ifndef RR_KERNEL_SUPPLEMENT_H_
 #define RR_KERNEL_SUPPLEMENT_H_
 
+#include <linux/if_tun.h>
 #include <linux/mman.h>
 #include <linux/seccomp.h>
 #include <linux/usbdevice_fs.h>
@@ -48,6 +49,14 @@ namespace rr {
 #define PTRACE_SYSEMU_SINGLESTEP 32
 #endif
 
+#ifndef PTRACE_GETSIGMASK
+#define PTRACE_GETSIGMASK 0x420a
+#endif
+
+#ifndef PTRACE_SETSIGMASK
+#define PTRACE_SETSIGMASK 0x420b
+#endif
+
 #ifndef PTRACE_O_TRACESECCOMP
 #define PTRACE_O_TRACESECCOMP 0x00000080
 #define PTRACE_EVENT_SECCOMP_OBSOLETE 8 // ubuntu 12.04
@@ -133,6 +142,19 @@ struct usbdevfs_streams {
 #define USBDEVFS_FREE_STREAMS _IOR('U', 29, struct usbdevfs_streams)
 #endif
 
+#ifndef TUNSETVNETLE
+#define TUNSETVNETLE _IOW('T', 220, int)
+#endif
+#ifndef TUNGETVNETLE
+#define TUNGETVNETLE _IOR('T', 221, int)
+#endif
+#ifndef TUNSETVNETBE
+#define TUNSETVNETBE _IOW('T', 222, int)
+#endif
+#ifndef TUNGETVNETBE
+#define TUNGETVNETBE _IOR('T', 223, int)
+#endif
+
 #ifndef TIOCGPKT
 #define TIOCGPKT _IOR('T', 0x38, int)
 #endif
@@ -143,6 +165,16 @@ struct usbdevfs_streams {
 #define TIOCGEXCL _IOR('T', 0x40, int)
 #endif
 
+struct rr_input_mask {
+  uint32_t type;
+  uint32_t codes_size;
+  uint64_t codes_ptr;
+};
+
+#ifndef EVIOCGMASK
+#define EVIOCGMASK _IOR('E', 0x92, struct rr_input_mask)
+#endif
+
 #ifndef MADV_FREE
 #define MADV_FREE 8
 #endif
@@ -170,6 +202,13 @@ struct usbdevfs_streams {
 #define SO_SET_REPLACE 64
 #endif
 
+#ifndef HCIGETDEVLIST
+#define HCIGETDEVLIST _IOR('H', 210, int)
+#endif
+#ifndef HCIGETDEVINFO
+#define HCIGETDEVINFO _IOR('H', 211, int)
+#endif
+
 // Unfortuantely the header that defines these is not C++ safe, we we'll
 // have to redefine them here
 #ifndef KEYCTL_GET_KEYRING_ID
@@ -252,6 +291,18 @@ struct usbdevfs_streams {
 #define PR_CAP_AMBIENT_CLEAR_ALL 4
 #endif
 
+// New in the 4.6 kernel.
+#ifndef CLONE_NEWCGROUP
+#define CLONE_NEWCGROUP 0x02000000
+#endif
+// New in the 4.12 kernel
+#ifndef ARCH_GET_CPUID
+#define ARCH_GET_CPUID 0x1011
+#endif
+#ifndef ARCH_SET_CPUID
+#define ARCH_SET_CPUID 0x1012
+#endif
+
 } // namespace rr
 
 #endif /* RR_KERNEL_SUPPLEMENT_H_ */
