@@ -4,13 +4,17 @@
 
 #include <stdlib.h>
 
+#include <deque>
+#include <fstream>
 #include <memory>
 #include <sstream>
 #include <unordered_map>
 
 #include "Flags.h"
+#include "GdbConnection.h"
 #include "GdbServer.h"
 #include "RecordSession.h"
+#include "core.h"
 #include "ftrace.h"
 #include "kernel_metadata.h"
 #include "util.h"
@@ -73,13 +77,18 @@ static bool log_globals_initialized = false;
 static LogLevel default_level = LOG_error;
 
 // These need to be available to other static constructors, so we need to be
-// sure that they const constant-initialized. Unfortunately some versions of C++
-// libraries have a bug that causes them not to be. _CONSTANT_STATIC should
+// sure that they can be constant-initialized. Unfortunately some versions of
+// C++ libraries have a bug that causes them not to be. _CONSTANT_STATIC should
 // turn this into a compile error rather than a runtime crash for compilers
 // that support the attribute.
 _CONSTANT_STATIC unique_ptr<unordered_map<string, LogLevel>> level_map;
 _CONSTANT_STATIC unique_ptr<unordered_map<const char*, LogModule>> log_modules;
 _CONSTANT_STATIC std::unique_ptr<stringstream> logging_stream;
+_CONSTANT_STATIC deque<char>* log_buffer;
+_CONSTANT_STATIC std::ostream* log_file;
+size_t log_buffer_size;
+
+static void flush_log_file() { log_file->flush(); }
 
 static void init_log_globals() {
   if (log_globals_initialized) {
@@ -92,6 +101,30 @@ static void init_log_globals() {
       new unordered_map<const char*, LogModule>());
   logging_stream = unique_ptr<stringstream>(new stringstream());
 
+  const char* buffer = getenv("RR_LOG_BUFFER");
+  if (buffer) {
+    log_buffer_size = atoi(buffer);
+    if (log_buffer_size) {
+      log_buffer = new deque<char>();
+      atexit(flush_log_buffer);
+    }
+  }
+
+  const char* filename = getenv("RR_LOG_FILE");
+  if (filename) {
+    auto file = new ofstream(filename);
+    if (!file->good()) {
+      delete file;
+    } else {
+      log_file = file;
+      atexit(flush_log_file);
+    }
+  }
+
+  if (!log_file) {
+    log_file = &cerr;
+  }
+
   const char* log_env = "RR_LOG";
   if (running_under_rr()) {
     log_env = "RR_UNDER_RR_LOG";
@@ -99,7 +132,7 @@ static void init_log_globals() {
   char* env = getenv(log_env);
   if (env) {
     env = strdup(env);
-    assert(env);
+    DEBUG_ASSERT(env);
     for (int i = 0; env[i]; ++i) {
       env[i] = simple_to_lower(env[i]);
     }
@@ -209,11 +242,40 @@ ostream& log_stream() {
 }
 
 static void flush_log_stream() {
-  cerr << logging_stream->str();
-  ftrace::write(logging_stream->str());
+  string s = logging_stream->str();
+  ftrace::write(s);
+  if (log_buffer) {
+    size_t len = s.size();
+    if (len >= log_buffer_size) {
+      log_buffer->clear();
+      log_buffer->insert(log_buffer->end(), s.c_str() + (len - log_buffer_size),
+                         s.c_str() + len);
+    } else {
+      if (log_buffer->size() + len > log_buffer_size) {
+        log_buffer->erase(log_buffer->begin(),
+                          log_buffer->begin() +
+                              (log_buffer->size() + len - log_buffer_size));
+      }
+      log_buffer->insert(log_buffer->end(), s.c_str(), s.c_str() + len);
+    }
+  } else {
+    *log_file << s;
+  }
+
   logging_stream->str(string());
 }
 
+void flush_log_buffer() {
+  if (log_buffer) {
+    for (char c : *log_buffer) {
+      // We could accumulate in a string to speed things up, but this could get
+      // called in low-memory situations so be safe.
+      *log_file << c;
+    }
+    log_buffer->clear();
+  }
+}
+
 template <typename T>
 static void write_prefix(T& stream, LogLevel level, const char* file, int line,
                          const char* function) {
@@ -223,12 +285,17 @@ static void write_prefix(T& stream, LogLevel level, const char* file, int line,
     stream << file << ":" << line << ":";
   }
   stream << function << "()";
-  if (level <= LOG_warn) {
+  if (level <= LOG_warn && err) {
     stream << " errno: " << errno_name(err);
   }
   stream << "] ";
 }
 
+bool is_logging_enabled(LogLevel level, const char* file) {
+  LogModule& m = get_log_module(file);
+  return level <= m.level;
+}
+
 NewlineTerminatingOstream::NewlineTerminatingOstream(LogLevel level,
                                                      const char* file, int line,
                                                      const char* function)
@@ -249,11 +316,23 @@ NewlineTerminatingOstream::~NewlineTerminatingOstream() {
     log_stream() << std::endl;
     flush_log_stream();
     if (Flags::get().fatal_errors_and_warnings && level <= LOG_warn) {
-      abort();
+      notifying_abort();
     }
   }
 }
 
+CleanFatalOstream::CleanFatalOstream(const char* file, int line,
+                                     const char* function) {
+  errno = 0;
+  write_prefix(*this, LOG_fatal, file, line, function);
+}
+
+CleanFatalOstream::~CleanFatalOstream() {
+  cerr << std::endl;
+  flush_log_stream();
+  exit(1);
+}
+
 FatalOstream::FatalOstream(const char* file, int line, const char* function) {
   write_prefix(*this, LOG_fatal, file, line, function);
 }
@@ -261,7 +340,7 @@ FatalOstream::FatalOstream(const char* file, int line, const char* function) {
 FatalOstream::~FatalOstream() {
   log_stream() << std::endl;
   flush_log_stream();
-  abort();
+  notifying_abort();
 }
 
 static void emergency_debug(Task* t) {
@@ -279,12 +358,15 @@ static void emergency_debug(Task* t) {
     record_session->trace_writer().close();
   }
 
-  if (probably_not_interactive() && !Flags::get().force_things) {
+  if (probably_not_interactive() && !Flags::get().force_things &&
+      !getenv("RUNNING_UNDER_TEST_MONITOR")) {
     errno = 0;
     FATAL()
         << "(session doesn't look interactive, aborting emergency debugging)";
   }
 
+  flush_log_buffer();
+
   GdbServer::emergency_debug(t);
   FATAL() << "Can't resume execution from invalid state";
 }
