@@ -5,6 +5,7 @@
 #include <limits.h>
 #include <linux/version.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <sys/utsname.h>
 
@@ -13,7 +14,10 @@
 #include "Command.h"
 #include "Flags.h"
 #include "RecordCommand.h"
+#include "ReplayCommand.h"
+#include "core.h"
 #include "log.h"
+#include "util.h"
 
 using namespace std;
 
@@ -43,66 +47,12 @@ void assert_prerequisites(bool use_syscall_buffer) {
   }
 }
 
-void check_performance_settings() {
-  if (Flags::get().suppress_environment_warnings) {
-    return;
-  }
-
-  // NB: we hard-code "cpu0" here because rr pins itself and all
-  // tracees to cpu 0.  We don't care about the other CPUs.
-  ScopedFd fd("/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor",
-              O_RDONLY);
-  if (0 > fd) {
-    // If the file doesn't exist, the system probably
-    // doesn't have the ability to frequency-scale, for
-    // example a VM.
-    LOG(info) << "Unable to check CPU-frequency governor.";
-    return;
-  }
-  char governor[PATH_MAX];
-  ssize_t nread = read(fd, governor, sizeof(governor) - 1);
-  if (0 > nread) {
-    FATAL() << "Unable to read cpu0's frequency governor.";
-  }
-  governor[nread] = '\0';
-  ssize_t len = strlen(governor);
-  if (len > 0) {
-    // Eat the '\n'.
-    governor[len - 1] = '\0';
-  }
-  LOG(info) << "cpu0's frequency governor is '" << governor << "'";
-  if (strcmp("performance", governor)) {
-    fprintf(stderr,
-            "\n"
-            "rr: Warning: Your CPU frequency governor is '%s'.  rr strongly\n"
-            "    recommends that you use the 'performance' governor.  Not "
-            "using the\n"
-            "    'performance' governor can cause rr to be at least 2x slower\n"
-            "    on laptops.\n"
-            "\n"
-            "    On Fedora-based systems, you can enable the 'performance' "
-            "governor\n"
-            "    by running the following commands:\n"
-            "\n"
-            "    $ sudo dnf install kernel-tools\n"
-            "    $ sudo cpupower frequency-set -g performance\n"
-            "\n",
-            governor);
-    // TODO: It would be nice to bail here or do something
-    // clever to enable the 'performance' just for rr, but
-    // that seems too hard at the moment.
-  }
-}
-
 void print_version(FILE* out) { fprintf(out, "rr version %s\n", RR_VERSION); }
 
-void print_usage(FILE* out) {
-  print_version(out);
-  fputs("Usage:\n", out);
-  Command::print_help_all(out);
+void print_global_options(FILE* out) {
   fputs(
-      "\n"
-      "Common options:\n"
+      "Global options:\n"
+      "  --disable-cpuid-faulting   disable use of CPUID faulting\n"
       "  -A, --microarch=<NAME>     force rr to assume it's running on a CPU\n"
       "                             with microarch NAME even if runtime "
       "detection\n"
@@ -145,13 +95,22 @@ void print_usage(FILE* out) {
       "                             suppress warnings about issues in the\n"
       "                             environment that rr has no control over\n"
       "  -T, --dump-at=TIME         dump memory at global timepoint TIME\n"
-      "  -V, --verbose              log messages that may not be urgently \n"
-      "                             critical to the user\n"
       "\n"
       "Use RR_LOG to control logging; e.g. RR_LOG=all:warn,Task:debug\n",
       out);
 }
 
+void print_usage(FILE* out) {
+  print_version(out);
+  fputs("\nUsage:\n", out);
+  Command::print_help_all(out);
+  fputs("\nIf no subcommand is provided, we check if the first non-option\n"
+        "argument is a directory. If it is, we assume the 'replay' subcommand\n"
+        "otherwise we assume the 'record' subcommand.\n\n",
+        out);
+  print_global_options(out);
+}
+
 static void init_random() {
   // Not very good, but good enough for our non-security-sensitive needs.
   int key = time(nullptr) ^ getpid();
@@ -161,18 +120,18 @@ static void init_random() {
 
 bool parse_global_option(std::vector<std::string>& args) {
   static const OptionSpec options[] = {
-    { 'C', "checksum", HAS_PARAMETER },
-    { 'K', "check-cached-mmaps", NO_PARAMETER },
-    { 'U', "cpu-unbound", NO_PARAMETER },
-    { 'T', "dump-at", HAS_PARAMETER },
-    { 'D', "dump-on", HAS_PARAMETER },
-    { 'F', "force-things", NO_PARAMETER },
+    { 0, "disable-cpuid-faulting", NO_PARAMETER },
     { 'A', "microarch", HAS_PARAMETER },
-    { 'M', "mark-stdio", NO_PARAMETER },
-    { 'S', "suppress-environment-warnings", NO_PARAMETER },
+    { 'C', "checksum", HAS_PARAMETER },
+    { 'D', "dump-on", HAS_PARAMETER },
     { 'E', "fatal-errors", NO_PARAMETER },
-    { 'V', "verbose", NO_PARAMETER },
-    { 'N', "version", NO_PARAMETER }
+    { 'F', "force-things", NO_PARAMETER },
+    { 'K', "check-cached-mmaps", NO_PARAMETER },
+    { 'M', "mark-stdio", NO_PARAMETER },
+    { 'N', "version", NO_PARAMETER },
+    { 'S', "suppress-environment-warnings", NO_PARAMETER },
+    { 'T', "dump-at", HAS_PARAMETER },
+    { 'U', "cpu-unbound", NO_PARAMETER },
   };
 
   ParsedOption opt;
@@ -182,6 +141,9 @@ bool parse_global_option(std::vector<std::string>& args) {
 
   Flags& flags = Flags::get_for_init();
   switch (opt.short_name) {
+    case 0:
+      flags.disable_cpuid_faulting = true;
+      break;
     case 'A':
       flags.forced_uarch = opt.value;
       break;
@@ -193,7 +155,7 @@ bool parse_global_option(std::vector<std::string>& args) {
         LOG(info) << "checksumming on all events";
         flags.checksum = Flags::CHECKSUM_ALL;
       } else {
-        flags.checksum = atoi(opt.value.c_str());
+        flags.checksum = strtoll(opt.value.c_str(), NULL, 10);
         LOG(info) << "checksumming on at event " << flags.checksum;
       }
       break;
@@ -201,7 +163,7 @@ bool parse_global_option(std::vector<std::string>& args) {
       if (opt.value == "RDTSC") {
         flags.dump_on = Flags::DUMP_ON_RDTSC;
       } else {
-        flags.dump_on = atoi(opt.value.c_str());
+        flags.dump_on = strtoll(opt.value.c_str(), NULL, 10);
       }
       break;
     case 'E':
@@ -220,13 +182,13 @@ bool parse_global_option(std::vector<std::string>& args) {
       flags.suppress_environment_warnings = true;
       break;
     case 'T':
-      flags.dump_at = atoi(opt.value.c_str());
+      flags.dump_at = strtoll(opt.value.c_str(), NULL, 10);
       break;
     case 'N':
       show_version = true;
       break;
     default:
-      assert(0 && "Invalid flag");
+      DEBUG_ASSERT(0 && "Invalid flag");
   }
   return true;
 }
@@ -237,6 +199,7 @@ using namespace rr;
 
 int main(int argc, char* argv[]) {
   init_random();
+  raise_resource_limits();
 
   vector<string> args;
   for (int i = 1; i < argc; ++i) {
@@ -264,7 +227,11 @@ int main(int argc, char* argv[]) {
       print_usage(stderr);
       return 1;
     }
-    command = RecordCommand::get();
+    if (is_directory(args[0].c_str())) {
+      command = ReplayCommand::get();
+    } else {
+      command = RecordCommand::get();
+    }
   }
 
   return command->run(args);
