@@ -3,15 +3,82 @@
 #ifndef RR_PRELOAD_INTERFACE_H_
 #define RR_PRELOAD_INTERFACE_H_
 
+/* Bump this whenever the interface between syscallbuf and rr changes in a way
+ * that would require changes to replay. So be very careful making changes to
+ * this file! Many changes would require a bump in this value, and support
+ * code in rr to handle old protocol versions. And when we bump it we'll need
+ * to figure out a way to test the old protocol versions.
+ * To be clear, changes that only affect recording and not replay, such as
+ * changes to the layout of syscall_patch_hook, do not need to bump this.
+ * Note also that SYSCALLBUF_PROTOCOL_VERSION is stored in the trace header, so
+ * replay always has access to the SYSCALLBUF_PROTOCOL_VERSION used during
+ * recording, even before the preload library is ever loaded.
+ *
+ * Version 0: initial rr 5.0.0 release
+ */
+#define SYSCALLBUF_PROTOCOL_VERSION 0
+
+#ifdef RR_IMPLEMENT_PRELOAD
+/* Avoid using <string.h> library functions */
+static inline int streq(const char* s1, const char* s2) {
+  while (1) {
+    if (*s1 != *s2) {
+      return 0;
+    }
+    if (!*s1) {
+      return 1;
+    }
+    ++s1;
+    ++s2;
+  }
+  return 1;
+}
+static inline size_t rrstrlen(const char* s) {
+  size_t ret = 0;
+  while (*s) {
+    ++s;
+    ++ret;
+  }
+  return ret;
+}
+#else
+#include <string.h>
+static inline int streq(const char* s1, const char* s2) {
+  return !strcmp(s1, s2);
+}
+static inline size_t rrstrlen(const char* s) { return strlen(s); }
+#include "../remote_ptr.h"
+#endif
+
 #include <signal.h>
 #include <stdint.h>
-#include <string.h>
 #include <sys/types.h>
 #include <sys/user.h>
 
-#ifndef RR_IMPLEMENT_PRELOAD
-#include "../remote_ptr.h"
-#endif
+static inline int strprefix(const char* s1, const char* s2) {
+  while (1) {
+    if (!*s1) {
+      return 1;
+    }
+    if (*s1 != *s2) {
+      return 0;
+    }
+    ++s1;
+    ++s2;
+  }
+  return 1;
+}
+
+static inline const char* extract_file_name(const char* s) {
+  const char* ret = s;
+  while (*s) {
+    if (*s == '/') {
+      ret = s + 1;
+    }
+    ++s;
+  }
+  return ret;
+}
 
 /* This header file is included by preload.c and various rr .cc files. It
  * defines the interface between the preload library and rr. preload.c
@@ -64,7 +131,7 @@
 /* PRELOAD_THREAD_LOCALS_ADDR should not change.
  * Tools depend on this address. */
 #define PRELOAD_THREAD_LOCALS_ADDR (RR_PAGE_ADDR + PAGE_SIZE)
-#define PRELOAD_THREAD_LOCALS_SIZE 88
+#define PRELOAD_THREAD_LOCALS_SIZE 104
 
 /* "Magic" (rr-implemented) syscalls that we use to initialize the
  * syscallbuf.
@@ -117,11 +184,15 @@
 #ifdef RR_IMPLEMENT_PRELOAD
 #define TEMPLATE_ARCH
 #define PTR(T) T*
+#define PTR_ARCH(T) T*
 #define VOLATILE volatile
+#define SIGNED_LONG long
 #else
 #define TEMPLATE_ARCH template <typename Arch>
 #define PTR(T) typename Arch::template ptr<T>
+#define PTR_ARCH(T) typename Arch::template ptr<T<Arch>>
 #define VOLATILE
+#define SIGNED_LONG typename Arch::signed_long
 #endif
 
 /**
@@ -134,11 +205,15 @@
  * instruction that follows a syscall instruction.
  * Each instance of this struct describes an instruction that can follow a
  * syscall and a hook function to patch with.
+ *
+ * This is not (and must not ever be) used during replay so we can change it
+ * without bumping SYSCALLBUF_PROTOCOL_VERSION.
  */
 struct syscall_patch_hook {
   uint8_t is_multi_instruction;
   uint8_t next_instruction_length;
-  uint8_t next_instruction_bytes[6];
+  /* Avoid any padding or anything that would make the layout arch-specific. */
+  uint8_t next_instruction_bytes[14];
   uint64_t hook_address;
 };
 
@@ -194,6 +269,16 @@ struct preload_globals {
   struct mprotect_record mprotect_records[MPROTECT_RECORD_COUNT];
 };
 
+/**
+ * Represents syscall params.  Makes it simpler to pass them around,
+ * and avoids pushing/popping all the data for calls.
+ */
+TEMPLATE_ARCH
+struct syscall_info {
+  SIGNED_LONG no;
+  SIGNED_LONG args[6];
+};
+
 /**
  * Can be architecture dependent. The rr process does not manipulate
  * these except to save and restore the values on task switches so that
@@ -205,16 +290,40 @@ struct preload_globals {
  */
 TEMPLATE_ARCH
 struct preload_thread_locals {
-  /* Pointer to alt-stack used by syscallbuf stubs (allocated at the end of
-   * the scratch buffer */
+  /* The offset of this field MUST NOT CHANGE, it is part of the preload ABI
+   * rr depends on.
+   * Offset of this field is hardcoded in syscall_hook.S and
+   * assembly_templates.py.
+   * Pointer to alt-stack used by syscallbuf stubs (allocated at the end of
+   * the scratch buffer.
+   */
   PTR(void) syscallbuf_stub_alt_stack;
-  /* Where syscall result will be (or during replay, has been) saved.
-   * The offset of this field MUST NOT CHANGE, it is part of the preload ABI
-   * tools can depend on. */
+  /* The offset of this field MUST NOT CHANGE, it is part of the preload ABI
+   * tools can depend on.
+   * Where syscall result will be (or during replay, has been) saved.
+   */
   PTR(int64_t) pending_untraced_syscall_result;
-  /* scratch space used by stub code */
+  /* The offset of this field MUST NOT CHANGE, it is part of the preload ABI
+   * rr depends on.
+   * Scratch space used by stub code.
+   */
   PTR(void) stub_scratch_1;
+  /* The offset of this field MUST NOT CHANGE, it is part of the preload ABI
+   * rr depends on.
+   */
   int alt_stack_nesting_level;
+  /**
+   * We could use this later.
+   */
+  int unused_padding;
+  /* The offset of this field MUST NOT CHANGE, it is part of the preload ABI
+   * rr depends on. It contains the parameters to the patched syscall, or
+   * zero if we're not processing a buffered syscall. Do not depend on this
+   * existing during replay, some traces with SYSCALLBUF_PROTOCOL_VERSION 0
+   * don't have it.
+   */
+  PTR_ARCH(const struct syscall_info) original_syscall_parameters;
+
   /* Nonzero when thread-local state like the syscallbuf has been
    * initialized.  */
   int thread_inited;
@@ -281,6 +390,11 @@ struct rrcall_init_preload_params {
   int syscall_patch_hook_count;
   PTR(struct syscall_patch_hook) syscall_patch_hooks;
   PTR(void) syscallhook_vsyscall_entry;
+  PTR(void) syscallbuf_code_start;
+  PTR(void) syscallbuf_code_end;
+  PTR(void) get_pc_thunks_start;
+  PTR(void) get_pc_thunks_end;
+  PTR(void) syscallbuf_final_exit_instruction;
   PTR(struct preload_globals) globals;
   /* Address of the first entry of the breakpoint table.
    * After processing a sycallbuf record (and unlocking the syscallbuf),
@@ -385,17 +499,21 @@ struct syscallbuf_hdr {
    * When it's zero, the desched signal can safely be
    * discarded. */
   volatile uint8_t desched_signal_may_be_relevant;
+  /* A copy of the tasks's signal mask. Updated by preload when a buffered
+   * rt_sigprocmask executes.
+   */
+  volatile uint64_t blocked_sigs;
+  /* Incremented by preload every time a buffered rt_sigprocmask executes.
+   * Cleared during syscallbuf reset.
+   */
+  volatile uint32_t blocked_sigs_generation;
   /* Nonzero when preload is in the process of calling an untraced
    * sigprocmask; the real sigprocmask may or may not match blocked_sigs.
    */
   volatile uint8_t in_sigprocmask_critical_section;
-  /* A copy of the tasks's signal mask. Updated by both preload and rr in
-   * response to events that change the signal mask. RR stores this in
-   * RecordTask::blocked_sigs, but since some signal mask affecting system
-   * calls may be buffered, that value may be out of date until rr reloads it
-   * from here.
-   */
-  volatile uint64_t blocked_sigs;
+  /* Nonzero when the syscall was aborted during preparation without doing
+   * anything */
+  volatile uint8_t failed_during_preparation;
 
   struct syscallbuf_record recs[0];
 } __attribute__((__packed__));
@@ -446,38 +564,54 @@ inline static long stored_record_size(size_t length) {
  * too much by piling on.
  */
 inline static int is_blacklisted_filename(const char* filename) {
-  return !strncmp("/dev/dri/", filename, 9) ||
-         !strcmp("/dev/nvidiactl", filename) ||
-         !strcmp("/usr/share/alsa/alsa.conf", filename);
+  const char* f;
+  if (strprefix("/dev/dri/", filename) || streq("/dev/nvidiactl", filename) ||
+      streq("/usr/share/alsa/alsa.conf", filename)) {
+    return 1;
+  }
+  f = extract_file_name(filename);
+  return strprefix("rr-test-blacklist-file_name", f) ||
+         strprefix("pulse-shm-", f);
+}
+
+inline static int is_blacklisted_memfd(const char* name) {
+  return streq("pulseaudio", name);
+}
+
+inline static int is_blacklisted_socket(const char* filename) {
+  /* Blacklist the nscd socket because glibc communicates with the daemon over
+   * shared memory rr can't handle.
+   */
+  return streq("/var/run/nscd/socket", filename);
 }
 
 inline static int is_gcrypt_deny_file(const char* filename) {
-  return !strcmp("/etc/gcrypt/hwf.deny", filename);
+  return streq("/etc/gcrypt/hwf.deny", filename);
 }
 
 inline static int is_terminal(const char* filename) {
-  return !strncmp("/dev/tty", filename, 8) || !strncmp("/dev/pts", filename, 8);
+  return strprefix("/dev/tty", filename) || strprefix("/dev/pts", filename);
 }
 
 inline static int is_proc_mem_file(const char* filename) {
-  if (strncmp("/proc/", filename, 6)) {
+  if (!strprefix("/proc/", filename)) {
     return 0;
   }
-  return !strcmp(filename + strlen(filename) - 4, "/mem");
+  return streq(filename + rrstrlen(filename) - 4, "/mem");
 }
 
 inline static int is_proc_fd_dir(const char* filename) {
-  if (strncmp("/proc/", filename, 6)) {
+  if (!strprefix("/proc/", filename)) {
     return 0;
   }
 
-  int len = strlen(filename);
+  int len = rrstrlen(filename);
   const char* fd_bit = filename + len;
   if (*fd_bit == '/') {
     fd_bit--;
   }
 
-  return !strncmp(fd_bit - 3, "/fd", 3);
+  return strprefix("/fd", fd_bit - 3);
 }
 
 /**
