@@ -2,7 +2,6 @@
 
 #include "record_signal.h"
 
-#include <assert.h>
 #include <fcntl.h>
 #include <linux/perf_event.h>
 #include <sched.h>
@@ -23,6 +22,8 @@
 #include "RecordSession.h"
 #include "RecordTask.h"
 #include "TraceStream.h"
+#include "VirtualPerfCounterMonitor.h"
+#include "core.h"
 #include "kernel_metadata.h"
 #include "log.h"
 #include "util.h"
@@ -41,65 +42,109 @@ static size_t sigaction_sigset_size(SupportedArch arch) {
   RR_ARCH_FUNCTION(sigaction_sigset_size_arch, arch);
 }
 
-/**
- * Restore the blocked-ness and sigaction for |sig| from |t|'s local
- * copy.
- */
-static void restore_signal_state(RecordTask* t, int sig) {
-  AutoRemoteSyscalls remote(t);
-  size_t sigset_size = sigaction_sigset_size(remote.arch());
-  {
+static void restore_sighandler_if_not_default(RecordTask* t, int sig) {
+  if (t->sig_disposition(sig) != SIGNAL_DEFAULT) {
+    LOG(debug) << "Restoring signal handler for " << signal_name(sig);
+    AutoRemoteSyscalls remote(t);
+    size_t sigset_size = sigaction_sigset_size(remote.arch());
     const vector<uint8_t>& sa = t->signal_action(sig);
     AutoRestoreMem child_sa(remote, sa.data(), sa.size());
     remote.infallible_syscall(syscall_number_for_rt_sigaction(remote.arch()),
                               sig, child_sa.get().as_int(), nullptr,
                               sigset_size);
   }
-  if (t->is_sig_blocked(sig)) {
+}
+
+/**
+ * Restore the blocked-ness and sigaction for |sig| from |t|'s local
+ * copy.
+ */
+static void restore_signal_state(RecordTask* t, int sig,
+                                 SignalBlocked signal_was_blocked) {
+  restore_sighandler_if_not_default(t, sig);
+  if (signal_was_blocked) {
+    LOG(debug) << "Restoring signal blocked-ness for " << signal_name(sig);
+    AutoRemoteSyscalls remote(t);
+    size_t sigset_size = sigaction_sigset_size(remote.arch());
     vector<uint8_t> bytes;
     bytes.resize(sigset_size);
     memset(bytes.data(), 0, sigset_size);
-    uint64_t mask = uint64_t(1) << (sig - 1);
+    sig_set_t mask = signal_bit(sig);
     ASSERT(t, sigset_size >= sizeof(mask));
     memcpy(bytes.data(), &mask, sizeof(mask));
     AutoRestoreMem child_block(remote, bytes.data(), bytes.size());
     remote.infallible_syscall(syscall_number_for_rt_sigprocmask(remote.arch()),
                               SIG_BLOCK, child_block.get().as_int(), nullptr,
                               sigset_size);
+    // We just changed the sigmask ourselves.
+    t->invalidate_sigmask();
   }
 }
 
-/** Return true iff |t->ip()| points at a RDTSC instruction. */
-static const uint8_t rdtsc_insn[] = { 0x0f, 0x31 };
-static bool is_ip_rdtsc(RecordTask* t) {
-  uint8_t insn[sizeof(rdtsc_insn)];
-  if (sizeof(insn) !=
-      t->read_bytes_fallible(t->ip().to_data_ptr<uint8_t>(), sizeof(insn),
-                             insn)) {
-    return false;
-  }
-  return !memcmp(insn, rdtsc_insn, sizeof(insn));
-}
+static const uintptr_t CPUID_RDRAND_FLAG = 1 << 30;
+static const uintptr_t CPUID_RTM_FLAG = 1 << 11;
+static const uintptr_t CPUID_RDSEED_FLAG = 1 << 18;
 
 /**
  * Return true if |t| was stopped because of a SIGSEGV resulting
- * from a rdtsc and |t| was updated appropriately, false otherwise.
+ * from a disabled instruction and |t| was updated appropriately, false
+ * otherwise.
  */
-static bool try_handle_rdtsc(RecordTask* t, siginfo_t* si) {
+static bool try_handle_disabled_insn(RecordTask* t, siginfo_t* si) {
   ASSERT(t, si->si_signo == SIGSEGV);
 
-  if (!is_ip_rdtsc(t) || t->tsc_mode == PR_TSC_SIGSEGV) {
-    return false;
+  auto disabled_insn = disabled_insn_at(t, t->ip());
+  switch (disabled_insn) {
+    case DisabledInsn::RDTSC:
+    case DisabledInsn::RDTSCP:
+      if (t->tsc_mode == PR_TSC_SIGSEGV) {
+        return false;
+      }
+      break;
+    case DisabledInsn::CPUID:
+      if (t->cpuid_mode == 0) {
+        return false;
+      }
+      break;
+    default:
+      return false;
   }
 
-  unsigned long long current_time = rdtsc();
+  size_t len = disabled_insn_len(disabled_insn);
+  ASSERT(t, len > 0);
+
   Registers r = t->regs();
-  r.set_rdtsc_output(current_time);
-  r.set_ip(r.ip() + sizeof(rdtsc_insn));
+  if (disabled_insn == DisabledInsn::RDTSC ||
+      disabled_insn == DisabledInsn::RDTSCP) {
+    unsigned long long current_time = rdtsc();
+    r.set_rdtsc_output(current_time);
+
+    LOG(debug) << " trapped for rdtsc: returning " << current_time;
+  } else if (disabled_insn == DisabledInsn::CPUID) {
+    auto eax = r.syscallno();
+    auto ecx = r.cx();
+    auto cpuid_data = cpuid(eax, ecx);
+    switch (eax) {
+      case CPUID_GETFEATURES:
+        cpuid_data.ecx &= ~CPUID_RDRAND_FLAG;
+        break;
+      case CPUID_GETEXTENDEDFEATURES:
+        if (ecx == 0) {
+          cpuid_data.ebx &= ~(CPUID_RDSEED_FLAG | CPUID_RTM_FLAG);
+        }
+        break;
+      default:
+        break;
+    }
+    r.set_cpuid_output(cpuid_data.eax, cpuid_data.ebx, cpuid_data.ecx,
+                       cpuid_data.edx);
+    LOG(debug) << " trapped for cpuid: " << HEX(eax) << ":" << HEX(ecx);
+  }
+
+  r.set_ip(r.ip() + len);
   t->set_regs(r);
 
-  t->push_event(Event(EV_SEGV_RDTSC, HAS_EXEC_INFO, t->arch()));
-  LOG(debug) << "  trapped for rdtsc: returning " << current_time;
+  t->push_event(Event::instruction_trap());
   return true;
 }
 
@@ -177,9 +222,8 @@ static bool try_grow_map(RecordTask* t, siginfo_t* si) {
   t->trace_writer().write_mapped_region(t, km, km.fake_stat());
   // No need to flush syscallbuf here. It's safe to map these pages "early"
   // before they're really needed.
-  t->record_event(Event(EV_GROW_MAP, NO_EXEC_INFO, t->arch()),
-                  RecordTask::DONT_FLUSH_SYSCALLBUF);
-  t->push_event(Event::noop(t->arch()));
+  t->record_event(Event::grow_map(), RecordTask::DONT_FLUSH_SYSCALLBUF);
+  t->push_event(Event::noop());
   LOG(debug) << "try_grow_map " << addr << ": extended map "
              << t->vm()->mapping_of(addr).map;
   return true;
@@ -197,14 +241,102 @@ void arm_desched_event(RecordTask* t) {
   }
 }
 
+template <typename Arch>
+static remote_code_ptr get_stub_scratch_1_arch(RecordTask* t) {
+  auto locals = t->read_mem(AddressSpace::preload_thread_locals_start()
+                                .cast<preload_thread_locals<Arch>>());
+  return locals.stub_scratch_1.rptr().as_int();
+}
+
+static remote_code_ptr get_stub_scratch_1(RecordTask* t) {
+  RR_ARCH_FUNCTION(get_stub_scratch_1_arch, t->arch(), t);
+}
+
+/**
+ * This function is responsible for handling breakpoints we set in syscallbuf
+ * code to detect sigprocmask calls and syscallbuf exit. It's called when we
+ * get a SIGTRAP. Returns true if the SIGTRAP was called by one of our
+ * breakpoints and should be hidden from the application.
+ * If it was triggered by one of our breakpoints, we have to call
+ * restore_sighandler_if_not_default(t, SIGTRAP) to make sure the SIGTRAP
+ * handler is properly restored if the kernel cleared it.
+ */
+bool handle_syscallbuf_breakpoint(RecordTask* t) {
+  if (t->is_at_syscallbuf_final_instruction_breakpoint()) {
+    LOG(debug) << "Reached final syscallbuf instruction, singlestepping to "
+                  "enable signal dispatch";
+    // This is a single instruction that jumps to the location stored in
+    // preload_thread_locals::stub_scratch_1. Emulate it.
+    Registers r = t->regs();
+    r.set_ip(get_stub_scratch_1(t));
+    t->set_regs(r);
+
+    restore_sighandler_if_not_default(t, SIGTRAP);
+    // Now we're back in application code so any pending stashed signals
+    // will be handled.
+    return true;
+  }
+
+  if (!t->is_at_syscallbuf_syscall_entry_breakpoint()) {
+    return false;
+  }
+
+  Registers r = t->regs();
+  r.set_ip(r.ip().decrement_by_bkpt_insn_length(t->arch()));
+  t->set_regs(r);
+
+  if (t->is_at_traced_syscall_entry()) {
+    // We will automatically dispatch stashed signals now since this is an
+    // allowed place to dispatch signals.
+    LOG(debug) << "Allowing signal dispatch at traced-syscall breakpoint";
+    restore_sighandler_if_not_default(t, SIGTRAP);
+    return true;
+  }
+
+  // We're at an untraced-syscall entry point.
+  // To allow an AutoRemoteSyscall, we need to make sure desched signals are
+  // disarmed (and rearmed afterward).
+  bool armed_desched_event = t->read_mem(
+      REMOTE_PTR_FIELD(t->syscallbuf_child, desched_signal_may_be_relevant));
+  if (armed_desched_event) {
+    disarm_desched_event(t);
+  }
+  restore_sighandler_if_not_default(t, SIGTRAP);
+  if (armed_desched_event) {
+    arm_desched_event(t);
+  }
+
+  // This is definitely a native-arch syscall.
+  if (is_rt_sigprocmask_syscall(r.syscallno(), t->arch())) {
+    // Don't proceed with this syscall. Emulate it returning EAGAIN.
+    // Syscallbuf logic will retry using a traced syscall instead.
+    r.set_syscall_result(-EAGAIN);
+    r.set_ip(r.ip().increment_by_syscall_insn_length(t->arch()));
+    t->set_regs(r);
+    t->canonicalize_regs(t->arch());
+    LOG(debug) << "Emulated EAGAIN to avoid untraced sigprocmask with pending "
+                  "stashed signal";
+    // Leave breakpoints enabled since we want to break at the traced-syscall
+    // fallback for rt_sigprocmask.
+    return true;
+  }
+
+  // We can proceed with the untraced syscall. Either it will complete and
+  // execution will continue until we reach some point where we can deliver our
+  // signal, or it will block at which point we'll be able to deliver our
+  // signal.
+  LOG(debug) << "Disabling breakpoints at untraced syscalls";
+  t->break_at_syscallbuf_untraced_syscalls = false;
+  return true;
+}
+
 /**
  * Return the event needing to be processed after this desched of |t|.
  * The tracee's execution may be advanced, and if so |regs| is updated
  * to the tracee's latest state.
  */
 static void handle_desched_event(RecordTask* t, const siginfo_t* si) {
-  ASSERT(t, (SYSCALLBUF_DESCHED_SIGNAL == si->si_signo &&
-             si->si_code == POLL_IN && si->si_fd == t->desched_fd_child))
+  ASSERT(t, SYSCALLBUF_DESCHED_SIGNAL == si->si_signo && si->si_code == POLL_IN)
       << "Tracee is using SIGPWR??? (siginfo=" << *si << ")";
 
   /* If the tracee isn't in the critical section where a desched
@@ -228,7 +360,7 @@ static void handle_desched_event(RecordTask* t, const siginfo_t* si) {
      * has cleared the relevancy flag, but not yet
      * disarmed the event itself. */
     disarm_desched_event(t);
-    t->push_event(Event::noop(t->arch()));
+    t->push_event(Event::noop());
     return;
   }
 
@@ -309,6 +441,16 @@ static void handle_desched_event(RecordTask* t, const siginfo_t* si) {
       }
       break;
     }
+    if (t->ptrace_event() == PTRACE_EVENT_SECCOMP) {
+      ASSERT(t,
+             t->session().syscall_seccomp_ordering() ==
+                 Session::SECCOMP_BEFORE_PTRACE_SYSCALL);
+      // This is the old kernel event ordering. This must be a SECCOMP event
+      // for the buffered syscall; it's not rr-generated because this is an
+      // untraced syscall, but it could be generated by a tracee's
+      // seccomp filter.
+      break;
+    }
 
     // Completely ignore spurious desched signals and
     // signals that aren't going to be delivered to the
@@ -326,6 +468,12 @@ static void handle_desched_event(RecordTask* t, const siginfo_t* si) {
     // and the tracee never switched out.
     int sig = t->stop_sig();
     ASSERT(t, sig) << "expected stop-signal, got " << t->status();
+    if (SIGTRAP == sig && handle_syscallbuf_breakpoint(t)) {
+      // We stopped at a breakpoint on an untraced may-block syscall.
+      // This can't be relevant to us since sigprocmask isn't may-block.
+      LOG(debug) << " disabling breakpoints on untraced syscalls";
+      continue;
+    }
     if (SYSCALLBUF_DESCHED_SIGNAL == sig ||
         PerfCounters::TIME_SLICE_SIGNAL == sig || t->is_sig_ignored(sig)) {
       LOG(debug) << "  dropping ignored " << signal_name(sig);
@@ -339,7 +487,7 @@ static void handle_desched_event(RecordTask* t, const siginfo_t* si) {
   if (t->is_disarm_desched_event_syscall()) {
     LOG(debug)
         << "  (at disarm-desched, so finished buffered syscall; resuming)";
-    t->push_event(Event::noop(t->arch()));
+    t->push_event(Event::noop());
     return;
   }
 
@@ -352,8 +500,9 @@ static void handle_desched_event(RecordTask* t, const siginfo_t* si) {
      * reset until we've finished guiding the tracee through this
      * interrupted call.  We use the record counter for
      * assertions. */
-    ASSERT(t, !t->delay_syscallbuf_reset);
-    t->delay_syscallbuf_reset = true;
+    ASSERT(t, !t->delay_syscallbuf_reset_for_desched);
+    t->delay_syscallbuf_reset_for_desched = true;
+    LOG(debug) << "Desched initiated";
 
     /* The tracee is (re-)entering the buffered syscall.  Stash
      * away this breadcrumb so that we can figure out what syscall
@@ -361,7 +510,7 @@ static void handle_desched_event(RecordTask* t, const siginfo_t* si) {
      * off the syscallbuf, if needed. */
     remote_ptr<const struct syscallbuf_record> desched_rec =
         t->next_syscallbuf_record();
-    t->push_event(DeschedEvent(desched_rec, t->arch()));
+    t->push_event(DeschedEvent(desched_rec));
     int call = t->read_mem(REMOTE_PTR_FIELD(t->desched_rec(), syscallno));
 
     /* The descheduled syscall was interrupted by a signal, like
@@ -388,7 +537,7 @@ static void handle_desched_event(RecordTask* t, const siginfo_t* si) {
   // state to ENTERING_SYSCALL
 
   LOG(debug) << "  resuming (and probably switching out) blocked `"
-             << t->syscall_name(call) << "'";
+             << syscall_name(call, ev.arch()) << "'";
 }
 
 static bool is_safe_to_deliver_signal(RecordTask* t, siginfo_t* si) {
@@ -408,12 +557,19 @@ static bool is_safe_to_deliver_signal(RecordTask* t, siginfo_t* si) {
                << " because in traced syscall";
     return true;
   }
+  if (t->is_at_traced_syscall_entry()) {
+    LOG(debug) << "Safe to deliver signal at " << t->ip()
+               << " because at entry to traced syscall";
+    return true;
+  }
 
   if (t->is_in_untraced_syscall() && t->desched_rec()) {
+    // Untraced syscalls always use the architecture of the process
     LOG(debug) << "Safe to deliver signal at " << t->ip()
                << " because tracee interrupted by desched of "
-               << t->syscall_name(t->read_mem(
-                      REMOTE_PTR_FIELD(t->desched_rec(), syscallno)));
+               << syscall_name(t->read_mem(REMOTE_PTR_FIELD(t->desched_rec(),
+                                                            syscallno)),
+                               t->arch());
     return true;
   }
 
@@ -434,10 +590,14 @@ static bool is_safe_to_deliver_signal(RecordTask* t, siginfo_t* si) {
       return true;
     }
 
-    // Our emulation of SYS_rrcall_notify_syscall_hook_exit clears this flag.
-    t->write_mem(
-        REMOTE_PTR_FIELD(t->syscallbuf_child, notify_on_syscall_hook_exit),
-        (uint8_t)1);
+    // A signal (e.g. seccomp SIGSYS) interrupted a untraced syscall in a
+    // non-restartable way. Defer it until SYS_rrcall_notify_syscall_hook_exit.
+    if (t->is_in_untraced_syscall()) {
+      // Our emulation of SYS_rrcall_notify_syscall_hook_exit clears this flag.
+      t->write_mem(
+          REMOTE_PTR_FIELD(t->syscallbuf_child, notify_on_syscall_hook_exit),
+          (uint8_t)1);
+    }
   }
 
   LOG(debug) << "Not safe to deliver signal at " << t->ip();
@@ -445,60 +605,65 @@ static bool is_safe_to_deliver_signal(RecordTask* t, siginfo_t* si) {
 }
 
 SignalHandled handle_signal(RecordTask* t, siginfo_t* si,
-                            SignalDeterministic deterministic) {
-  LOG(debug) << t->tid << ": handling signal " << signal_name(si->si_signo)
+                            SignalDeterministic deterministic,
+                            SignalBlocked signal_was_blocked) {
+  int sig = si->si_signo;
+  LOG(debug) << t->tid << ": handling signal " << signal_name(sig)
              << " (pevent: " << ptrace_event_name(t->ptrace_event())
              << ", event: " << t->ev();
 
-  /* We have to check for a desched event first, because for
-   * those we *do not* want to (and cannot, most of the time)
-   * step the tracee out of the syscallbuf code before
-   * attempting to deliver the signal. */
-  if (SYSCALLBUF_DESCHED_SIGNAL == si->si_signo) {
-    handle_desched_event(t, si);
-    return SIGNAL_HANDLED;
-  }
+  // Conservatively invalidate the sigmask in case just accepting a signal has
+  // sigmask effects.
+  t->invalidate_sigmask();
 
-  if (!is_safe_to_deliver_signal(t, si)) {
-    return DEFER_SIGNAL;
-  }
-
-  if (!t->set_siginfo_for_synthetic_SIGCHLD(si)) {
-    return DEFER_SIGNAL;
-  }
-
-  int sig = si->si_signo;
-  // When a deterministic signal is triggered, but the signal is currently
-  // blocked or ignored, the kernel (in |force_sig_info|) unblocks it and
-  // sets its disposition to SIG_DFL. It never undoes this (probably
-  // because it expects the signal to be fatal, which it always would be
-  // unless a ptracer intercepts the signal as we do). Therefore, if the signal
-  // was generated for rr's purposes, we need to restore the signal state
-  // ourselves. If the signal was not generated for rr's purposes, we'll
-  // modify our copy of the signal state to match what the kernel did.
-
-  /* See if this signal occurred because of an rr implementation detail,
-   * and fudge t appropriately. */
-  switch (sig) {
-    case SIGSEGV:
-      if (deterministic == DETERMINISTIC_SIG &&
-          (try_handle_rdtsc(t, si) || try_grow_map(t, si))) {
-        if (t->is_sig_blocked(sig) || t->is_sig_ignored(sig)) {
-          restore_signal_state(t, sig);
-        }
-        return SIGNAL_HANDLED;
+  if (deterministic == DETERMINISTIC_SIG) {
+    // When a deterministic signal is triggered, but the signal is currently
+    // blocked or ignored, the kernel (in |force_sig_info|) unblocks it and
+    // sets its disposition to SIG_DFL. It never undoes this (probably
+    // because it expects the signal to be fatal, which it always would be
+    // unless a ptracer intercepts the signal as we do). Therefore, if the
+    // signal was generated for rr's purposes, we need to restore the signal
+    // state ourselves.
+    if (sig == SIGSEGV &&
+        (try_handle_disabled_insn(t, si) || try_grow_map(t, si))) {
+      if (signal_was_blocked || t->is_sig_ignored(sig)) {
+        restore_signal_state(t, sig, signal_was_blocked);
       }
-      break;
-
-    case PerfCounters::TIME_SLICE_SIGNAL:
-      t->push_event(Event(EV_SCHED, HAS_EXEC_INFO, t->arch()));
       return SIGNAL_HANDLED;
+    }
+
+    // Since we're not undoing the kernel's changes, update our signal handler
+    // state to match the kernel's.
+    if (signal_was_blocked || t->is_sig_ignored(sig)) {
+      t->set_sig_handler_default(sig);
+    }
   }
 
-  if (deterministic == DETERMINISTIC_SIG &&
-      (t->is_sig_blocked(sig) || t->is_sig_ignored(sig))) {
-    t->set_sig_blocked(sig, false);
-    t->set_sig_handler_default(sig);
+  if (!VirtualPerfCounterMonitor::is_virtual_perf_counter_signal(si)) {
+    /* We have to check for a desched event first, because for
+     * those we *do not* want to (and cannot, most of the time)
+     * step the tracee out of the syscallbuf code before
+     * attempting to deliver the signal. */
+    if (SYSCALLBUF_DESCHED_SIGNAL == si->si_signo) {
+      handle_desched_event(t, si);
+      return SIGNAL_HANDLED;
+    }
+
+    if (!is_safe_to_deliver_signal(t, si)) {
+      return DEFER_SIGNAL;
+    }
+
+    if (!t->set_siginfo_for_synthetic_SIGCHLD(si)) {
+      return DEFER_SIGNAL;
+    }
+
+    if (sig == PerfCounters::TIME_SLICE_SIGNAL) {
+      t->push_event(Event::sched());
+      return SIGNAL_HANDLED;
+    }
+  } else {
+    // Clear the magic flag so it doesn't leak into the program.
+    si->si_errno = 0;
   }
 
   /* This signal was generated by the program or an external
@@ -513,13 +678,11 @@ SignalHandled handle_signal(RecordTask* t, siginfo_t* si,
     // currently assumes it won't encounter a deterministic SIGTRAP (due to
     // a hardcoded breakpoint in the tracee).
     if (deterministic == DETERMINISTIC_SIG) {
-      t->push_event(SignalEvent(*si, deterministic, t));
-      t->record_current_event();
-      t->pop_event(EV_SIGNAL);
+      t->record_event(Event(EV_SIGNAL, SignalEvent(*si, deterministic,
+                                                   t->sig_resolved_disposition(
+                                                       sig, deterministic))));
     } else {
-      t->push_event(Event(EV_SCHED, HAS_EXEC_INFO, t->arch()));
-      t->record_current_event();
-      t->pop_event(EV_SCHED);
+      t->record_event(Event::sched());
     }
     // ptracer has been notified, so don't deliver the signal now.
     // The signal won't be delivered for real until the ptracer calls
@@ -527,7 +690,9 @@ SignalHandled handle_signal(RecordTask* t, siginfo_t* si,
     return SIGNAL_PTRACE_STOP;
   }
 
-  t->push_event(SignalEvent(*si, deterministic, t));
+  t->push_event(Event(
+      EV_SIGNAL, SignalEvent(*si, deterministic,
+                             t->sig_resolved_disposition(sig, deterministic))));
   return SIGNAL_HANDLED;
 }
 
