@@ -436,7 +436,7 @@ sethostname = EmulatedSyscall(x86=74, x64=170)
 # by other means.
 setrlimit = EmulatedSyscall(x86=75, x64=160)
 
-getrlimit = EmulatedSyscall(x64=97, arg2="typename Arch::rlimit")
+getrlimit = EmulatedSyscall(x86=76, x64=97, arg2="typename Arch::rlimit")
 
 #  int getrusage(int who, struct rusage *usage)
 #
@@ -639,7 +639,12 @@ setdomainname = EmulatedSyscall(x86=121, x64=171)
 uname = EmulatedSyscall(x86=122, x64=63, arg1="typename Arch::utsname")
 
 modify_ldt = IrregularEmulatedSyscall(x86=123, x64=154)
-adjtimex = UnsupportedSyscall(x86=124, x64=159)
+
+#  int adjtimex(struct timex *buf);
+#
+# adjtimex() takes a pointer to a timex structure, reads it, and returns
+# the same structure updated with the current kernel values.
+adjtimex = EmulatedSyscall(x86=124, x64=159, arg1="typename Arch::timex")
 
 #  int mprotect(const void *addr, size_t len, int prot)
 #
@@ -1206,7 +1211,7 @@ sched_getaffinity = IrregularEmulatedSyscall(x86=242, x64=204)
 set_thread_area = IrregularEmulatedSyscall(x86=243, x64=205)
 
 get_thread_area = IrregularEmulatedSyscall(x86=244, x64=211)
-io_setup = UnsupportedSyscall(x86=245, x64=206)
+io_setup = IrregularEmulatedSyscall(x86=245, x64=206)
 io_destroy = UnsupportedSyscall(x86=246, x64=207)
 io_getevents = UnsupportedSyscall(x86=247, x64=208)
 io_submit = UnsupportedSyscall(x86=248, x64=209)
@@ -1331,8 +1336,8 @@ fadvise64_64 = EmulatedSyscall(x86=272)
 
 vserver = InvalidSyscall(x86=273, x64=236)
 mbind = EmulatedSyscall(x86=274, x64=237)
-get_mempolicy = UnsupportedSyscall(x86=275, x64=239)
-set_mempolicy = UnsupportedSyscall(x86=276, x64=238)
+get_mempolicy = IrregularEmulatedSyscall(x86=275, x64=239)
+set_mempolicy = EmulatedSyscall(x86=276, x64=238)
 
 mq_open = EmulatedSyscall(x86=277, x64=240)
 mq_unlink = EmulatedSyscall(x86=278, x64=241)
@@ -1473,7 +1478,7 @@ tee = UnsupportedSyscall(x86=315, x64=276)
 vmsplice = UnsupportedSyscall(x86=316, x64=278)
 move_pages = UnsupportedSyscall(x86=317, x64=279)
 getcpu = EmulatedSyscall(x86=318, x64=309, arg1="unsigned int", arg2="unsigned int")
-epoll_pwait = UnsupportedSyscall(x86=319, x64=281)
+epoll_pwait = IrregularEmulatedSyscall(x86=319, x64=281)
 
 #  int utimensat(int dirfd, const char *pathname, const struct timespec
 #times[2], int flags);
@@ -1633,7 +1638,8 @@ sched_getattr = UnsupportedSyscall(x86=352, x64=315)
 renameat2 = EmulatedSyscall(x86=353, x64=316)
 seccomp = IrregularEmulatedSyscall(x86=354, x64=317)
 getrandom = IrregularEmulatedSyscall(x86=355, x64=318)
-memfd_create = EmulatedSyscall(x86=356, x64=319)
+memfd_create = IrregularEmulatedSyscall(x86=356, x64=319)
+arch_prctl = IrregularEmulatedSyscall(x86=384, x64=158)
 
 bpf = UnsupportedSyscall(x86=357, x64=321)
 execveat = UnsupportedSyscall(x86=358, x64=322)
@@ -1656,7 +1662,7 @@ rrcall_mprotect_record = IrregularEmulatedSyscall(x86=447, x64=447)
 
 # These syscalls are also subsumed under socketcall on x86.
 socket = EmulatedSyscall(x86=359, x64=41)
-connect = EmulatedSyscall(x86=362, x64=42)
+connect = IrregularEmulatedSyscall(x86=362, x64=42)
 accept = IrregularEmulatedSyscall(x64=43)
 sendto = IrregularEmulatedSyscall(x86=369, x64=44)
 recvfrom = IrregularEmulatedSyscall(x86=371, x64=45)
@@ -1687,7 +1693,6 @@ msgctl = IrregularEmulatedSyscall(x64=71)
 semtimedop = IrregularEmulatedSyscall(x64=220)
 
 # These syscalls simply don't exist on x86.
-arch_prctl = IrregularEmulatedSyscall(x64=158)
 tuxcall = InvalidSyscall(x64=184)
 security = InvalidSyscall(x64=185)
 epoll_ctl_old = UnsupportedSyscall(x64=214)
