@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 static void fault_handler(__attribute__((unused)) int sig,
                           __attribute__((unused)) siginfo_t* si,
