@@ -1,7 +1,7 @@
 import sys
 import re
 
-syscall_re = re.compile("`SYSCALL: futex' \\(state:EXITING_SYSCALL\\)")
+syscall_re = re.compile("`SIGNAL: ")
 time_re = re.compile("global_time:(\d+)")
 futex_time = 999999999
 
