@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 static void overwrite_file(const char* path, ssize_t num_bytes) {
   const int magic = 0x5a5a5a5a;
@@ -17,8 +17,8 @@ int main(void) {
   int ret;
   int status;
 
-  char name[] = "/tmp/rr-clone-newflags-XXXXXX";
-  int fd = mkstemp(name);
+  static const char name[] = "temp";
+  int fd = open(name, O_CREAT | O_RDWR | O_EXCL, 0600);
 
   overwrite_file(name, 0x1000);
 
