@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 static int not_shared;
 static int* shared;
@@ -32,7 +32,7 @@ int main(int argc, char* argv[]) {
                 CLONE_VFORK | SIGCHLD, (void*)exe);
 
   /* This should not execute until after the vfork child has execed */
-  test_assert(*shared = 1);
+  test_assert(*shared == 1);
 
   test_assert(child == waitpid(child, &status, 0));
   test_assert(WIFEXITED(status) && 0 == WEXITSTATUS(status));
