@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 static int lock;
 static int accumulator = 9999;
@@ -8,7 +8,8 @@ static int accumulator = 9999;
 static void hle_abort(void) {
   int j;
 #if defined(__x86_64__) || defined(__i386__)
-  asm("xacquire; lock addl $1,%0\n\t" : : "m"(lock));
+  /* Use .byte 0xf2 instead of XACQUIRE for old assemblers */
+  asm(".byte 0xf2; lock addl $1,%0\n\t" : : "m"(lock));
 #else
 #error Unknown architecture
 #endif
@@ -23,7 +24,8 @@ static void hle_abort(void) {
   /* Force abort */
   sched_yield();
 #if defined(__x86_64__) || defined(__i386__)
-  asm("xrelease; lock subl $1,%0\n\t" : : "m"(lock));
+  /* Use .byte 0xf3 instead of XRELEASE for old assemblers */
+  asm(".byte 0xf3; lock subl $1,%0\n\t" : : "m"(lock));
 #else
 #error Unknown architecture
 #endif
