@@ -5,7 +5,7 @@ static void breakpoint(void) {
    * Tests rely on that. */
 }
 
-#include "rrutil.h"
+#include "util.h"
 
 static void handle_sigtrap(__attribute__((unused)) int sig) {
   atomic_puts("EXIT-SUCCESS");
