@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 #define NUM_ITERATIONS (1 << 27)
 
@@ -41,6 +41,6 @@ int main(void) {
 
   spin();
 
-  atomic_printf("EXIT-SUCCESS\n");
+  atomic_puts("EXIT-SUCCESS");
   return 0;
 }
