@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 static void* thread(__attribute__((unused)) void* p) {
   sigset_t mask;
@@ -26,7 +26,7 @@ static void handle_signal(int sig, __attribute__((unused)) siginfo_t* si,
   } else if (SIGUSR2 == sig) {
     ++usr2_hit;
   } else {
-    assert("Unexpected signal" && 0);
+    test_assert("Unexpected signal" && 0);
   }
 }
 
