@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 const char* sockaddr_name(const struct sockaddr* addr) {
   const struct sockaddr_in* sin = (const struct sockaddr_in*)addr;
@@ -97,24 +97,44 @@ static void get_ifconfig(int sockfd, struct ifreq* req, struct ifreq* eth_req) {
   }
 }
 
+static void generic_request_by_name(int sockfd, struct ifreq* req, int nr,
+                                    const char* nr_str) {
+  int ret;
+  memset(&req->ifr_ifru, 0xff, sizeof(req->ifr_ifru));
+  ret = ioctl(sockfd, nr, req);
+  VERIFY_GUARD(req);
+  atomic_printf("%s(ret:%d): %s ", nr_str, ret, req->ifr_name);
+  test_assert(0 == ret);
+}
+
 #define GENERIC_REQUEST_BY_NAME(nr)                                            \
-  memset(&req->ifr_ifru, 0xff, sizeof(req->ifr_ifru));                         \
-  ret = ioctl(sockfd, nr, req);                                                \
-  VERIFY_GUARD(req);                                                           \
-  atomic_printf(#nr "(ret:%d): %s ", ret, req->ifr_name);                      \
-  test_assert(0 == ret)
+  generic_request_by_name(sockfd, req, nr, #nr)
+
+static int generic_wireless_request_by_name_internal(int sockfd,
+                                                     struct iwreq** wreq,
+                                                     const char* name, int nr,
+                                                     const char* nr_str) {
+  int err;
+  int ret;
+  /* Note that we use sizeof(struct ifreq) here, not iwreq, because of
+   * https://bugzilla.kernel.org/show_bug.cgi?id=195869
+   */
+  *wreq = allocate_guard(sizeof(struct ifreq), 'd');
+  strcpy((*wreq)->ifr_ifrn.ifrn_name, name);
+  ret = ioctl(sockfd, nr, *wreq);
+  VERIFY_GUARD(*wreq);
+  err = errno;
+  atomic_printf("%s(ret:%d): %s: ", nr_str, ret, (*wreq)->ifr_name);
+  if (-1 == ret) {
+    atomic_printf("WARNING: %s doesn't support ioctl %s\n", name, nr_str);
+    test_assert(EOPNOTSUPP == err || EPERM == err || EINVAL == err);
+  }
+  return ret;
+}
 
 #define GENERIC_WIRELESS_REQUEST_BY_NAME_INTERNAL(nr, nr_str, args)            \
-  ALLOCATE_GUARD(wreq, 'd');                                                   \
-  strcpy(wreq->ifr_ifrn.ifrn_name, name);                                      \
-  ret = ioctl(sockfd, nr, wreq);                                               \
-  VERIFY_GUARD(wreq);                                                          \
-  err = errno;                                                                 \
-  atomic_printf(nr_str "(ret:%d): %s: ", ret, wreq->ifr_name);                 \
-  if (-1 == ret) {                                                             \
-    atomic_printf("WARNING: %s doesn't support ioctl\n", name);                \
-    test_assert(EOPNOTSUPP == err || EPERM == err || EINVAL == err);           \
-  } else {                                                                     \
+  if (generic_wireless_request_by_name_internal(sockfd, &wreq, name, nr,       \
+                                                nr_str) == 0) {                \
     atomic_printf args;                                                        \
     atomic_puts("");                                                           \
   }                                                                            \
@@ -182,6 +202,13 @@ int main(void) {
   GENERIC_REQUEST_BY_NAME(SIOCGIFMETRIC);
   atomic_printf("metric is %d\n", req->ifr_metric);
 
+  GENERIC_REQUEST_BY_NAME(SIOCGIFMAP);
+  atomic_printf("map is %llu,%llu,%u,%d,%d,%d\n",
+                (unsigned long long)req->ifr_map.mem_start,
+                (unsigned long long)req->ifr_map.mem_end,
+                req->ifr_map.base_addr, req->ifr_map.irq, req->ifr_map.dma,
+                req->ifr_map.port);
+
   memset(&req->ifr_metric, 0xff, sizeof(req->ifr_metric));
   ret = ioctl(sockfd, SIOCGIFMEM, req);
   VERIFY_GUARD(req);
@@ -223,15 +250,20 @@ int main(void) {
                   etc->phy_address, etc->maxtxpkt, etc->maxrxpkt);
   }
 
-  GENERIC_WIRELESS_PARAM_REQUEST_BY_NAME(SIOCGIWRATE, bitrate);
-
   GENERIC_WIRELESS_REQUEST_BY_NAME(SIOCGIWNAME,
                                    ("wireless protocol name:%s", wreq->u.name));
 
   GENERIC_WIRELESS_REQUEST_BY_NAME(SIOCGIWMODE,
                                    (" wireless mode:%d", wreq->u.mode));
 
-  ALLOCATE_GUARD(wreq, 'e');
+  GENERIC_WIRELESS_PARAM_REQUEST_BY_NAME(SIOCGIWSENS, sens);
+
+  GENERIC_WIRELESS_PARAM_REQUEST_BY_NAME(SIOCGIWRATE, bitrate);
+
+  /* Note that we use sizeof(struct ifreq) here, not iwreq, because of
+   * https://bugzilla.kernel.org/show_bug.cgi?id=195869
+   */
+  wreq = allocate_guard(sizeof(struct ifreq), 'e');
   strcpy(wreq->ifr_ifrn.ifrn_name, name);
   wreq->u.essid.length = sizeof(buf);
   wreq->u.essid.pointer = buf;
@@ -247,8 +279,6 @@ int main(void) {
     atomic_printf("wireless ESSID:%s\n", buf);
   }
 
-  GENERIC_WIRELESS_PARAM_REQUEST_BY_NAME(SIOCGIWSENS, sens);
-
   atomic_puts("EXIT-SUCCESS");
   return 0;
 }
